using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using UnityEngine.XR;
using Leap;

namespace JayJayWon
{
    public class BrowserAssist : MVRScript
    {
        public static BrowserUI mainBrowserUI =null;
        public static Vector3 mainBrowserVAMUIOffset = new Vector3();

        SuperController.OnAtomUIDRename atomRenameHandle;
        SuperController.OnAtomRemoved atomRemoveHandle;
        SuperController.OnAtomAdded atomAddedHandle;
        SuperController.OnSceneLoaded sceneLoadedHandle;

        private Canvas launchButtonCanvas;
        private Canvas topScreenLaunchButtonCanvas;

        private UIDynamicButton vamUILaunchButton;
        private UIDynamicButton vamTopMenuLaunchButton;

        private GameObject moreOptionsGameObject;

        private Button vamSceneBrowserButton;

        private bool isActive = false;
        public override void Init()
        {
            try
            {
                Globals.mvrScript = this;
                if (XRSettings.enabled) Globals.isVREnabled = true;
                else Globals.isVREnabled = false;

                if (containingAtom.name == "CoreControl" && containingAtom.type == "SessionPluginManager")
                {
                    int browserAssistCount = 0;
                    foreach (var plugin in PluginUtils.GetSessionPlugins())
                    {
                        if (plugin.name.EndsWith("_JayJayWon.BrowserAssist")) browserAssistCount++;
                        
                    }

                    if (browserAssistCount < 2)
                    {
                        atomRenameHandle = new SuperController.OnAtomUIDRename(this.OnAtomNameUpdate);
                        SuperController.singleton.onAtomUIDRenameHandlers += atomRenameHandle;

                        atomRemoveHandle = new SuperController.OnAtomRemoved(this.OnAtomRemoved);
                        SuperController.singleton.onAtomRemovedHandlers += atomRemoveHandle;

                        sceneLoadedHandle = new SuperController.OnSceneLoaded(OnSceneLoaded);
                        SuperController.singleton.onSceneLoadedHandlers += sceneLoadedHandle;

                        atomAddedHandle = new SuperController.OnAtomAdded(OnAtomAdded);
                        SuperController.singleton.onAtomAddedHandlers += atomAddedHandle;

                        // Initiate integration with Acidbubbles KeyBindings
                        SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

//                        SuperController.singleton.BroadcastMessage("OnBrowserAssistAvailable", this, SendMessageOptions.DontRequireReceiver);

                        Utils.CreatePluginDataFolder();

                        DateTime scanTime = DateTime.Now;
                        HashSet<int> notNewResourceTypes = VARPackageManifest.Init(scanTime);
                        ResourceManifest.LocalResourcesInit(scanTime, notNewResourceTypes);

                        BAPluginUI.InitUI();

                        Settings.LoadSettings();

                        mainBrowserUI = new BrowserUI();
                        mainBrowserUI.resourceBrowserUI.RefactorButtonRows((int)Settings.maxButtonsPerPage, 6);

                        CreateVAMUILaunchButton();
                        CreateTopScreenLaunchButton();

                        CopySymLinkBatScripts();
                        CheckSymLinkAccess();

                        DateTime initCompleteTime = DateTime.Now;

                        if (Globals.debugLoadTimingEnabled) SuperController.LogMessage("TOTAL PLUGIN STARTUP TIME TAKEN " + (initCompleteTime - scanTime).TotalSeconds.ToString("0.00") + " seconds");

                        isActive = true;
                    }
                    else
                    {
                        JSONStorableString text = new JSONStorableString("text", "Only one instance of Browser Assist can be active at a time.");
                        CreateTextField(text);
                    }
                    
                }
                else
                {
                    JSONStorableString text = new JSONStorableString("text", "Browser Assist must be loaded as a Session Plugin");
                    CreateTextField(text);
                }

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Init)}: {e}");
                OnDestroy();
            }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
            jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
            jc["BAFixVersion"].AsInt = Const.BAFixVersion;
            jc["BAInteropVersion"].AsInt = Const.BAInteropVersion;

            return jc;
        }
        private void CreateTopScreenLaunchButton()
        {
            GameObject canvasLaunchButtonObject = new GameObject();
            topScreenLaunchButtonCanvas = canvasLaunchButtonObject.AddComponent<Canvas>();
            topScreenLaunchButtonCanvas.renderMode = RenderMode.WorldSpace;
            topScreenLaunchButtonCanvas.pixelPerfect = false;
            canvasLaunchButtonObject.layer = SuperController.singleton.topWorldUI.gameObject.layer;

            topScreenLaunchButtonCanvas.transform.SetParent(SuperController.singleton.topWorldUI, false);
            CanvasScaler topScreenLaunchButtonCS = canvasLaunchButtonObject.AddComponent<CanvasScaler>();
            topScreenLaunchButtonCS.scaleFactor = 100f;
            topScreenLaunchButtonCS.dynamicPixelsPerUnit = 1f;
            canvasLaunchButtonObject.AddComponent<GraphicRaycaster>();

            vamTopMenuLaunchButton = UI.CreateUIButton(topScreenLaunchButtonCanvas.gameObject, 100f, 100f, "Browser Assist", 40);
            vamTopMenuLaunchButton.buttonColor = Color.white;
            string thumbnailPath = Utils.GetLatestVARPath("JayJayWon.BrowserAssist.latest:/Custom/Scripts/JayJayWon/BrowserAssist/BrowserAssist.jpg");
            ImageUtils.QueueLoadTexture(thumbnailPath, TopScreenLaunchButtonImageLoadedCallback, true);
            vamTopMenuLaunchButton.button.onClick.AddListener(delegate
            {
                mainBrowserUI.ToggleActive();
            });

            SuperController.singleton.AddCanvas(topScreenLaunchButtonCanvas);

            topScreenLaunchButtonCanvas.transform.localPosition = new Vector3(-170f, -800f, 0f);

            ToggleTopMenuLanchButton();
        }
        private void TopScreenLaunchButtonImageLoadedCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi.tex != null)
            {
                
                Texture2D tex = qi.tex;
                vamTopMenuLaunchButton.button.image.sprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
                vamTopMenuLaunchButton.buttonText.text = "";
            }
        }
        public void ToggleTopMenuLanchButton()
        {
            if (Settings.enableTopScreenLaunchButtonJSB.val) topScreenLaunchButtonCanvas.gameObject.SetActive(true);
            else topScreenLaunchButtonCanvas.gameObject.SetActive(false);
        }

        private void CreateVAMUILaunchButton()
        {
            var vamSceneBrowserTransform = SuperController.singleton.mainHUD.Find("LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Toolbar/AutoLayoutArea/ButtonOpenSceneFullDialog");
            vamSceneBrowserButton = vamSceneBrowserTransform.GetComponent<Button>();
            

            var moreOptionstransform = SuperController.singleton.mainHUD.Find("LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Panel/ClickForMoreOptions/OtherOptionsPopup");

            moreOptionsGameObject = moreOptionstransform.gameObject;

            GameObject canvasLaunchButtonObject = new GameObject();
            launchButtonCanvas = canvasLaunchButtonObject.AddComponent<Canvas>();
            launchButtonCanvas.renderMode = RenderMode.WorldSpace;
            launchButtonCanvas.pixelPerfect = false;

            canvasLaunchButtonObject.layer = SuperController.singleton.mainHUD.gameObject.layer;
            launchButtonCanvas.transform.SetParent(SuperController.singleton.mainHUD, false);
            CanvasScaler launchButtonCS = canvasLaunchButtonObject.AddComponent<CanvasScaler>();
            launchButtonCS.scaleFactor = 100f;
            launchButtonCS.dynamicPixelsPerUnit = 1f;
            GraphicRaycaster gr = canvasLaunchButtonObject.AddComponent<GraphicRaycaster>();

            vamUILaunchButton = UI.CreateUIButton(launchButtonCanvas.gameObject, 38f, 36f, "Browser Assist", 20);
            vamUILaunchButton.buttonColor = Color.white;
            string thumbnailPath = Utils.GetLatestVARPath("JayJayWon.BrowserAssist.latest:/Custom/Scripts/JayJayWon/BrowserAssist/BrowserAssist.jpg");
            ImageUtils.QueueLoadTexture(thumbnailPath, LaunchButtonImageLoadedCallback, true);
            vamUILaunchButton.button.onClick.AddListener(delegate
            {
                mainBrowserUI.ToggleActive();
            });
            vamUILaunchButton.button.gameObject.AddComponent <LaunchButtonPointerEnterExitHelp>();
            SuperController.singleton.AddCanvas(launchButtonCanvas);

            launchButtonCanvas.transform.localScale = new Vector3(0.001f, 0.001f, 0.001f);
            RePositionLaunchButton();

        }

        public void RePositionLaunchButton()
        {
            vamSceneBrowserButton.interactable = true;
            launchButtonCanvas.gameObject.SetActive(true);
            switch (Settings.launchButtonPlacementJSESC.val)
            {
                case VAMUILaunchButtonPosition.toolBarGap:
                    launchButtonCanvas.transform.localPosition = new Vector3(-0.131f, -0.074f, 0f);
                    break;
                case VAMUILaunchButtonPosition.sceneBrowserSub:
                    launchButtonCanvas.transform.localPosition = new Vector3(0.2205f, -0.074f, 0f);
                    vamSceneBrowserButton.interactable = false;
                    break;
                case VAMUILaunchButtonPosition.toolBarLeftInnerUpper:
                    launchButtonCanvas.transform.localPosition = new Vector3(0.37f, -0.074f, 0f);
                    break;
                case VAMUILaunchButtonPosition.toolBarRightInnerUpper:
                    launchButtonCanvas.transform.localPosition = new Vector3(-0.37f, -0.074f, 0f);
                    break;
                case VAMUILaunchButtonPosition.disabled:
                    launchButtonCanvas.gameObject.SetActive(false);
                    break;
            }

            if (Globals.isVREnabled)
            {
                launchButtonCanvas.transform.localEulerAngles = new Vector3(32, 180, 0);
                launchButtonCanvas.transform.Translate(new Vector3(0f, -0.01f * mainBrowserUI.worldScaleAdjust, -0.039f * mainBrowserUI.worldScaleAdjust));
            }
            else
            {
                launchButtonCanvas.transform.localEulerAngles = new Vector3(0, 180, 0);
            }
        }

        private void LaunchButtonImageLoadedCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi.tex != null)
            {
                Texture2D tex = qi.tex;
                vamUILaunchButton.button.image.sprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
                vamUILaunchButton.buttonText.text = "";
            }
        }

        private void CopySymLinkBatScripts()
        {
            try
            {
                string pluginSlot = name.Substring(0, name.IndexOf('_'));               
                JSONClass pluginsJC = manager.GetJSON(true, true, true).AsObject["plugins"].AsObject;

                string pluginPath = pluginsJC[pluginSlot];
                if (pluginPath.Contains(":"))
                {
                    string varName = pluginPath.Substring(0, pluginPath.IndexOf(':'));
                    string createSymLinksSourceFileName = FileManagerSecure.NormalizePath(varName + ":\\" + Const.BAScriptFolder + "\\CreateBASymLinks.bat");
                    string removeSymLinksSourceFileName = FileManagerSecure.NormalizePath(varName + ":\\" + Const.BAScriptFolder + "\\RemoveBASymLinks.bat");

                    if (FileManagerSecure.FileExists(createSymLinksSourceFileName))
                    {
                        FileManagerSecure.DeleteFile(Const.BAScriptDataFolder + "\\CreateBASymLinks.bat");
                        FileManagerSecure.CopyFile(createSymLinksSourceFileName, Const.BAScriptDataFolder + "\\CreateBASymLinks.bat");
                    }
                    if (FileManagerSecure.FileExists(removeSymLinksSourceFileName))
                    {
                        FileManagerSecure.DeleteFile(Const.BAScriptDataFolder + "\\RemoveBASymLinks.bat");
                        FileManagerSecure.CopyFile(removeSymLinksSourceFileName, Const.BAScriptDataFolder + "\\RemoveBASymLinks.bat");
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(CopySymLinkBatScripts)}: {e}");
            }
        }
        private void CheckSymLinkAccess()
        {
            try
            {
                string guid = SuperController.singleton.GetTempUID();
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\Custom"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\Custom\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("Custom\\" + guid))
                    {
                        Globals.customFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\Custom\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\Saves\\scene"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\Saves\\scene\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("Saves\\scene\\" + guid))
                    {
                        Globals.scenesFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\Saves\\scene\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\AddonPackagesFilePrefs"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("AddonPackagesFilePrefs\\" + guid))
                    {
                        Globals.addonPackageFilePrefsFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\AddonPackages"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\AddonPackages\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("AddonPackages\\" + guid))
                    {
                        Globals.addonPackagesAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\AddonPackages\\" + guid);
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(CheckSymLinkAccess)}: {e}");
            }
            
        }

        public void OnBrowserAssistConsumerAvailable(MVRScript consumer)
        {
 //           SuperController.LogMessage("BA: OnBrowserAssistConsumerAvailable");

            consumer.SendMessage("OnBrowserAssistAvailable", this, SendMessageOptions.DontRequireReceiver);
        }

        public void OnGetTags(object[] objects)
        {
            JSONStorableString exceptionCallback = objects[4] as JSONStorableString;
            try
            {
                int resourceType = (int)objects[0];
                List<string> tagsList = objects[1] as List<string>;
                JSONStorableAction successCallback = objects[2] as JSONStorableAction;
                JSONStorableAction failCallback = objects[3] as JSONStorableAction;

                tagsList.Clear();
                foreach (var tag in TagManifest.GetTags(resourceType))
                {
                    tagsList.Add(tag.tagName);
                }

                successCallback.actionCallback.Invoke();
            }
            catch (Exception e)
            {
                exceptionCallback.val = e.ToString();
            }
        }

        public void OnBindingsListRequested(List<object> bindings)
        {
            try
            {
                int pluginSlot = int.Parse(this.name.Substring(7, this.name.IndexOf('_') - 7));

                bindings.Add(new JSONStorableAction("Open UI", () =>
                {
                    mainBrowserUI.SetActive();
                }
                ));

                bindings.Add(new JSONStorableAction("Close UI", () =>
                {
                    mainBrowserUI.SetInactive();
                }
                ));

                bindings.Add(new JSONStorableAction("Toggle UI Open/Close", () =>
                {
                    mainBrowserUI.ToggleActive();
                }
                ));

                bindings.Add(new JSONStorableAction("Open UI: Scenes", () =>
                {
                    mainBrowserUI.SetActive();
                    mainBrowserUI.browserFiltersUI.resourceType = ResourceType.scene;
                }
                ));

                bindings.Add(new JSONStorableAction("Open UI: Presets Appearance", () =>
                {
                    mainBrowserUI.SetActive();
                    mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetAppearance;
                }
                ));

                if (PC.patreonContentEnabled)
                {
                    bindings.Add(new JSONStorableAction("Open UI: Presets Breast Physics", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetBreastPhys;
                    }
               ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Clothing", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetClothing;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets General", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetGeneral;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Glute Phys", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetGlutePhysics;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Hair", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetHair;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Morphs", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetMorphs;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Plugins", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetPlugins;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Pose", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetPose;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Skin", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.presetSkin;
                    }
                    ));

                    bindings.Add(new JSONStorableAction("Open UI: Plugins", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.resourceType = ResourceType.plugins;
                    }
                    ));
                }
               

                bindings.Add(new JSONStorableAction("Toggle 'Only Favorites' filter", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.browserFiltersUI.onlyFavouritesJSB.val = !mainBrowserUI.browserFiltersUI.onlyFavouritesJSB.val;
                }
                ));

                bindings.Add(new JSONStorableAction("Toggle 'Show Hidden' filter", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.browserFiltersUI.showHiddenJSB.val = !mainBrowserUI.browserFiltersUI.showHiddenJSB.val;
                }
                ));

                bindings.Add(new JSONStorableAction("Toggle 'Only New' filter", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.browserFiltersUI.onlyNewJSB.val = !mainBrowserUI.browserFiltersUI.onlyNewJSB.val;
                }
                ));
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnBindingsListRequested)}: {e}");
            }
        }

        
        private void OnSceneLoaded()
        {
            try
            {
                if (mainBrowserUI!=null && mainBrowserUI.isVREnabled) mainBrowserUI.SetInactive();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnSceneLoaded)}: {e}");
            }
        }
        private void OnAtomAdded(Atom atom)
        {
         
            try
            {
                if (mainBrowserUI != null) mainBrowserUI.resourceActionsUI.OnAtomAdded(atom);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomAdded)}: {e}");
            }
        }

        private void OnAtomNameUpdate(string oldName, string newName)
        {           
            try
            {
                if (mainBrowserUI != null) mainBrowserUI.resourceActionsUI.AtomNameUpdated(oldName, newName);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomNameUpdate)}: {e}");
            }
        }

        private void OnAtomRemoved(Atom atom)
        {
            try
            {
                if (mainBrowserUI != null) mainBrowserUI.resourceActionsUI.OnAtomRemoved(atom);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomRemoved)}: {e}");
            }
        }

        public void OnEnable()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnEnable)}: {e}");
            }
        }

        public void OnDisable()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnDisable)}: {e}");
            }
        }

        public void OnDestroy()
        {
            try
            {
                if (isActive)
                {
                    SuperController.singleton.onAtomUIDRenameHandlers -= atomRenameHandle;
                    SuperController.singleton.onAtomRemovedHandlers -= atomRemoveHandle;
                    SuperController.singleton.onAtomAddedHandlers -= atomAddedHandle;
                    SuperController.singleton.onSceneLoadedHandlers -= sceneLoadedHandle;

                    //Deregisters with Acidbubbles KeyBindings plugin
                    SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);

                    SuperController.singleton.BroadcastMessage("OnBrowserAssistDestroyed", SendMessageOptions.DontRequireReceiver);

                    if (SuperController.singleton != null)
                    {
                        SuperController.singleton.RemoveCanvas(launchButtonCanvas);
                        SuperController.singleton.RemoveCanvas(topScreenLaunchButtonCanvas);
                    }
                    if (launchButtonCanvas != null) launchButtonCanvas.transform.SetParent(null, false);
                    if (launchButtonCanvas.gameObject != null) GameObject.Destroy(launchButtonCanvas.gameObject);
                    if (topScreenLaunchButtonCanvas != null) topScreenLaunchButtonCanvas.transform.SetParent(null, false);
                    if (topScreenLaunchButtonCanvas.gameObject != null) GameObject.Destroy(topScreenLaunchButtonCanvas.gameObject);

                    vamSceneBrowserButton.interactable = true;

                    if (mainBrowserUI != null) mainBrowserUI.OnDestroy();
                    ResourceManifest.OnDestoy();
                    Resources.UnloadUnusedAssets();
                    GC.Collect();
                }

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnDestroy)}: {e}");
            }
        }


        void Start()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Start)}: {e}");
            }
        }

        void Update()
        {
            try
            {
                if (isActive)
                {
                    if (vamUILaunchButton != null && moreOptionsGameObject != null)
                    {
                        if (moreOptionsGameObject.activeSelf) vamUILaunchButton.gameObject.SetActive(false);
                        else vamUILaunchButton.gameObject.SetActive(true);
                    }

                    if (mainBrowserUI != null) mainBrowserUI.Update();

                    DataFileManifest.Update();
                }

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Update)}: {e}");
            }

        }
    }

}
