﻿
using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;



namespace JayJayWon
{
    public static class BAPluginUI
    {
        private static CustomRowGridlessUIElement topTabMenuOptions;

        private static int currentTopTabSelection = TopUITabs.generalSettings;

        private static BaseUITab currentTabUI;

        private static GeneralUITab generalSettingsUITab;
        private static DesktopUITab desktopSettingsUITab;
        private static VRSettingsUITab vrSettingsUITab;
        private static ResourceSettingsUITab resourceSettingsUITab;

        private static UIDynamicButton generalSettingsUID;
        private static UIDynamicButton VRSettingsUID;
        private static UIDynamicButton desktopSettingsUID;
        private static UIDynamicPopup resourceSettingsUID;

        private static JSONStorableEnumStringChooser resourceSettingsJSESC;

        public static void InitUI()
        {
            topTabMenuOptions = new CustomRowGridlessUIElement(80);

            generalSettingsUITab = new GeneralUITab();
            desktopSettingsUITab = new DesktopUITab();
            vrSettingsUITab = new VRSettingsUITab();
            resourceSettingsUITab = new ResourceSettingsUITab();
            topTabMenuOptions.InitUI(CustomRowWidth.full);

            resourceSettingsJSESC = new JSONStorableEnumStringChooser("resourceSettings", ResourceType.enumManifestName, ResourceType.scene, "Resource\nSettings", ResourceSettingsTabSelected);

            generalSettingsUID = topTabMenuOptions.AddGridlessButtonToRow("Global\nSettings", TopUITabs.generalSettings, TopTabSelected,200,0,0);
            VRSettingsUID = topTabMenuOptions.AddGridlessButtonToRow("VR\nSettings", TopUITabs.vrSettings, TopTabSelected, 200, 200, 0);
            desktopSettingsUID = topTabMenuOptions.AddGridlessButtonToRow("Desktop\nSettings", TopUITabs.desktopSettings, TopTabSelected, 200, 400,0);
            
            resourceSettingsUID = topTabMenuOptions.AddGridlessPopupToRow(resourceSettingsJSESC, null, 450,600,0,800,false);
            resourceSettingsUID.popup.currentValue = "Select...";

            generalSettingsUITab.RefreshGroupUI(1);

            topTabMenuOptions.SetButtonColorActive(currentTopTabSelection);

            currentTabUI = generalSettingsUITab;
        }

        private static void ResourceSettingsTabSelected(int val)
        {
            if (currentTopTabSelection!= TopUITabs.resourceSettings)
            {
                currentTopTabSelection = TopUITabs.resourceSettings;
                ClearCurrentTopTabsUI();
                
                currentTabUI = resourceSettingsUITab;

                
            }
            resourceSettingsUITab.resourceType = val;

            if (PC.patreonContentEnabled || val == ResourceType.scene || val == ResourceType.presetAppearance) BrowserAssist.mainBrowserUI.browserFiltersUI.resourceType = val;
            resourceSettingsUITab.RefreshGroupUI(1);
        }
        private static void TopTabSelected(int tabRef)
        {
            if (tabRef != currentTopTabSelection)
            {
                ClearCurrentTopTabsUI();
                switch (tabRef)
                {
                    case TopUITabs.generalSettings:
                        generalSettingsUITab.RefreshGroupUI(1);
                        currentTabUI = generalSettingsUITab;
                        break;
                    case TopUITabs.vrSettings:
                        vrSettingsUITab.RefreshGroupUI(1);
                        currentTabUI = vrSettingsUITab;
                        break;
                    case TopUITabs.desktopSettings:
                        desktopSettingsUITab.RefreshGroupUI(1);
                        currentTabUI = desktopSettingsUITab;
                        break;

                }
                currentTopTabSelection = tabRef;
            }
        }

        private static void ClearCurrentTopTabsUI()
        {
            currentTabUI.ClearAllGroupsUI();

        }

        public static void CloseAllPopups(UIDynamicPopup exception)
        {
            currentTabUI.CloseAllPopups(exception);
        }
    }

    public class GeneralUITab : BaseUITab
    {
        private CustomUIElement generalSettingsCustomUI;
        private CustomUIElement helpTextCustomUI;
        
        public GeneralUITab()
        {
            generalSettingsCustomUI = new CustomUIElement(890);
            helpTextCustomUI = new CustomUIElement(200);
            defaultHelpText = "The Global Settings tab contain settings that impact general behavior of the Browser Assist UI - which are not specific to VR, Desktop or Resource types.";
            helpTextJSS.val = defaultHelpText;
            
        }

        protected override void CreateGroupUI(int groupID)
        {
            generalSettingsCustomUI.InitUI(CustomRowWidth.full);

            var launchBAToggle = generalSettingsCustomUI.AddToggle(Settings.enableTopScreenLaunchButtonJSB, 1020, 50, 0, 0, "Enable BrowserAssist Launch button on VAM Top Menu");
            var launchButtonPlacementPopup = generalSettingsCustomUI.AddPopup(Settings.launchButtonPlacementJSESC, null, 1020, 80, 0, -60, labelProportion: 0.7f);
            var defaultResourceTypePopup = generalSettingsCustomUI.AddPopup(Settings.defaultResourceTypeJESC, null, 1020, 80, 0, -150, labelProportion: 0.7f);
            var resourceNameSearchModePopup = generalSettingsCustomUI.AddPopup(Settings.resourceNameSearchModeJSESC, null, 1020, 80, 0, -240, labelProportion: 0.7f);
            var defaultTargetPersonType = generalSettingsCustomUI.AddPopup(Settings.defaultTargetPersonTypeJSE, null, 1020, 80, 0, -330, labelProportion: 0.7f);
            var creatorFilterSortOrder = generalSettingsCustomUI.AddPopup(Settings.creatorFilterOrderJSEC, null, 1020, 80, 0, -420, labelProportion: 0.7f);

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            ApplyHelp(launchBAToggle, "When this setting is enabled, a Browser Assist launch button will be displayed embeded into the VAM 'Top Menu' UI. The 'Top Menu' is the UI which is displayed when VAM first starts. It can also be accessed by pressing the star icon on the tool bar in the main VAM UI.");

            ApplyHelp(launchButtonPlacementPopup, "The launch button placement setting determines where the main BrowserAssist launch button is placed on the VAM UI toolbar. Options allow it to be placed over the VAM Scene Browser button (effectively replacing it) or to the left or right of the toolbar.");

            ApplyHelp(defaultResourceTypePopup, "This setting determines which Resource Type (e.g. Scene or Appearance preset) is selected by default when Browser Assist first starts.");

            ApplyHelp(resourceNameSearchModePopup, "The 'Resource Filter Name Search Mode' setting determines the behaviour of the resource name search filter in Browser Assist. The 'Starts With' option will filter to resources where the resource name starts with the search text. The two 'Contains' option will filter to resource names that contain the search text. 'Contains (Full Path)' will search for any matching text in the entire path name including the package name and folders.");

            ApplyHelp(defaultTargetPersonType, "This setting determines the default person gender when selecting a resource that has a person as it's target (e.g. Appearance Presets will select either a male or female Atom where available in the current scene). The setting can be overridden by setting the 'male' or 'female' tag on individual resources.");

            ApplyHelp(creatorFilterSortOrder, "This setting determines the order which VAR Creators are listed in the Creator filter dropdown. This can either be alphabetical or by the quantity  of resources within VARs packaged by that Creator.");

            RegisterCustomUIElement(groupID, helpTextCustomUI);

            helpTextBox.backgroundColor = new Color(0.4f,0.5f,0.2f,0.4f);

            RegisterCustomUIElement(groupID, generalSettingsCustomUI);
        }
    }

    public class DesktopUITab : BaseUITab
    {
        private CustomUIElement desktopSettingsCustomUI;
        private CustomUIElement helpTextCustomUI;
        public DesktopUITab()
        {
            desktopSettingsCustomUI = new CustomUIElement(890);
            helpTextCustomUI = new CustomUIElement(200);
            defaultHelpText = "The Desktop Settings tab contain settings that impact behavior of the Browser Assist UI when in Desktop mode (as opposed to VR mode)";
            helpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(int groupID)
        {
            desktopSettingsCustomUI.InitUI(CustomRowWidth.full);

            var uiScaleSlider = desktopSettingsCustomUI.AddSlider(Settings.uiScaleJSF, 1020, 120, 0, 0);
            uiScaleSlider.valueFormat = "F2";
            uiScaleSlider.ConfigureQuickButtons(-0.1f, -0.25f, -0.5f, -1f, 0.1f, 0.25f, 0.5f, 1f);

            var xPosSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.positionXJSF, 500, 120, 0, -130);
            xPosSlider.valueFormat = "F3";
            xPosSlider.label = "New Atom: Default Position X";

            var yPosSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.positionYJSF, 500, 120, 0, -250);
            yPosSlider.valueFormat = "F3";
            yPosSlider.label = "New Atom: Default Position Y";

            var zPosSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.positionZJSF, 500, 120, 0, -370);
            zPosSlider.valueFormat = "F3";
            zPosSlider.label = "New Atom: Default Position Z";

            var xRotSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.rotationXJSF, 500, 120, 510, -130);
            xRotSlider.valueFormat = "F3";
            xRotSlider.label = "New Atom: Default Rotation X";

            var yRotSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.rotationYJSF, 500, 120, 510, -250);
            yRotSlider.valueFormat = "F3";
            yRotSlider.label = "New Atom: Default Rotation Y";

            var zRotSlider = desktopSettingsCustomUI.AddSlider(Settings.desktopDefaultNewAtomPosRot.rotationZJSF, 500, 120, 510, -370);
            zRotSlider.valueFormat = "F3";
            zRotSlider.label = "New Atom: Default Rotation Z";

            var xPosAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.posXRelAbsJSE,null,500,60,0,-490,140,60,"X Pos: Relative or Absolute",0.8f);

            var yPosAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.posYRelAbsJSE, null, 500, 60, 0, -550, 140, 60, "Y Pos: Relative or Absolute", 0.8f);

            var zPosAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.posZRelAbsJSE, null, 500, 60, 0, -610, 140, 60, "Z Pos: Relative or Absolute", 0.8f);

            var xRotAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.rotXRelAbsJSE, null, 500, 60, 510, -490, 140, 60, "X Rot: Relative or Absolute", 0.8f);

            var yRotAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.rotYRelAbsJSE, null, 500, 60, 510, -550, 140, 60, "Y Rot: Relative or Absolute", 0.8f);

            var zRotAbsRelPopup = desktopSettingsCustomUI.AddPopup(Settings.desktopDefaultNewAtomPosRot.rotZRelAbsJSE, null, 500, 60, 510, -610, 140, 60, "Z Rot: Relative or Absolute", 0.8f);

            var maxButtonsSlider = desktopSettingsCustomUI.AddSlider(Settings.maxButtonsPerPageDesktopJSF, 1020, 120, 0, -670);
            maxButtonsSlider.labelText.text = "Maximum Number of Resources per Browser page";
            maxButtonsSlider.valueFormat = "F0";
            maxButtonsSlider.ConfigureQuickButtons(-10f, -50f, -100f, -1000f, 10f, 50f, 100f, 1000f);

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            ApplyHelp(uiScaleSlider, "The 'Desktop UI Scale' setting determines the scale of Browser Assist UI when in Desktop mode. Higher scales numbers make the UI larger and lower scale makes the UI smaller. ");

            ApplyHelp(xPosSlider, "The 'New Atom Default Position X' setting determines the default X coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the desktop view point. If the position is relative, the X coordinate will offset the atom position to the left or right relative to the direction the user is facing.");

            ApplyHelp(yPosSlider, "The 'New Atom Default Position Y' setting determines the default Y coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the desktop view point. If the position is relative, the Y coordinate will offset the atom position to the up or down relative to the direction the user is facing.");

            ApplyHelp(zPosSlider, "The 'New Atom Default Position Z' setting determines the default Z coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the desktop view point. If the position is relative, the Z coordinate will offset the atom position to the further away or closer relative to the users view point.");

            ApplyHelp(xRotSlider, "The 'New Atom Default Rotation X' setting determines the default X rotation (Pitch) for new atoms created by Browser Assist. The rotation can be absolute or relative to the desktop view point. If the rotation is relative, the X rotation will offset the atom's pitch relative to the direction the user is facing.");

            ApplyHelp(yRotSlider, "The 'New Atom Default Rotation Y' setting determines the default Y rotation (Yaw) for new atoms created by Browser Assist. The rotation can be absolute or relative to the desktop view point. If the rotation is relative, the Y rotation will offset the atom's yaw relative to the direction the user is facing.");

            ApplyHelp(zRotSlider, "The 'New Atom Default Rotation Z' setting determines the default Z rotation (Roll) for new atoms created by Browser Assist. The rotation can be absolute or relative to the desktop view point. If the rotation is relative, the Z rotation will offset the atom's roll relative to the direction the user is facing.");

            ApplyHelp(xPosAbsRelPopup, "The 'X Pos: Relative or Absolute' setting determines whether default X position for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(yPosAbsRelPopup, "The 'Y Pos: Relative or Absolute' setting determines whether default Y position for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(zPosAbsRelPopup, "The 'Z Pos: Relative or Absolute' setting determines whether default Z position for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(xRotAbsRelPopup, "The 'X Rot: Relative or Absolute' setting determines whether default X rotation (Pitch) for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(yRotAbsRelPopup, "The 'Y Rot: Relative or Absolute' setting determines whether default Y rotation (Yaw) for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(zRotAbsRelPopup, "The 'Z Rot: Relative or Absolute' setting determines whether default Z rotation (Roll) for new atoms (created by Browser Assist) is absolute or relative to the desktop view point.");

            ApplyHelp(maxButtonsSlider, "This setting determines the maximum number of resources icons that will be displayed on a single page of browser results, whilst in Desktop mode.");

            RegisterCustomUIElement(groupID, desktopSettingsCustomUI);
            RegisterCustomUIElement(groupID, helpTextCustomUI);
        }
    }

    public class VRSettingsUITab : BaseUITab
    {
        private CustomUIElement vrSettingsCustomUI;
        private CustomUIElement helpTextCustomUI;
        public VRSettingsUITab()
        {
            vrSettingsCustomUI = new CustomUIElement(890);
            helpTextCustomUI = new CustomUIElement(200);
            defaultHelpText = "The VR Settings tab contain settings that impact behavior of the Browser Assist UI when in VR mode (as opposed to Desktop mode)";
            helpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(int groupID)
        {
            vrSettingsCustomUI.InitUI(CustomRowWidth.full);

            var dockVAMUIToggle = vrSettingsCustomUI.AddToggle(Settings.dockToVAMUIJSB, 500, 60, 0, 0);
            dockVAMUIToggle.label = "Dock BrowserAssist to VAM UI";

            var xPosSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.positionXJSF, 500, 120, 0, -130);
            xPosSlider.valueFormat = "F3";
            xPosSlider.label = "New Atom: Default Position X";

            var yPosSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.positionYJSF, 500, 120, 0, -250);
            yPosSlider.valueFormat = "F3";
            yPosSlider.label = "New Atom: Default Position Y";

            var zPosSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.positionZJSF, 500, 120, 0, -370);
            zPosSlider.valueFormat = "F3";
            zPosSlider.label = "New Atom: Default Position Z";

            var xRotSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.rotationXJSF, 500, 120, 510, -130);
            xRotSlider.valueFormat = "F3";
            xRotSlider.label = "New Atom: Default Rotation X";

            var yRotSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.rotationYJSF, 500, 120, 510, -250);
            yRotSlider.valueFormat = "F3";
            yRotSlider.label = "New Atom: Default Rotation Y";

            var zRotSlider = vrSettingsCustomUI.AddSlider(Settings.vrDefaultNewAtomPosRot.rotationZJSF, 500, 120, 510, -370);
            zRotSlider.valueFormat = "F3";
            zRotSlider.label = "New Atom: Default Rotation Z";

            var xPosAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.posXRelAbsJSE, null, 500, 60, 0, -490, 140, 60, "X Pos: Relative or Absolute", 0.8f);

            var yPosAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.posYRelAbsJSE, null, 500, 60, 0, -550, 140, 60, "Y Pos: Relative or Absolute", 0.8f);

            var zPosAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.posZRelAbsJSE, null, 500, 60, 0, -610, 140, 60, "Z Pos: Relative or Absolute", 0.8f);

            var xRotAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.rotXRelAbsJSE, null, 500, 60, 510, -490, 140, 60, "X Rot: Relative or Absolute", 0.8f);

            var yRotAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.rotYRelAbsJSE, null, 500, 60, 510, -550, 140, 60, "Y Rot: Relative or Absolute", 0.8f);

            var zRotAbsRelPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.rotZRelAbsJSE, null, 500, 60, 510, -610, 140, 60, "Z Rot: Relative or Absolute", 0.8f);

            var relAtomPopup = vrSettingsCustomUI.AddPopup(Settings.vrDefaultNewAtomPosRot.relativeAtomNameJSME, null,500, 60, 510, -680, 200, 60,"Relative to:", 0.6f);

            var maxButtonsSlider = vrSettingsCustomUI.AddSlider(Settings.maxButtonsPerPageVRJSF, 1020, 120, 0, -750);
            maxButtonsSlider.labelText.text = "Maximum Number of Resources per Browser page";
            maxButtonsSlider.valueFormat = "F0";
            maxButtonsSlider.ConfigureQuickButtons(-10f, -50f, -100f, -1000f, 10f, 50f, 100f, 1000f);

            helpTextCustomUI.InitUI(CustomRowWidth.full);

            var helpTextBox = helpTextCustomUI.AddTextBox(helpTextJSS, 1050f, 200f, 0f, 0f);
            helpTextBox.textColor = Color.white;
            helpTextBox.backgroundColor = new Color(0.4f, 0.5f, 0.2f, 0.4f);

            ApplyHelp(dockVAMUIToggle, "This setting determines if Browser Assist window is docked to VAM Main UI when in VR. If enabled, the Browser Assist window will behave simlarly to other settings available in VAM UI and can be repostioned using the VR controller B button.");

            ApplyHelp(xPosSlider, "The 'New Atom Default Position X' setting determines the default X coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the VR headset or VR controllers. If the position is relative, the X coordinate will offset the atom position to the left or right relative to the direction the user is facing.");

            ApplyHelp(yPosSlider, "The 'New Atom Default Position Y' setting determines the default Y coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the VR headset or VR controllers. If the position is relative, the Y coordinate will offset the atom position to the up or down relative to the direction the user is facing.");

            ApplyHelp(zPosSlider, "The 'New Atom Default Position Z' setting determines the default Z coordinate for new atoms created by Browser Assist. The position can be absolute or relative to the VR headset or VR controllers. If the position is relative, the Z coordinate will offset the atom position to the further away or closer relative to the users view point.");

            ApplyHelp(xRotSlider, "The 'New Atom Default Rotation X' setting determines the default X rotation (Pitch) for new atoms created by Browser Assist. The rotation can be absolute or relative to the VR headset or VR controllers. If the rotation is relative, the X rotation will offset the atom's pitch relative to the direction the user is facing.");

            ApplyHelp(yRotSlider, "The 'New Atom Default Rotation Y' setting determines the default Y rotation (Yaw) for new atoms created by Browser Assist. The rotation can be absolute or relative to the VR headset or VR controllers. If the rotation is relative, the Y rotation will offset the atom's yaw relative to the direction the user is facing.");

            ApplyHelp(zRotSlider, "The 'New Atom Default Rotation Z' setting determines the default Z rotation (Roll) for new atoms created by Browser Assist. The rotation can be absolute or relative to the VR headset or VR controllers. If the rotation is relative, the Z rotation will offset the atom's roll relative to the direction the user is facing.");

            ApplyHelp(xPosAbsRelPopup, "The 'X Pos: Relative or Absolute' setting determines whether default X position for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(yPosAbsRelPopup, "The 'Y Pos: Relative or Absolute' setting determines whether default Y position for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(zPosAbsRelPopup, "The 'Z Pos: Relative or Absolute' setting determines whether default Z position for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(xRotAbsRelPopup, "The 'X Rot: Relative or Absolute' setting determines whether default X rotation (Pitch) for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(yRotAbsRelPopup, "The 'Y Rot: Relative or Absolute' setting determines whether default Y rotation (Yaw) for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(zRotAbsRelPopup, "The 'Z Rot: Relative or Absolute' setting determines whether default Z rotation (Roll) for new atoms (created by Browser Assist) is absolute or relative to the VR headset or VR controllers.");

            ApplyHelp(relAtomPopup, "If any of the atoms positions or rotations are set as 'Relative', then this setting determines whether that position or rotation is relative to the VR Headset, VR Left Controller or the VR right Controller.");

            ApplyHelp(maxButtonsSlider, "This setting determines the maximum number of resources icons that will be displayed on a single page of browser results, whilst in VR mode.");

            RegisterCustomUIElement(groupID, vrSettingsCustomUI);
            RegisterCustomUIElement(groupID, helpTextCustomUI);
        }
    }


}
