﻿using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using UnityEngine;

namespace JayJayWon
{
    public class VARPackageManifest
    {
        public static Dictionary<string, VARPackageManifestEntry> varPackages = new Dictionary<string, VARPackageManifestEntry>();

        public static Dictionary<string, VARPackageVersionGroupEntry> varPackageVersionGroups = new Dictionary<string, VARPackageVersionGroupEntry>();

        public static HashSet<int> Init(DateTime scanTime)
        {
            string[] legacyVARMetaFiles =null;
            if (FileManagerSecure.DirectoryExists(Const.VARLegacyCoreDataFolder)) legacyVARMetaFiles = FileManagerSecure.GetFiles(Const.VARLegacyCoreDataFolder, "*" + Const.CoreDataFileExt);

            string[] legacyVARUserFiles = null;
            if (FileManagerSecure.DirectoryExists(Const.VARLegacyUserDataFolder)) legacyVARUserFiles = FileManagerSecure.GetFiles(Const.VARLegacyUserDataFolder, "*" + Const.UserDataFileExt);

            string[] varGroupCoreFiles = FileManagerSecure.GetFiles(Const.VARGroupCoreDataFilesFolder, "*" + Const.CoreDataFileExt);
            string[] varResourceCoreFiles = FileManagerSecure.GetFiles(Const.VARResourcesCoreDataFilesFolder, "*" + Const.CoreDataFileExt);

            
            string[] varGroupUserFiles = FileManagerSecure.GetFiles(Const.VARGroupUserDataFilesFolder, "*" + Const.UserDataFileExt);
            string[] varResourceUserFiles = FileManagerSecure.GetFiles(Const.VARResourcesUserDataFilesFolder, "*" + Const.UserDataFileExt);

            HashSet<string> varFiles = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var",true);

            DateTime varFilesScanCompleteTime = DateTime.Now;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAR files found in ", varFiles.Count, (varFilesScanCompleteTime - scanTime).TotalSeconds);

            HashSet<string> varFileNames = new HashSet<string>() ;
            foreach (string varFile in varFiles)
            {
                if (FileManagerSecure.DirectoryExists(varFile + ":")) varFileNames.Add(FileManagerSecure.GetFileName(varFile));
            }

            HashSet<int> notNewResourceTypes = new HashSet<int>();
            if ((legacyVARMetaFiles!=null && legacyVARMetaFiles.Count() == 0) || (legacyVARMetaFiles == null && varGroupCoreFiles.Count() == 0)) notNewResourceTypes = new HashSet<int>(EnumManifest.GetEnumsList(ResourceType.enumManifestName));

            ReloadVARManifest(varGroupCoreFiles, varResourceCoreFiles, varFileNames, scanTime);
            DateTime varMetaFilesProcessCompleteTime = DateTime.Now;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" BA VAR Manifest files processed in ", varResourceCoreFiles.Count()+ varGroupCoreFiles.Count(), (varMetaFilesProcessCompleteTime - varFilesScanCompleteTime).TotalSeconds);

            bool legacyRestore = false;
            if (varGroupUserFiles.Count() == 0 && legacyVARUserFiles != null && legacyVARUserFiles.Count() > 0) legacyRestore = true;

            DateTime varUserFilesProcessCompleteTime = varMetaFilesProcessCompleteTime;
            if (!legacyRestore)
            {
                legacyRestore = false;
                ReloadGroupVARUserData(varGroupUserFiles);
                ResourceManifest.ReloadResourceVARUserData(varResourceUserFiles);

                varUserFilesProcessCompleteTime = DateTime.Now;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAR User files processed in ", varResourceUserFiles.Count(), (varUserFilesProcessCompleteTime - varMetaFilesProcessCompleteTime).TotalSeconds);
            }

            RefreshManifestWithNewVARs(varFiles, scanTime, notNewResourceTypes);
            DateTime varFilesProcessCompleteTime = DateTime.Now;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAM VAR files processed in ", varFiles.Count(), (varFilesProcessCompleteTime - varUserFilesProcessCompleteTime).TotalSeconds);

            if (legacyRestore)
            {
                ReloadLegacyVARUserData(legacyVARUserFiles);
                DateTime varLegacyUserFilesProcessCompleteTime = DateTime.Now;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAR User files processed in ", varResourceUserFiles.Count(), (varLegacyUserFilesProcessCompleteTime - varFilesProcessCompleteTime).TotalSeconds);
                varFilesProcessCompleteTime = varLegacyUserFilesProcessCompleteTime;
            }

            RefreshVARUserFilePrefs();
            DateTime vaFilePrefsProcessCompleteTime = DateTime.Now;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAR Prefs files processed in ", 0, (vaFilePrefsProcessCompleteTime - varFilesProcessCompleteTime).TotalSeconds);

            if (legacyRestore)
            {
                try
                {
                    if (!FileManagerSecure.DirectoryExists(Const.BAScriptDataFolder + "\\Archive")) FileManagerSecure.CreateDirectory(Const.BAScriptDataFolder + "\\Archive");
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.Init creating folder " + Const.BAScriptDataFolder + "\\Archive" + ": " + e); }

                string uniqueArchiveFolderName = Utils.GetUniqueFolderName(Const.BAScriptDataFolder + "\\Archive\\VARMetaData");
                try
                {
                    if (FileManagerSecure.DirectoryExists(Const.VARLegacyCoreDataFolder)) FileManagerSecure.MoveFile(Const.VARLegacyCoreDataFolder, uniqueArchiveFolderName);
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.Init Moving legacy VARMetaData folder to " + uniqueArchiveFolderName + ": " + e); }

                uniqueArchiveFolderName = Utils.GetUniqueFolderName(Const.BAScriptDataFolder + "\\Archive\\VARUserData");
                try
                {
                    if (FileManagerSecure.DirectoryExists(Const.VARLegacyUserDataFolder)) FileManagerSecure.MoveFile(Const.VARLegacyUserDataFolder, uniqueArchiveFolderName);
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.Init Moving legacy VARUserData folder to " + uniqueArchiveFolderName + ": " + e); }
            }
            

            return notNewResourceTypes;
        }

        public static string GetVPMEKey(string creatorName, string packageName,int varVersion)
        {
            return creatorName + "." + packageName + "."+varVersion.ToString();
        }
        public static string GetVPMEKey(VARPackageManifestEntry vpme)
        {
            return GetVPMEKey(vpme.creatorName, vpme.packageName, vpme.varVersion);

        }
        public static string GetVPVGEKey(string creatorName, string packgageName)
        {
            return creatorName + "." + packgageName;
        }
        public static string GetVPVGEKey(VARPackageManifestEntry vpme)
        {
            return GetVPVGEKey(vpme.creatorName, vpme.packageName);
        }

        public static string GetVARPath(PackagedRVGE packagedRVGE, int version)
        {
            string vpmeKey = GetVPMEKey(packagedRVGE.creatorName, packagedRVGE.packageName, version);
            if (varPackages.ContainsKey(vpmeKey))
            {
                return varPackages[vpmeKey].varFilePath;
            }
            return null;
        }
        public static void ReloadVARManifest(string[] varGroupCoreFileNames, string[] varResourceCoreFileNames, HashSet<string> varFileNames, DateTime scanTime)
        {
            HashSet<string> varGroupCoreFileNamesToSave = new HashSet<string>();
            foreach (string varGroupCoreFileName in varGroupCoreFileNames)
            {
                try
                {
                    JSONClass varGroupCoreFileJC = SuperController.singleton.LoadJSON(varGroupCoreFileName).AsObject;
                    if (varGroupCoreFileJC["VARManifestStoreFormat"].AsInt == Const.VARManifestStoreFormat)
                    {
                        JSONArray varGroupsJA = varGroupCoreFileJC["varGroups"].AsArray;
                        foreach (JSONClass varGroupJC in varGroupsJA)
                        {
                            JSONArray versionsJA = varGroupJC["varVersions"].AsArray;

                            string creatorName = varGroupJC["creatorName"];
                            string packageName = varGroupJC["packageName"];
                            string vpvgeKey = GetVPVGEKey(creatorName, packageName);

                            JSONArray currentVARVersionsJA = new JSONArray();
                            foreach (JSONClass jc in versionsJA)
                            {
                                int version = jc["varVersion"].AsInt;
                                string varFileNameFromMetaFile = vpvgeKey + "." + version.ToString() + ".var";
                                if (varFileNames.Contains(varFileNameFromMetaFile))  currentVARVersionsJA.Add(jc);
                                else varGroupCoreFileNamesToSave.Add(varGroupCoreFileName);
                            }

                            VARPackageVersionGroupEntry vpvge;
                            if (!varPackageVersionGroups.ContainsKey(vpvgeKey))
                            {
                                if (currentVARVersionsJA.Count > 0)
                                {
                                    vpvge = new VARPackageVersionGroupEntry(creatorName, packageName, currentVARVersionsJA, scanTime);
                                    varPackageVersionGroups.Add(vpvgeKey, vpvge);
                                    DataFileManifest.RegisterExisitingDFCoreVARGroup(vpvge, varGroupCoreFileName);
                                    CreatorManifest.RegisterCreator(vpvge.creatorName);
                                }
                                else varGroupCoreFileNamesToSave.Add(varGroupCoreFileName);


                            }
                            else SuperController.LogMessage("BA.VARPackageManifest.ReloadVARManifest: Unexpected duplicated key for Group Meta File " + varGroupCoreFileName);
                        }                  
                    }
                    else varGroupCoreFileNamesToSave.Add(varGroupCoreFileName);
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.ReloadVARManifest for Group Meta File " + varGroupCoreFileName + ": " + e); }
            }
            foreach (string varGroupCoreFileName in varGroupCoreFileNamesToSave) DataFileManifest.RegisterVARGroupCoreSaveRequest(varGroupCoreFileName);

            HashSet<string> varResourceCoreFileNamesToSave = new HashSet<string>();
            foreach (string varResourceCoreFileName in varResourceCoreFileNames)
            {
                try
                {
                    JSONClass varResourcesCoreFileJC = SuperController.singleton.LoadJSON(varResourceCoreFileName).AsObject;
                    if (varResourcesCoreFileJC["VARManifestStoreFormat"].AsInt == Const.VARManifestStoreFormat)
                    {

                        JSONArray varResourcesJA = varResourcesCoreFileJC["resources"].AsArray;
                        foreach (JSONClass varResourceJC in varResourcesJA)
                        {
                            string creatorName = varResourceJC["creatorName"];
                            string packageName = varResourceJC["packageName"];
                            string varGroupName = GetVPVGEKey(creatorName, packageName);
                            if (varPackageVersionGroups.ContainsKey(varGroupName))
                            {
                                bool result = ResourceManifest.RegisterVARResourceFromManifest(varPackageVersionGroups[varGroupName], varResourceJC, varResourceCoreFileName);
                                if (!result) varResourceCoreFileNamesToSave.Add(varResourceCoreFileName);
                            }
                            else varResourceCoreFileNamesToSave.Add(varResourceCoreFileName);
                        }
                    }
                    else varResourceCoreFileNamesToSave.Add(varResourceCoreFileName);

                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.ReloadVARManifest for Resource Meta File " + varResourceCoreFileName + ": " + e); }
            }
            foreach (string varResourceCoreFileName in varResourceCoreFileNamesToSave) DataFileManifest.RegisterVARResourceCoreSaveRequest(varResourceCoreFileName);
        }

        public static void RefreshManifestWithNewVARs(HashSet<string> varFiles, DateTime scanTime, HashSet<int> notNewResourceTypes)
        {            
            foreach (string varFile in varFiles)
            {
                try
                {
                    if (FileManagerSecure.DirectoryExists(varFile + ":"))
                    {
                        VARPackageManifestEntry vpme = new VARPackageManifestEntry(varFile);
                        string vpmeKey = GetVPMEKey(vpme.creatorName, vpme.packageName, vpme.varVersion);

                        if (!varPackages.ContainsKey(vpmeKey))
                        {
                            string vpvgeKey = GetVPVGEKey(vpme.creatorName, vpme.packageName);

                            if (!varPackageVersionGroups.ContainsKey(vpvgeKey) || !varPackageVersionGroups[vpvgeKey].varVersionLicenses.Keys.Contains(vpme.varVersion))
                            {
                                ProcessNewVARFile(vpme, scanTime, notNewResourceTypes);
                            }
                            varPackages.Add(vpmeKey, vpme);
                        }
                    }                    
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.RefreshManifest for " + varFile + ": " + e); }               
            }
        }
        public static void PurgeVARFile (string creatorName, string packageName, int varVersion)
        {
            string vpmeKey = GetVPMEKey(creatorName, packageName,varVersion);
            if (varPackages.ContainsKey(vpmeKey))
            {
                string vpvgeKey = GetVPVGEKey(creatorName, packageName);
                if (varPackageVersionGroups.ContainsKey(vpvgeKey))
                {
                    var vpvge = varPackageVersionGroups[vpvgeKey];
                    vpvge.PurgeVARVersion(varVersion);

                    if (vpvge.versionsCount == 0)
                    {
                        varPackageVersionGroups.Remove(vpvgeKey);
                        vpvge.coreDataFile.UnRegisterVARGroup(vpvge);
                        vpvge.userDataFile.UnRegisterVARGroup(vpvge);
                    }
                }
                varPackages.Remove(vpmeKey);
            }
        }
        public static void ReloadLegacyVARUserData(string[] files)
        {
            foreach (string file in files)
            {
                try
                {
                    JSONClass jc = SuperController.singleton.LoadJSON(file).AsObject;

                    string creatorName = jc["creatorName"];
                    string packageName = jc["packageName"];

                    string vpvgeKey = GetVPVGEKey(creatorName, packageName);

                    if (varPackageVersionGroups.ContainsKey(vpvgeKey))
                    {
                        varPackageVersionGroups[vpvgeKey].ReloadLegacyVARUserData(jc);
                    }
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.ReloadVARUserData for " + file + ": " + e); }

            }
        }

        public static void ReloadGroupVARUserData(string[] varGroupUserFiles)
        {
            foreach (string varGroupUserFileName in varGroupUserFiles)
            {
                try
                {
                    JSONClass jc = SuperController.singleton.LoadJSON(varGroupUserFileName).AsObject;

                    JSONArray varGroupsJA = jc["varGroups"].AsArray;

                    HashSet<string> varGroupUserFileNamesToSave = new HashSet<string>();

                    foreach (JSONClass varGroupJC in varGroupsJA)
                    {
                        string creatorName = varGroupJC["creatorName"];
                        string packageName = varGroupJC["packageName"];

                        string vpvgeKey = GetVPVGEKey(creatorName, packageName);

                        if (varPackageVersionGroups.ContainsKey(vpvgeKey))
                        {
                            VARPackageVersionGroupEntry vpvge = varPackageVersionGroups[vpvgeKey];
                            DataFileManifest.RegisterExisitingDFUserVARGroup(vpvge, varGroupUserFileName);

                            JSONArray varVersionJA = varGroupJC["varVersions"].AsArray;

                            JSONArray currentVARVersionsJA = new JSONArray();
                            foreach (JSONClass varVersionJC in varVersionJA)
                            {
                                int version = varVersionJC["varVersion"].AsInt;
                                if (vpvge.varVersionLicenses.Keys.Contains(version)) currentVARVersionsJA.Add(varVersionJC);
                                else varGroupUserFileNamesToSave.Add(varGroupUserFileName);
                            }


                            vpvge.ReloadVARVersionsUserData(currentVARVersionsJA);
                        }
                        else varGroupUserFileNamesToSave.Add(varGroupUserFileName);
                    }

                    foreach (string varGroupUserFileNameToSave in varGroupUserFileNamesToSave) DataFileManifest.RegisterVARGroupUserSaveRequest(varGroupUserFileNameToSave);

                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.ReloadGroupVARUserData for " + varGroupUserFileName + ": " + e); }
            }
        }


        public static void RefreshVARUserFilePrefs()
        {
            ResourceManifest.ClearAllUserFilePrefs();
            foreach(var vpme in varPackages)
            {
                if (FileManagerSecure.DirectoryExists("AddonPackagesFilePrefs\\" + vpme.Value.varFileNameNoExtension))
                {
                    string searchFolder = "AddonPackagesFilePrefs\\" + vpme.Value.varFileNameNoExtension;
                    HashSet<string> varFilePrefs = Utils.GetFilesAtPathRecursive(searchFolder, "*.*");
                    foreach (string varFilePref in varFilePrefs)
                    {
                        try
                        {
                            int fileExtIndex = varFilePref.LastIndexOf('.');
                            int filePathLength = varFilePref.Length;

                            string rvgeKey = ResourceManifest.GetRVGKey(vpme.Value, varFilePref.Substring(searchFolder.Length + 1, fileExtIndex - searchFolder.Length - 1));
                            if (varFilePref.EndsWith(".hide")) ResourceManifest.UpdateVARUserHideFlag(rvgeKey, vpme.Value.varVersion, true);
                            if (varFilePref.EndsWith(".fav")) ResourceManifest.UpdateVARUserFavFlag(rvgeKey, vpme.Value.varVersion, true);
                        }
                        catch (Exception e) { SuperController.LogError("Exception in BA.RefreshVARUserFilePrefs for VAR " + vpme.Value.varFileNameNoExtension + "and file "+ varFilePref+": " + e); }

                    }
                }

            }
        }

        private static void ProcessNewVARFile(VARPackageManifestEntry vpme, DateTime scanTime, HashSet<int> notNewResourceTypes)
        {
            string vpvgeKey = GetVPVGEKey(vpme);
            VARPackageVersionGroupEntry vpvge;
            if (!varPackageVersionGroups.ContainsKey(vpvgeKey))
            {
                vpvge = new VARPackageVersionGroupEntry(vpme,scanTime);
                varPackageVersionGroups.Add(vpvgeKey, vpvge);
                CreatorManifest.RegisterCreator(vpvge.creatorName);
                DataFileManifest.RegisterNewDFCoreVARGroup(vpvge);
            }
            else
            {
                vpvge = varPackageVersionGroups[vpvgeKey];
                if (vpvge.varVersionLicenses.Keys.Contains(vpme.varVersion)) return;
                vpvge.AddVARVersion(vpme.varVersion,vpme.varFilePath, scanTime);
                
            }

            vpvge.SetPersistVARCoreData();
            vpvge.SetPersistVARGroupUserData();
            vpme.vpvge = vpvge;
            ResourceManifest.RegisterNewVARResources(vpme, notNewResourceTypes);

        }
    }

    public class VARPackageManifestEntry
    {
        public string varFilePath { get; private set; }
        public string varFileNameNoExtension { get; private set; }
        public string creatorName { get; private set; }
        public string packageName { get; private set; }
        public int varVersion { get; private set; }

        public VARPackageVersionGroupEntry vpvge;

        public VARPackageManifestEntry(string varFile)
        {
            varFilePath = varFile;
            string varFileName = FileManagerSecure.GetFileName(varFilePath);
            varFileNameNoExtension = varFileName.Substring(0, varFileName.Length - 4);

            string[] parts = varFileNameNoExtension.Split('.');
            creatorName = parts[0];
            packageName = parts[1];
            varVersion = int.Parse(parts[2]);    
        }
    }
    public class VARPackageVersionGroupEntry
    {
        public string creatorName { get; private set; }
        public string packageName { get; private set; }

        public string varGroupName { get { return creatorName + "." + packageName; } }

        private HashSet<int> varVersions = new HashSet<int>();
        public Dictionary<int,int> varVersionLicenses = new Dictionary<int, int>();

        private Dictionary<int, DateTime> varVersionFirstScanTimes = new Dictionary<int, DateTime>();
        private Dictionary<int, DateTime> varVersionFirstCacheTimes = new Dictionary<int, DateTime>();

        private VARGroupUserDataFile _userDataFile = null;

        public VARGroupCoreDataFile coreDataFile { get; private set;}
        public VARGroupUserDataFile userDataFile {
            get
            {
                if (_userDataFile == null) DataFileManifest.RegisterNewDFUserVARGroup(this);
                return _userDataFile;
            }
            private set { _userDataFile = value; } }

        public DateTime firstCataloguedDateTime
        {
            get
            {
                DateTime firstScanDateTime = varVersionFirstScanTimes.Values.Min(d => d);
                if (varVersionFirstCacheTimes.Count == 0) return firstScanDateTime;
                DateTime firstCacheDateTime = varVersionFirstCacheTimes.Values.Min(d => d);
                return new DateTime(Math.Min(firstScanDateTime.Ticks, firstCacheDateTime.Ticks));
            }
        }
        public DateTime lastCataloguedDateTime
        {
            get
            {
                DateTime lastScanDateTime = varVersionFirstScanTimes[highestVARVersion];
                if (!varVersionFirstCacheTimes.ContainsKey(highestVARVersion)) return lastScanDateTime;
                DateTime lastCacheDateTime = varVersionFirstCacheTimes[highestVARVersion];
                return new DateTime(Math.Min(lastScanDateTime.Ticks, lastCacheDateTime.Ticks));
            }
        }


        public int highestVARVersion { get; private set; }
        public int versionsCount { get { return varVersions.Count; } }

        private HashSet<PackagedRVGE> packageRVGEs = new HashSet<PackagedRVGE>();
        private Dictionary<string,PackagedRVGE> packageRVGEDict = new Dictionary<string, PackagedRVGE>();

        public static bool anyPersistVARMetaDataPending { get; private set; } = false;
        public static bool anyPersistVARUserDataPending { get; private set; } = false;

        public VARPackageVersionGroupEntry(VARPackageManifestEntry vpme, DateTime scanTime)
        {
            creatorName = vpme.creatorName;
            packageName = vpme.packageName;

            varVersions.Add(vpme.varVersion);
            varVersionLicenses.Add(vpme.varVersion, GetLicenseTypeFromMeta(vpme.varFilePath));
            varVersionFirstScanTimes.Add(vpme.varVersion, scanTime);
            highestVARVersion = vpme.varVersion;

            CheckFirstCachedDate(vpme.varVersion);
        }

        public VARPackageVersionGroupEntry(string _creatorName, string _packageName, JSONArray varVersionsJA, DateTime scanTime)
        {
            creatorName = _creatorName;
            packageName = _packageName;
            highestVARVersion = 0;
            foreach (JSONClass jc in varVersionsJA)
            {
                int varVersion = jc["varVersion"].AsInt;
                varVersions.Add(varVersion);
                varVersionLicenses.Add(varVersion, EnumManifest.GetEnumValFromStoreVal(LicenseType.enumManifestName,jc["varLicense"]));
                varVersionFirstScanTimes.Add(varVersion, scanTime);

                if (highestVARVersion < varVersion) highestVARVersion = varVersion;
            }
            
        }

        public bool AnyExistingVARVersions(JSONArray varResourcesVersionsJA)
        {
            foreach (JSONNode versionJN in varResourcesVersionsJA) if (varVersions.Contains(versionJN.AsInt)) return true;
            return false;
        }

        public bool AnyMismatchedVARVersions(JSONArray varResourcesVersionsJA)
        {
            foreach (JSONNode versionJN in varResourcesVersionsJA) if (!varVersions.Contains(versionJN.AsInt)) return true;
            return false;
        }

        public void RegisterCoreDataFile(VARGroupCoreDataFile coreDataFile)
        {
            this.coreDataFile = coreDataFile;
        }

        public void RegisterUserDataFile(VARGroupUserDataFile userDataFile)
        {
            this.userDataFile = userDataFile;
        }

        private void CheckFirstCachedDate(int varVersion)
        {
            string cacheMetaFileName = "Cache\\PackageJSON\\" + varGroupName + "." + varVersion.ToString() + ".vamcachemeta";
            DateTime? firstCachedDateTime = null;

            if (FileManagerSecure.FileExists(cacheMetaFileName))
            {
                JSONClass cacheMetaFileJC = SuperController.singleton.LoadJSON(cacheMetaFileName).AsObject;
                if (cacheMetaFileJC["date"] != null) firstCachedDateTime = DateTime.Parse(cacheMetaFileJC["date"], new CultureInfo("en-US", false));
                SetPersistVARGroupUserData();
            }
            if (firstCachedDateTime != null)
            {
                if (varVersionFirstCacheTimes.ContainsKey(varVersion)) varVersionFirstCacheTimes[varVersion] = (DateTime)firstCachedDateTime;
                else varVersionFirstCacheTimes.Add(varVersion, (DateTime)firstCachedDateTime);
            }
        }

        private int GetLicenseTypeFromMeta(string varFileName)
        {
            int varLicenseType = LicenseType.Unknown;
            if (FileManagerSecure.FileExists(varFileName + ":\\meta.json"))
            {
                JSONClass varMetaJSON = null;
                try
                {
                    varMetaJSON = SuperController.singleton.LoadJSON(varFileName + ":\\meta.json").AsObject;
                }
                catch (Exception e) { SuperController.LogError("VARPackageVersionGroupEntry.GetLicenseTypeFromMeta: Error loading meta.json in '" + varFileName+"'. Possibly malformed JSON format. Exception message - " + e); }

                if (varMetaJSON != null && varMetaJSON["licenseType"] != null)
                {
                    string varLicenseTypeStoreVal = varMetaJSON["licenseType"];
                    if (EnumManifest.GetStoreValEnumsDict(LicenseType.enumManifestName).Keys.Contains(varLicenseTypeStoreVal))
                    {
                        varLicenseType = EnumManifest.GetEnumValFromStoreVal(LicenseType.enumManifestName, varLicenseTypeStoreVal);
                    }
                    else
                    {
                        SuperController.LogMessage(varFileName + " contains unrecognised License Type. Possibly caused by creator manually editing VAR meta.json");
                    }
                    
                }
            }
            else SuperController.LogMessage("Meta Not found: "+ varFileName);
            return varLicenseType;
        }
        public void SetPersistVARCoreData()
        {
            DataFileManifest.RegisterVARGroupCoreSaveRequest(coreDataFile);
        }

        public void SetPersistVARGroupUserData()
        {
            DataFileManifest.RegisterVARGroupUserSaveRequest(userDataFile);
        }

        public void AddVARVersion(int varVersion, string varFilePath, DateTime scanTime)
        {
            varVersions.Add(varVersion);
            varVersionLicenses.Add(varVersion, GetLicenseTypeFromMeta(varFilePath));
            if (varVersion> highestVARVersion) highestVARVersion = varVersion;
            varVersionFirstScanTimes.Add(varVersion, scanTime);
            CheckFirstCachedDate(varVersion);
        }

        public void PurgeVARVersion(int varVersion)
        {
            HashSet<PackagedRVGE> tempPackageRVGEs = new HashSet<PackagedRVGE>(packageRVGEs);
            foreach (PackagedRVGE rvge in tempPackageRVGEs)
            {
                rvge.PurgeVARVersion(varVersion);
                if (rvge.versionsCount == 0)
                {
                    ResourceManifest.PurgePackageRVGE(rvge);
                    string rvgeKey = ResourceManifest.GetRVGKey(rvge);
                    packageRVGEDict.Remove(rvgeKey);
                    packageRVGEs.Remove(rvge);
                    rvge.SetUserDataPersist();
                    rvge.coreDataFile.UnRegisterVARResource(rvge);
                    rvge.userDataFile.UnRegisterVARResource(rvge);
                }
                rvge.SetCoreDataPersist();
            }
            varVersions.Remove(varVersion);
            varVersionFirstCacheTimes.Remove(varVersion);
            varVersionFirstScanTimes.Remove(varVersion);
            varVersionLicenses.Remove(varVersion);  
            if (highestVARVersion == varVersion)
            {
                highestVARVersion = -1;
                foreach(int remainingVersion in varVersions)
                {
                    if (remainingVersion > highestVARVersion) highestVARVersion = remainingVersion;
                }
            }

            SetPersistVARCoreData();
            SetPersistVARGroupUserData();
            
            
        }

        public void RegisterRVGE(PackagedRVGE packagedRVGE)
        {
            packageRVGEs.Add(packagedRVGE);
            packageRVGEDict.Add(ResourceManifest.GetRVGKey(this, packagedRVGE.resourceFolderPathFileName),packagedRVGE);
        }

        private void ReloadVARVersionUserData(JSONClass varVersionJC, int varVersion)
        {
            DateTime firstScanTime = DateTime.Parse(varVersionJC["firstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
            if (varVersionFirstScanTimes.ContainsKey(varVersion)) varVersionFirstScanTimes[varVersion] = firstScanTime;
            else varVersionFirstScanTimes.Add(varVersion, firstScanTime);

            string cacheMetaFileName = "Cache\\PackageJSON\\" + varGroupName + "." + varVersion.ToString() + ".vamcachemeta";
            if (varVersionJC["firstCacheDateTime"] != null)
            {
                DateTime firstCachedDateTime = DateTime.Parse(varVersionJC["firstCacheDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
                if (varVersionFirstCacheTimes.ContainsKey(varVersion)) varVersionFirstCacheTimes[varVersion] = firstCachedDateTime;
                else varVersionFirstCacheTimes.Add(varVersion, firstCachedDateTime);

            }
            else CheckFirstCachedDate(varVersion);
        }
        public void ReloadVARVersionsUserData(JSONArray varVersionsJA)
        {
            HashSet<int> previouslyScannedVARVersions = new HashSet<int>();

            foreach (JSONClass varVersionJC in varVersionsJA)
            {
                int varVersion = varVersionJC["varVersion"].AsInt;
                previouslyScannedVARVersions.Add(varVersion);
                ReloadVARVersionUserData(varVersionJC, varVersion);
            }

            HashSet<int> newVARVersions = new HashSet<int>(varVersions.Except(previouslyScannedVARVersions));
            if (newVARVersions.Count > 0)
            {
                foreach (ResourceVersionGroupEntry rvge in packageRVGEs)
                {
                    if (!rvge.localFileResource) (rvge as PackagedRVGE).NewVARVersions(newVARVersions);
                }

            }
        }
        public void ReloadLegacyVARUserData(JSONClass varUserDataJC)
        {
            int varUserDataStoreFormat = varUserDataJC["VARUserDataStoreFormat"].AsInt;
            JSONArray varVersionsJA = varUserDataJC["varVersions"].AsArray;
            JSONArray resourcesJA = varUserDataJC["resources"].AsArray;

            HashSet<int> previouslyScannedVARVersions = new HashSet<int>();

            if (varUserDataStoreFormat == 1)
            {
                foreach (JSONNode varVersionJC in varVersionsJA)
                {
                    int varVersion = varVersionJC.AsInt;
                    previouslyScannedVARVersions.Add(varVersion);
                }
                foreach (JSONClass resourceJC in resourcesJA)
                {
                    JSONArray resourceVARVersionsJA = resourceJC["varVersions"].AsArray;
                    foreach (JSONClass resourceVARVersionJC in resourceVARVersionsJA)
                    {
                        int varVersion = resourceVARVersionJC["varVersion"].AsInt;
                        DateTime firstScanTime = DateTime.Parse(resourceVARVersionJC["firstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);

                        if (varVersionFirstScanTimes.ContainsKey(varVersion)) varVersionFirstScanTimes[varVersion] = firstScanTime;
                        else varVersionFirstScanTimes.Add(varVersion, firstScanTime);
                    }
                }
                SetPersistVARGroupUserData();
            }
            else
            {
                foreach (JSONClass varVersionJC in varVersionsJA)
                {
                    int varVersion = varVersionJC["varVersion"].AsInt;
                    previouslyScannedVARVersions.Add(varVersion);

                    ReloadVARVersionUserData(varVersionJC, varVersion);
                }              
            }

            HashSet<int> newVARVersions = new HashSet<int>(varVersions.Except(previouslyScannedVARVersions));
            if (newVARVersions.Count > 0)
            {
                foreach (ResourceVersionGroupEntry rvge in packageRVGEs)
                {
                    if (!rvge.localFileResource) (rvge as PackagedRVGE).NewVARVersions(newVARVersions);
                }

            }

            foreach (JSONClass resourceJC in resourcesJA)
            {
                string rvgeKey = ResourceManifest.GetRVGKey(creatorName, packageName, resourceJC["resourceFullFileName"]);
                if (packageRVGEDict.ContainsKey(rvgeKey))
                {
                    ResourceVersionGroupEntry rvge = packageRVGEDict[rvgeKey];
                    if (!rvge.localFileResource)
                    {
                        (rvge as PackagedRVGE).RestoreVARUserData(resourceJC);
                        (rvge as PackagedRVGE).SetUserDataPersist();
                    }
                }
            }

        }
        public JSONClass GetVARGroupUserDataJSON()
        {
            JSONClass varJC = new JSONClass();
            varJC["creatorName"] = creatorName;
            varJC["packageName"] = packageName;

            JSONArray varVersionsJA = new JSONArray();
            varJC["varVersions"] = varVersionsJA;
            foreach (var varVersionFirstScanTime in varVersionFirstScanTimes)
            {
                JSONClass varVersionJC = new JSONClass();
                varVersionJC["varVersion"].AsInt = varVersionFirstScanTime.Key;
                varVersionJC["firstScanDateTime"] = varVersionFirstScanTime.Value.ToString("O");
                if (varVersionFirstCacheTimes.ContainsKey(varVersionFirstScanTime.Key))
                {
                    varVersionJC["firstCacheDateTime"] = varVersionFirstCacheTimes[varVersionFirstScanTime.Key].ToString("O");
                }

                varVersionsJA[-1] = varVersionJC;
            }

            return varJC;
        }

        public JSONClass GetVARGroupMetaDataJSON()
        {
            JSONClass jc = new JSONClass();


            jc["creatorName"] = creatorName;
            jc["packageName"] = packageName;
//            jc["varGroupName"] = varGroupName;
            JSONArray varVersionsJA = new JSONArray();
            jc["varVersions"] = varVersionsJA;
            foreach (var varVersion in varVersionLicenses)
            {
                JSONClass varVersionJC = new JSONClass();
                varVersionJC["varVersion"].AsInt = varVersion.Key;
                varVersionJC["varLicense"] = EnumManifest.GetStoreVal(LicenseType.enumManifestName, varVersion.Value);
                varVersionsJA[-1] = varVersionJC;
            }
            return jc;
        }
    }

    public class ResourceManifest
    {
        private static Dictionary<string, ResourceVersionGroupEntry> resourceVersionGroupsDict = new Dictionary<string, ResourceVersionGroupEntry>();

        private static HashSet<ResourceVersionGroupEntry> resourceVersionGroups = new HashSet<ResourceVersionGroupEntry>();

        private static HashSet<LocalRVGE> localResourceVersionGroups = new HashSet<LocalRVGE>();
        private static HashSet<PackagedRVGE> varResourceVersionGroups = new HashSet<PackagedRVGE>();

        public static void PurgeLocalRVGE(LocalRVGE localRVGE)
        {
            if (localRVGE.localFileResource)
            {
                localResourceVersionGroups.Remove(localRVGE);
                resourceVersionGroups.Remove(localRVGE);
                resourceVersionGroupsDict.Remove(localRVGE.resourceFullPathReference);
                localRVGE.userDataFile.PurgeLocalRVGE(localRVGE);
            }
        }

        public static void OnDestoy()
        {
            foreach (ResourceVersionGroupEntry rvge in resourceVersionGroups)
            {
                if (rvge.resourceSprite != null)
                {
                    GameObject.Destroy(rvge.resourceSprite.texture);
                    GameObject.Destroy(rvge.resourceSprite);
                }
                
            }
            resourceVersionGroupsDict.Clear();
            resourceVersionGroupsDict = null;
            resourceVersionGroups.Clear();
            resourceVersionGroups = null;
            localResourceVersionGroups.Clear();
            localResourceVersionGroups = null;
            varResourceVersionGroups.Clear();
            varResourceVersionGroups = null;
        }

        public static void PurgePackageRVGE(PackagedRVGE packagedRVGE)
        {
            if (!packagedRVGE.localFileResource)
            {
                varResourceVersionGroups.Remove(packagedRVGE);
                resourceVersionGroups.Remove(packagedRVGE);
                resourceVersionGroupsDict.Remove(GetRVGKey(packagedRVGE));
                CreatorManifest.RemoveRVGE(packagedRVGE);
            }
        }

        public static void ResetRVGEFolderPath(LocalRVGE rvge)
        {
            string oldRVGEKey = rvge.resourceFolderPathFileName;
            rvge.ResetFolderPath();
            string newRVGEKey = rvge.resourceFolderPathFileName;
            resourceVersionGroupsDict.Remove(oldRVGEKey);
            resourceVersionGroupsDict.Add(newRVGEKey,rvge);
        }

        public static void RenameRVGEFilename(LocalRVGE rvge, string newFullPathFileName)
        {
            string oldRVGEKey = rvge.resourceFolderPathFileName;
            rvge.RenameRVGEFilename(newFullPathFileName);
            string newRVGEKey = rvge.resourceFolderPathFileName;
            resourceVersionGroupsDict.Remove(oldRVGEKey);
            resourceVersionGroupsDict.Add(newRVGEKey, rvge);
        }

        public static void MoveRVGELocalFolderPath(LocalRVGE rvge, Folder targetFolder)
        {
            string oldRVGEKey = rvge.resourceFolderPathFileName;
            rvge.MoveFolderPath(targetFolder);
            string newRVGEKey = rvge.resourceFolderPathFileName;
            resourceVersionGroupsDict.Remove(oldRVGEKey);
            resourceVersionGroupsDict.Add(newRVGEKey, rvge);
        }
        public static void DeleteLocalRVGE(LocalRVGE rvge)
        {
            string rvgeKey = rvge.resourceFolderPathFileName;
            resourceVersionGroupsDict.Remove(rvgeKey);
            localResourceVersionGroups.Remove(rvge);
            resourceVersionGroups.Remove(rvge);
        }

        public static void LocalResourcesInit(DateTime scanTime, HashSet<int> notNewResourceTypes)
        {
            DataFileManifest.ClearLocalResourceUserDataFile();
            foreach (LocalRVGE existingLocalRVGE in localResourceVersionGroups)
            {
                resourceVersionGroups.Remove(existingLocalRVGE);
                resourceVersionGroupsDict.Remove(existingLocalRVGE.resourceFolderPathFileName);
            }
            localResourceVersionGroups.Clear();

            bool legacyImport = false;
            if (FileManagerSecure.DirectoryExists(Const.LocalLegacyUserDataFolder)) legacyImport = true;
            if (!legacyImport)
            {
                DateTime startTimeReloadLocalDataFiles = DateTime.Now;
                int localFileCount = ReloadLocalDataFiles();

                DateTime endTimeReloadLocalDataFiles = DateTime.Now;
                double findTimeReload = (endTimeReloadLocalDataFiles - startTimeReloadLocalDataFiles).TotalSeconds;

                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Local User Data files reloaded in ", localFileCount, findTimeReload);
            }

            RefreshLocalSceneManifest(scanTime, notNewResourceTypes.Contains(ResourceType.scene));
            RefreshLocalPresetManifest( ResourceType.presetAppearance, scanTime, notNewResourceTypes.Contains(ResourceType.presetAppearance));
            RefreshLocalPresetManifest(ResourceType.presetAnimation, scanTime, notNewResourceTypes.Contains(ResourceType.presetAnimation));
            RefreshLocalPresetManifest( ResourceType.presetBreastPhys, scanTime, notNewResourceTypes.Contains(ResourceType.presetBreastPhys));
            RefreshLocalPresetManifest(ResourceType.presetClothing, scanTime, notNewResourceTypes.Contains(ResourceType.presetClothing));
            RefreshLocalPresetManifest(ResourceType.presetGlutePhysics, scanTime, notNewResourceTypes.Contains(ResourceType.presetGlutePhysics));
            RefreshLocalPresetManifest( ResourceType.presetHair, scanTime, notNewResourceTypes.Contains(ResourceType.presetHair));
            RefreshLocalPresetManifest( ResourceType.presetMorphs, scanTime, notNewResourceTypes.Contains(ResourceType.presetMorphs));
            RefreshLocalPresetManifest( ResourceType.presetPlugins, scanTime, notNewResourceTypes.Contains(ResourceType.presetPlugins));
            RefreshLocalPresetManifest( ResourceType.presetPose, scanTime, notNewResourceTypes.Contains(ResourceType.presetPose));
            RefreshLocalPresetManifest( ResourceType.presetSkin, scanTime, notNewResourceTypes.Contains(ResourceType.presetSkin));
            RefreshLocalPresetManifest( ResourceType.presetGeneral, scanTime, notNewResourceTypes.Contains(ResourceType.presetGeneral));

            DateTime startTime = DateTime.Now;
            string[] patterns = new string[3] {"*.cs","*.cslist","*.dll" };
            HashSet<string> localPluginFiles = Utils.GetFilesAtPathRecursive("Custom\\Scripts", patterns);

            DateTime scriptFilesScanCompleteTime = DateTime.Now;
            double findTime = (scriptFilesScanCompleteTime - startTime).TotalSeconds;

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Script files found in ", localPluginFiles.Count, findTime);

            RefreshLocalResourceManifest(localPluginFiles, ResourceType.plugins, scanTime, notNewResourceTypes.Contains(ResourceType.plugins));

            DateTime scriptFilesProcessCompleteTime = DateTime.Now;
            double processTime = (scriptFilesProcessCompleteTime - scriptFilesScanCompleteTime).TotalSeconds;

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Script files processed in ", localPluginFiles.Count, processTime);


            if (legacyImport)
            {
                try
                {
                    if (!FileManagerSecure.DirectoryExists(Const.BAScriptDataFolder + "\\Archive")) FileManagerSecure.CreateDirectory(Const.BAScriptDataFolder + "\\Archive");
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.ResourceManifest.LocalResourcesInit creating folder " + Const.BAScriptDataFolder + "\\Archive" + ": " + e); }

                string uniqueArchiveFolderName = Utils.GetUniqueFolderName(Const.BAScriptDataFolder + "\\Archive\\LocalUserData");
                try
                {
                    if (FileManagerSecure.DirectoryExists(Const.LocalLegacyUserDataFolder)) FileManagerSecure.MoveFile(Const.LocalLegacyUserDataFolder, uniqueArchiveFolderName);
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.ResourceManifest.LocalResourcesInit Moving legacy Local folder to " + uniqueArchiveFolderName +": " + e); }

            }

        }

        private static int ReloadLocalDataFiles()
        {
            string[] localDataUserFiles = FileManagerSecure.GetFiles(Const.LocalResourcesUserDataFolder, "*" + Const.UserDataFileExt);
            HashSet<string> localDataUserFilesToSave = new HashSet<string>();
            foreach (string localDataUserFile in localDataUserFiles)
            {
                try
                {
                    JSONClass localDataUserFileJC = SuperController.singleton.LoadJSON(localDataUserFile).AsObject;
                    JSONArray localResourcesJA = localDataUserFileJC["resources"].AsArray;

                    foreach (JSONClass localResourceJC in localResourcesJA)
                    {
                        string resourceFilePath = localResourceJC["resourceFullFileName"];
                        if (FileManagerSecure.FileExists(resourceFilePath))
                        {
                            int resourceType = EnumManifest.GetEnumValFromStoreVal(ResourceType.enumManifestName, localResourceJC["resourceType"]);
                            LocalRVGE rvge = new LocalRVGE(resourceFilePath, resourceType);

                            resourceVersionGroupsDict.Add(resourceFilePath, rvge);
                            resourceVersionGroups.Add(rvge);
                            localResourceVersionGroups.Add(rvge);
                            rvge.LoadLocalUserData(localResourceJC);
                            DataFileManifest.RegisterExisitingDFUserLocalResource(rvge, localDataUserFile);
                        }
                        else localDataUserFilesToSave.Add(localDataUserFile);
                    }
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.ResourceManifest.ReloadLocalDataFiles for File " + localDataUserFile + ": " + e); }
            }

            foreach (string localDataUserFile in localDataUserFilesToSave) DataFileManifest.RegisterLocalResourceUserSaveRequest(localDataUserFile);

            return localDataUserFiles.Count();
        }
        private static void RefreshLocalSceneManifest(DateTime scanTime, bool forceNotNew)
        {
            DateTime startTime = DateTime.Now;

            HashSet<string> localFiles = Utils.GetFilesAtPathRecursive("Saves\\scene", "*.json");

            DateTime sceneFilesScanCompleteTime = DateTime.Now;
            double findTime = (sceneFilesScanCompleteTime - startTime).TotalSeconds;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Scene files found in ", localFiles.Count, findTime);

            RefreshLocalResourceManifest(localFiles, ResourceType.scene, scanTime, forceNotNew);

            DateTime sceneFilesProcessCompleteTime = DateTime.Now;
            double processTime = (sceneFilesProcessCompleteTime - sceneFilesScanCompleteTime).TotalSeconds;

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Scene files processed in ", localFiles.Count, processTime);

        }
        private static void RefreshLocalPresetManifest(int resourceType, DateTime scanTime, bool forceNotNew)
        {
            DateTime startTime = DateTime.Now;

            HashSet<string> localFiles = new HashSet<string>();
            foreach (string resourcePath in Const.GetResourceRootFolderNames(resourceType))
            {
                localFiles.UnionWith(Utils.GetFilesAtPathRecursive(resourcePath, "*.vap"));
            }

            DateTime presetFilesScanCompleteTime = DateTime.Now;
            double findTime = (presetFilesScanCompleteTime - startTime).TotalSeconds;
            string presetName = EnumManifest.GetDisplayVal(ResourceType.enumManifestName, resourceType);

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" " + presetName + " files found in ", localFiles.Count, findTime);

            RefreshLocalResourceManifest(localFiles, resourceType, scanTime, forceNotNew);

            DateTime presetFilesProcessCompleteTime = DateTime.Now;
            double processTime = (presetFilesProcessCompleteTime - presetFilesScanCompleteTime).TotalSeconds;

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" " + presetName + " files processed in ", localFiles.Count, processTime);
        }

        private static void RefreshLocalResourceManifest(HashSet<string> localFiles, int resourceType, DateTime scanTime, bool forceNotNew)
        {
            bool legacyImport = false;
            if (FileManagerSecure.DirectoryExists(Const.LocalLegacyUserDataFolder)) legacyImport = true;
            foreach (string localFile in localFiles)
            {
                try
                {

                    RegisterLocalResourceOfType(localFile, resourceType, scanTime, forceNotNew, legacyImport);
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.RefreshLocalResourceManifest for " + localFile + ": " + e); }

            }
        }

        private static void RegisterLocalResourceOfType(string resourceFileName, int resourceType, DateTime scanTime, bool forceNotNew, bool legacyImport)
        {

            if (!resourceVersionGroupsDict.ContainsKey(resourceFileName))
            {
                LocalRVGE rvge = new LocalRVGE(resourceFileName, resourceType);

                resourceVersionGroupsDict.Add(resourceFileName, rvge);
                resourceVersionGroups.Add(rvge);
                localResourceVersionGroups.Add(rvge);

                string localUserDataManifestFilename = "";

                if (legacyImport) localUserDataManifestFilename = Const.LocalLegacyUserDataFolder + "\\" + rvge.resourceFolderPathFileName + Const.CoreDataFileExt;

                if (legacyImport && FileManagerSecure.FileExists(localUserDataManifestFilename))
                {
                    rvge.LoadLegacyLocalUserData(localUserDataManifestFilename);
                    rvge.SetUserDataPersist();
                }
                else
                {
                    rvge.SetLocalFileScanTime(scanTime);
                    if (!forceNotNew) rvge.isNew = true;
                }

            }
        }

        public static void ReloadResourceVARUserData(string[] varResourceUserFiles)
        {
            HashSet<string> varResourceCoreFileNamesToSave = new HashSet<string>();
            foreach (string varResourceUserFileName in varResourceUserFiles)
            {
                try
                {
                    JSONClass jc = SuperController.singleton.LoadJSON(varResourceUserFileName).AsObject;

                    JSONArray resourceJA = jc["resources"].AsArray;

                    foreach (JSONClass resourceJC in resourceJA)
                    {
                        string creatorName = resourceJC["creatorName"];
                        string packageName = resourceJC["packageName"];
                        string resourceFullFileName = resourceJC["resourceFullFileName"];
                        
                        string rvgeKey = GetRVGKey(creatorName, packageName, resourceFullFileName);

                        if (resourceVersionGroupsDict.ContainsKey(rvgeKey))
                        {
                            PackagedRVGE rvge = resourceVersionGroupsDict[rvgeKey] as PackagedRVGE;
                            DataFileManifest.RegisterExisitingDFUserVARResource(rvge, varResourceUserFileName);

                            rvge.RestoreVARUserData(resourceJC);
                            
                        }
                        else varResourceCoreFileNamesToSave.Add(varResourceUserFileName);
                    }

                    
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.ResourceManifest.ReloadResourceVARUserData for " + varResourceUserFileName + ": " + e); }
            }
            foreach (string varResourceUserFileName in varResourceCoreFileNamesToSave) DataFileManifest.RegisterVARResourceUserSaveRequest(varResourceUserFileName);
        }

        public static void UpdateVARUserHideFlag(string rvgKey,int varVersion, bool hidden)
        {
            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                if (!rvge.localFileResource) 
                {
                    (rvge as PackagedRVGE).SetHideUserVARVersion(varVersion, hidden);
                }
                
            }
            
        }
        public static void UpdateVARUserFavFlag(string rvgKey, int varVersion, bool favourite)
        {
            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                if (!rvge.localFileResource)
                {
                    (rvge as PackagedRVGE).SetFavUserVARVersion(varVersion, favourite);
                }
                    
            }

        }

        public static IEnumerable<ResourceVersionGroupEntry> GetTagFilteredResourceGroups(ResourceBrowserFilterSettings filters)
        {
            var filteredRVGs = resourceVersionGroups.Where(o =>
                        o.resourceType == filters.resourceType && o.resourcePresent &&
                        (filters.showHiddenJSB.val || (!o.isHidden && !o.isTagHidden)) &&
                        (!filters.onlyFavouritesJSB.val || o.isFavorite) &&
                        (!filters.onlyNewJSB.val || o.isNew) &&
                        !o.isCSLReferenced);

            IEnumerable<ResourceVersionGroupEntry> tagFilteredRVGEs = filteredRVGs;

            if (filters.combiTagFilterEnabled)
            {
                tagFilteredRVGEs = filteredRVGs.Where(o => filters.combiTag.GetState(o));
            }
            else
            {
                if (filters.tagsFilterJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
                {
                    tagFilteredRVGEs = filteredRVGs.Where(o => o.IsTagged(TagManifest.GetTag(filters.resourceType, filters.tagsFilterJSMultiEnum.mainVal)));
                }
                else
                {
                    if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.hiddenOnly) tagFilteredRVGEs = filteredRVGs.Where(o => o.isHidden || o.isTagHidden);
                    if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.noTags) tagFilteredRVGEs = filteredRVGs.Where(o => o.IsNotTagged());
                    if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.noUserTags) tagFilteredRVGEs = filteredRVGs.Where(o => o.IsNotUserTagged());
                }
            }

            return tagFilteredRVGEs;
        }

        private static IEnumerable<ResourceVersionGroupEntry> GetCreatorFilteredResourceGroups(ResourceBrowserFilterSettings filters)
        {
            var tagFilteredRVGEs = GetTagFilteredResourceGroups(filters);

            IEnumerable<ResourceVersionGroupEntry> creatorFilteredRVGEs = tagFilteredRVGEs;
            if (filters.creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs)
                {
                    creatorFilteredRVGEs = tagFilteredRVGEs.Where(o =>
                        !o.localFileResource).ToList();
                }
                else if (filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.localOnly)
                {
                    if (!filters.inludeSubFoldersJSB.val) creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.localFileResource && o.containingLocalFolder == filters.selectedFolderFilter).ToList();
                    else creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.localFileResource && o.allContainingLocalFolders.Contains(filters.selectedFolderFilter)).ToList();
                }
            }
            else
            {
                creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.creatorName == filters.creatorJSMultiEnum.displayVal).ToList();
            }

            return creatorFilteredRVGEs;
        }

        private static IEnumerable<ResourceVersionGroupEntry> GetLicenseFilteredResourceGroups(ResourceBrowserFilterSettings filters)
        {
            var creatorFilteredRVGEs = GetCreatorFilteredResourceGroups(filters);

            IEnumerable<ResourceVersionGroupEntry> licenseFilteredRVGEs = creatorFilteredRVGEs;

            if (filters.licenseFilterJSEnum.val != LicenseType.Any && (filters.creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag || filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs))
            {
                licenseFilteredRVGEs = creatorFilteredRVGEs.Where(o => ((PackagedRVGE)o).varLicenseType == filters.licenseFilterJSEnum.val);
            }

            return licenseFilteredRVGEs;
        }

        public static List<int> GetLicenseTypesByInverseFilters(ResourceBrowserFilterSettings filters)
        {
            List<int> licenseTypesNotPresent = new List<int>();

            var creatorFilteredRVGEs = GetCreatorFilteredResourceGroups(filters);

            foreach (int licenseType  in EnumManifest.GetEnumsList(LicenseType.enumManifestName))
            {
                if (creatorFilteredRVGEs.Where(o => ((PackagedRVGE)o).varLicenseType == licenseType).Count() == 0) licenseTypesNotPresent.Add(licenseType);
            }
            licenseTypesNotPresent.Remove(LicenseType.Any);
            return licenseTypesNotPresent;
        }


        public static List<ResourceVersionGroupEntry> GetResourceGroups(ResourceBrowserFilterSettings filters)
        {
            var licenseFilteredRVGEs = GetLicenseFilteredResourceGroups(filters);
            var globalFilterRVGs = licenseFilteredRVGEs.Where(o =>(filters.resourceNameFilterQuick == "" ||(Settings.resourceNameSearchModeJSESC.val == ResourceNameSearchMode.StartsWith && o.resourceDisplayNameLowerCaps.StartsWith(filters.resourceNameFilterQuick)) || (Settings.resourceNameSearchModeJSESC.val == ResourceNameSearchMode.Contains && o.resourceDisplayNameLowerCaps.Contains(filters.resourceNameFilterQuick)) || (Settings.resourceNameSearchModeJSESC.val == ResourceNameSearchMode.ContainsFullPath && o.resourceFullPathReferenceWithAliasLowerCaps.Contains(filters.resourceNameFilterQuick))));

            if (filters.sortModeJSEnum.val == BrowserSortMode.AtoZ) globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.ZtoA) globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.newDateTime)
                 globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.firstCataloguedDateTime).ThenBy(o =>o.creatorName+ o.packageName + o.folderPath).ThenBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.reverseNewDateTime)
                globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.firstCataloguedDateTime).ThenByDescending(o => o.creatorName + o.packageName + o.folderPath).ThenByDescending(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.updatedDateTime)
                globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.lastCataloguedDateTime).ThenBy(o => o.creatorName + o.packageName + o.folderPath).ThenBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.reverseUpdatedDateTime)
                globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.lastCataloguedDateTime).ThenByDescending(o => o.creatorName + o.packageName + o.folderPath).ThenByDescending(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.packageFolder) globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.creatorName + o.packageName + o.folderPath).ThenBy(o => o.resourceDisplayName);

            return globalFilterRVGs.ToList();
        }
        public static HashSet<ResourceVersionGroupEntry> GetAllResourceGroups(int resourceType)
        {
            var results = new HashSet<ResourceVersionGroupEntry>(resourceVersionGroups.Where(o =>
                        o.resourceType == resourceType));
            return results;
        }
        public static string GetRVGKey(string creatorName, string packageName, string resourceFileName)
        {
            return creatorName + "." + packageName + ":" + resourceFileName;
        }
        public static string GetRVGKey (VARPackageManifestEntry vpme, string resourceFileName)
        {
            return GetRVGKey(vpme.creatorName, vpme.packageName, resourceFileName);
        }
        public static string GetRVGKey(VARPackageVersionGroupEntry vpvge, string resourceFileName)
        {
            return GetRVGKey(vpvge.creatorName, vpvge.packageName, resourceFileName);
        }
        public static string GetRVGKey(PackagedRVGE packagedRVGE)
        {
            return GetRVGKey(packagedRVGE.creatorName, packagedRVGE.packageName, packagedRVGE.resourceFolderPathFileName);
        }

        public static void RegisterNewVARResources(VARPackageManifestEntry vpme, HashSet<int> notNewResourceTypes)
        {
            HashSet<string> contentFiles ;
            HashSet<string> hideFiles = new HashSet<string>();
            HashSet<string> favFiles = new HashSet<string>();

            contentFiles = Utils.GetFilesAtPathRecursive(vpme.varFilePath + ":", "*.*");

            int length = vpme.varFilePath.Length + 2;
            foreach (string contentFile in contentFiles)
            {
                if (contentFile.EndsWith(".hide")) hideFiles.Add(contentFile);
                else if (contentFile.EndsWith(".fav")) favFiles.Add(contentFile);
                else
                {
                    int resourceType = GetResourceType(contentFile.Substring(length));
                    bool notNew = notNewResourceTypes.Contains(resourceType);
                    if (resourceType!=ResourceType.unknown) RegisterNewVARResourceOfType(vpme, contentFile.Substring(length), resourceType, notNew);
                }
            }

            HashSet<string> csListFiles =  new HashSet<string>(contentFiles.Where(fileName => fileName.EndsWith(".cslist") && fileName.Contains(":\\Custom\\Scripts")));
            ProcessVARCSListReferences(vpme, csListFiles);

            RegisterCreatorHideFlags(vpme, hideFiles);
            RegisterCreatorFavFlags(vpme, favFiles);
        }

        public static bool RegisterVARResourceFromManifest(VARPackageVersionGroupEntry vpvge, JSONClass resourceJC, string varResourceMetaFileName)
        {

            string resourceFullFileName = resourceJC["resourceFullFileName"];
            string rvgKey = GetRVGKey(vpvge, resourceFullFileName);
            
            if (resourceVersionGroupsDict.ContainsKey(rvgKey)) return true;
            else
            {
                JSONArray varVersionsJA = resourceJC["varVersions"].AsArray;
                if (vpvge.AnyExistingVARVersions(varVersionsJA))
                {
                    PackagedRVGE rvge = new PackagedRVGE(resourceJC, vpvge);
                    resourceVersionGroupsDict.Add(rvgKey, rvge);
                    resourceVersionGroups.Add(rvge);
                    varResourceVersionGroups.Add(rvge);
                    CreatorManifest.RegisterResource(rvge);
                    DataFileManifest.RegisterExisitingDFCoreVARResource(rvge, varResourceMetaFileName);
                    vpvge.RegisterRVGE(rvge);

                    if (vpvge.AnyMismatchedVARVersions(varVersionsJA)) rvge.SetCoreDataPersist();
                    return true;
                }
                
            }

            return false;
        }
        private static void RegisterCreatorHideFlags(VARPackageManifestEntry vpme, HashSet<string> hideFiles)
        {
            foreach (string hideFile in hideFiles)
            {               
                string hideFileNoExt = hideFile.Substring(0, hideFile.Length - 5);
                string hideFilPath = hideFileNoExt.Substring(hideFileNoExt.IndexOf(":\\") + 2);

                string rvgKey = GetRVGKey(vpme, hideFilPath);
                
                if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                {
                    ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                    (rvge as PackagedRVGE).SetHideCreatorVARVersion(vpme.varVersion);
                }
            }
        }
        private static void RegisterCreatorFavFlags(VARPackageManifestEntry vpme, HashSet<string> favFiles)
        {
            foreach (string favFile in favFiles)
            {
                string favFileNoExt = favFile.Substring(0, favFile.Length - 4);
                string favFilPath = favFileNoExt.Substring(favFileNoExt.IndexOf(":\\") + 2);

                string rvgKey = GetRVGKey(vpme, favFilPath);
                if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                {
                    ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                    (rvge as PackagedRVGE).SetFavCreatorVARVersion(vpme.varVersion);
                }
            }
        }

        private static void ProcessVARCSListReferences(VARPackageManifestEntry vpme, HashSet<string> csListFiles)
        {

            string[] stringSeparators = new string[] { ".cs", "\n", "\r\n" };
            string csListFileContents = null;

            foreach (string cslFilePath in csListFiles)
            {
                try
                {
                    csListFileContents = FileManagerSecure.ReadAllText(cslFilePath);
                }
                catch (Exception e)
                {
                    Utils.LogDebug("<ResourceManifest.ProcessCSListReferences> '" + cslFilePath + "' could not be read by VAM - possible malformed file");
                    break;
                }
                if (csListFileContents != null)
                {
                    string[] csListPluginFiles = csListFileContents.Split(stringSeparators, System.StringSplitOptions.RemoveEmptyEntries);

                    string cslPath = FileManagerSecure.GetDirectoryName(cslFilePath.Substring(cslFilePath.IndexOf(":\\")+2));
                    foreach (string csFileRef in csListPluginFiles)
                    {
                        if (!(String.IsNullOrEmpty(csFileRef) || csFileRef.Trim().Length == 0) && !csFileRef.Contains(":"))
                        {
                            string rvgKey = GetRVGKey(vpme, cslPath+"\\"+csFileRef.Replace('/','\\') + ".cs");
                            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                            {
                                 var rvge = (PackagedRVGE) resourceVersionGroupsDict[rvgKey];
                                rvge.SetCSLFileReferenced(vpme.varVersion);
                            }
                            

                        }

                    }
                }
            }
                
        }

        private static int GetResourceType(string resourceFileName)
        {
            if (resourceFileName.StartsWith("Saves\\scene") && resourceFileName.EndsWith(".json")) return ResourceType.scene;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\AnimationPresets") && resourceFileName.EndsWith(".vap")) return ResourceType.presetAnimation;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Appearance") && resourceFileName.EndsWith(".vap")) return ResourceType.presetAppearance;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\BreastPhysics") && resourceFileName.EndsWith(".vap")) return ResourceType.presetBreastPhys;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Clothing") && resourceFileName.EndsWith(".vap")) return ResourceType.presetClothing;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\General") && resourceFileName.EndsWith(".vap")) return ResourceType.presetGeneral;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\GlutePhysics") && resourceFileName.EndsWith(".vap")) return ResourceType.presetGlutePhysics;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Hair") && resourceFileName.EndsWith(".vap")) return ResourceType.presetHair;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Morphs") && resourceFileName.EndsWith(".vap")) return ResourceType.presetMorphs;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Plugins") && resourceFileName.EndsWith(".vap")) return ResourceType.presetPlugins;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Pose") && resourceFileName.EndsWith(".vap")) return ResourceType.presetPose;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Skin") && resourceFileName.EndsWith(".vap")) return ResourceType.presetSkin;
            //           if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Textures") && (resourceFileName.EndsWith(".jpg")|| resourceFileName.EndsWith(".png"))) return ResourceType.personTexture;
            //           if (resourceFileName.StartsWith("Custom\\Textures") && (resourceFileName.EndsWith(".jpg") || resourceFileName.EndsWith(".png"))) return ResourceType.atomTexture;

            if (resourceFileName.StartsWith("Custom\\Scripts") && (resourceFileName.EndsWith(".cs") || resourceFileName.EndsWith(".cslist") || resourceFileName.EndsWith(".dll"))) return ResourceType.plugins;

            return ResourceType.unknown;
        }

        private static void RegisterNewVARResourceOfType(VARPackageManifestEntry vpme, string resourceFileName,int resourceType, bool forceNotNew)
        {
            string rvgKey = GetRVGKey(vpme, resourceFileName);

            if (!resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                PackagedRVGE rvge = new PackagedRVGE(vpme, resourceFileName, resourceType, !forceNotNew);
                resourceVersionGroupsDict.Add(rvgKey, rvge);
                resourceVersionGroups.Add(rvge);
                varResourceVersionGroups.Add(rvge);
                CreatorManifest.RegisterResource(rvge);
                DataFileManifest.RegisterNewDFCoreVARResource(rvge);
                vpme.vpvge.RegisterRVGE(rvge);
            }
            else
            {
                PackagedRVGE rvge = resourceVersionGroupsDict[rvgKey] as PackagedRVGE;
                rvge.AddVARVersion(vpme.varVersion, !forceNotNew);
                rvge.SetCoreDataPersist();
            }
        }



        public static void ClearAllUserFilePrefs()
        {
            foreach (PackagedRVGE packegedRVGE in varResourceVersionGroups)
            {
                packegedRVGE.ClearAllUserFilePrefs();
            }
        }


    }


    public abstract class ResourceVersionGroupEntry
    {
        
        protected int baFavorite = TriState.indeterminate;
        protected int baHidden = TriState.indeterminate;
        public virtual string creatorName { get; set; }
        public virtual string packageName { get; set; }
        public virtual bool resourcePresent { get { return true; } }

        // Path of Folder containing the Resource
        public string folderPath { get; protected set; }

        // Resource File Name with no path or file extension.
        public string resourceFileNameNoExt { get; private set; }

        public string resourceFileNameNoExtNoPrefix { get; private set; }

        public JSONStorableString resourceNameAliasJSS;
        public JSONStorableString notesJSS ;

        //Resource Filename (with no path or extension) with prefix removed for Presets
        public string resourceDisplayName { get
            {
                if (resourceNameAliasJSS.val == "") return resourceFileNameNoExtNoPrefix;
                return resourceNameAliasJSS.val;
            }
        }

        // Resource File Name in lower case with no path. Includes file extension.
        public string resourceDisplayNameLowerCaps { get; private set; }
       
        public virtual DateTime firstCataloguedDateTime { get; }
        public virtual DateTime lastCataloguedDateTime { get; }

        public string resourceFolderPathFileName { get; protected set; }
        
        public string resourceFolderPathFileNameNoExt { get; protected set; }
        public string resourceFullPathReference { get; protected set; }
        public string resourceFullPathReferenceWithAliasLowerCaps { get; protected set; }

        public string resourceFileNameExtension { get; protected set; }
        public int resourceType { get; protected set; }

        private ResourceSettings resourceTypeSettings { get { return Settings.resourceSettings[resourceType]; } }
        public int resourceCategory { get; protected set; }
        public int resourceTargetCategory { get; protected set; }
        public bool localFileResource { get; protected set; } = false;

        protected virtual bool vamUserHidden { get;  set; } = false;
        protected virtual bool vamUserFavorite { get;  set; } = false;

        public bool isTagHidden { get
            {
                return TagManifest.GetRVGETags(this).Where(a => a.hideTaggedResourcesJSB.val).Any();
            } }
        public bool isRenameable
        {
            get
            {
                if (!localFileResource) return false;
                if (resourceCategory == ResourceCategory.scene && !Globals.scenesFolderAccess) return false;
                if (resourceCategory == ResourceCategory.plugins && !Globals.customFolderAccess) return false;
                if (resourceCategory == ResourceCategory.preset && !Globals.customFolderAccess) return false;
                return true;
            }
        }
        public  bool isDeletable { get
            {
                if (!localFileResource && Globals.addonPackagesAccess) return true;
                if (localFileResource)
                {
                    if (resourceCategory == ResourceCategory.scene && Globals.scenesFolderAccess) return true;
                    if ((resourceCategory == ResourceCategory.plugins || resourceCategory == ResourceCategory.preset) && Globals.customFolderAccess) return true;
                }
                return false;
            } }
        public virtual bool isHidden
        {
            get
            {
                return (vamUserHidden || baHidden == TriState.active);
            }
            set
            {
                if (value)
                {
                    if (baHidden != TriState.active)
                    {
                        baHidden = TriState.active;
                        SetUserDataPersist();
                    }

                }
                else
                {
                    if (baHidden != TriState.inactive)
                    {
                        baHidden = TriState.inactive;
                        SetUserDataPersist();
                    }
                }
                SetVAMUserHidden(value);
                foreach (CombiTag combiTag in TagManifest.GetCombiTags(resourceType))
                {
                    if (combiTag.usesHidden) SetTag(combiTag);
                }               
            }
        }
        public abstract bool isCSLReferenced { get; }

        public virtual bool isMultiVARVersion { get { return false; } }
        public virtual bool isFavorite
        {
            get
            {
                return (vamUserFavorite || baFavorite == TriState.active);
            }
            set
            {
                if (value)
                {
                    if (baFavorite != TriState.active)
                    {
                        baFavorite = TriState.active;
                        SetUserDataPersist();
                    }

                }
                else
                {
                    if (baFavorite != TriState.inactive)
                    {
                        baFavorite = TriState.inactive;
                        SetUserDataPersist();
                    }
                }
                SetVAMUserFavourite(value);
                foreach (CombiTag combiTag in TagManifest.GetCombiTags(resourceType))
                {
                    if (combiTag.usesFavorite) SetTag(combiTag);
                }
            }
        }

        protected bool _isNew = false;
        public bool isNew
        {
            get
            {
                return _isNew;
            }
            set
            {
                if (_isNew != value)
                {
                    _isNew = value;
                    SetUserDataPersist();
                }
                foreach (CombiTag combiTag in TagManifest.GetCombiTags(resourceType))
                {
                    if (combiTag.usesNew) SetTag(combiTag);
                }
            }
        }


        public Sprite resourceSprite = null;

        public abstract void SetUserDataPersist();

        protected abstract void SetVAMUserFavourite(bool value);
        protected abstract void SetVAMUserHidden(bool value);

        public virtual string resourceThumbnailReference { get; }

        protected UIDynamicButton queuedButtonLoad=null;
        protected ImageLoaderThreaded.QueuedImage queuedImage = null;

        public ResourceVersionGroupEntry() {
            resourceNameAliasJSS = new JSONStorableString("resourceNameAlias", "", delegate (string x)
            {
                SetUserDataPersist();
                if (BrowserAssist.mainBrowserUI != null)
                {
                    BrowserAssist.mainBrowserUI.resourceBrowserUI. UpdateDisplayName(this);
                    BrowserAssist.mainBrowserUI.resourceSelectionsUI. UpdateDisplayName(this);
                }
            });
            notesJSS = new JSONStorableString("notes", "", delegate (string x)
            {
                SetUserDataPersist();
            });
        }

        private Folder _containingLocalFolder = null;
        public Folder containingLocalFolder { get { return _containingLocalFolder; } protected set
            {
                if (_containingLocalFolder != null) _containingLocalFolder.UnRegisterLocalRVGE(this);
                _containingLocalFolder = value;
                _containingLocalFolder.RegisterLocalRVGE(this);
            } } 
        protected virtual void Init()
        {
            int fileExtensionIndex = resourceFolderPathFileName.LastIndexOf('.');
            int fileNameIndex = 0;
            resourceCategory = ResourceType.GetResourceCategory(resourceType);


            if (resourceFolderPathFileName.Contains("\\")) fileNameIndex= resourceFolderPathFileName.LastIndexOf('\\');
            folderPath = resourceFolderPathFileName.Substring(0, fileNameIndex);
            resourceFileNameNoExt = resourceFolderPathFileName.Substring(folderPath.Length + 1, fileExtensionIndex - fileNameIndex - 1);
            if (resourceCategory == ResourceCategory.preset) resourceFileNameNoExtNoPrefix= resourceFileNameNoExt.Substring(7);
            else resourceFileNameNoExtNoPrefix = resourceFileNameNoExt;
           
            resourceDisplayNameLowerCaps = resourceDisplayName.ToLower();
            resourceFileNameExtension = resourceFolderPathFileName.Substring(fileExtensionIndex + 1);
            resourceFolderPathFileNameNoExt = resourceFolderPathFileName.Substring(0, fileExtensionIndex);

            resourceTargetCategory = ResourceType.GetResourceTargetCategory(resourceType);
            if (localFileResource && containingLocalFolder==null)
            {
                containingLocalFolder = FolderBrowser.GetFolder(folderPath, resourceType);
            }
        }

        public HashSet<Folder> allContainingLocalFolders { get
            {
                return containingLocalFolder.allContainingLocalFolders;
            } }

        public virtual JSONClass GetCoreJSON()
        {
            JSONClass jc = new JSONClass();

            jc["resourceFullFileName"] = resourceFolderPathFileName;
            jc["resourceType"] = EnumManifest.GetStoreVal(ResourceType.enumManifestName,resourceType);
            

            return (jc);
        }

        protected virtual void RestoreTags(JSONArray tagsArray)
        {
            foreach (JSONClass tagJC in tagsArray)
            {
                int tagCategory = EnumManifest.GetEnumValFromStoreVal(TagCategory.enumManifestName, tagJC["tagCategory"]);
                string tagName = tagJC["tagName"];
                if (tagCategory == TagCategory.userDefined) TagManifest.CreateTag(tagName, tagCategory, resourceType);
                if (tagCategory != TagCategory.combi) TagManifest.SetRVGETagged(this, tagName, true);
            }
        }
        protected virtual JSONNode GetTagsJSON()
        {
            JSONNode ja = new JSONArray();

            foreach(var tag in TagManifest.GetRVGETags(this))
            {
                JSONClass tagJC = new JSONClass();
                tagJC["tagName"] = tag.tagName;
                tagJC["tagCategory"] = tag.tagCategoryJSEnum.storeVal;

                ja[-1] = tagJC;
            }

            return (ja);
        }

        public IEnumerator Activate(ResourcePrimaryActionUI resourceActions)
        {
            resourceActions.parentBrowser.resourceBrowserUI.SetButtonLoading(this);
            yield return new WaitForSeconds(0.01f);
            if (isNew)
            {
                isNew = false;
                resourceActions.SetNewIcon(this);
            }

            if (resourceTypeSettings.closeBrowserUIonResourceLoadJSB.val) resourceActions.parentBrowser.SetInactive();

            if (resourceType == ResourceType.scene)
            {
                if (resourceActions.parentBrowser.isVREnabled) resourceActions.parentBrowser.SetInactive();
                if (resourceActions.mergePresetJSB.val) SuperController.singleton.LoadMerge(FileManagerSecure.NormalizePath(resourceFullPathReference));
                else SuperController.singleton.Load(FileManagerSecure.NormalizePath(resourceFullPathReference));
            }
            else if (resourceCategory == ResourceCategory.preset)
            {
                bool createNewAtom = false;
                string atomName;
                if (resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
                {
                    atomName = AtomUtils.GetNextAvailableAtomName("Person");

                    yield return SuperController.singleton.AddAtomByType("Person", atomName, true);
                    
                    createNewAtom = true;
                }
                else atomName = resourceActions.targetAtomJSMultiEnum.mainVal;

                yield return new WaitForEndOfFrame();               

                if (AtomUtils.GetAtomNames(includeNonPersonAtoms: false).Contains(atomName))
                {
                    if (createNewAtom)
                    {
                        AtomUtils.TeleportAtomToSpawnPoint(atomName, ResourcePositionActionUI.newAtomPosition.GetPosition(), ResourcePositionActionUI.newAtomPosition.GetRotation());
                        
                    }

                    Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                    PresetLockStore atomPresetLockStore = new PresetLockStore();
                    if (atom.type == "Person") atomPresetLockStore.StorePresetLocks(atom, true,resourceActions.suppressClothingLoadJSB.val);

                    if (resourceType == ResourceType.presetAppearance && resourceActions.suppressScaleChangeJSB.val && PC.patreonContentEnabled) PC.AppPresetSuppressScaleLoad(atom, resourceFullPathReference, resourceActions.mergePresetJSB.val);
                    else
                    {
                        JSONStorable js = null;
                        if (resourceType == ResourceType.presetAppearance) js = atom.GetStorableByID("AppearancePresets");
                        else if (PC.patreonContentEnabled) js = PC.GetPresetStorable(atom, resourceType);

                        if (js != null)
                        {
                            JSONStorableBool loadOnSelectJSB = js.GetBoolJSONParam("loadPresetOnSelect");

                            bool loadOnSelectPreState = loadOnSelectJSB.val;
                            loadOnSelectJSB.val = false;

                            bool preState1 = false;
                            bool preState2 = false;
                            bool preState3 = false;

                            if (resourceType == ResourceType.presetMorphs)
                            {
                                JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                preState1 = includePhysical.val;
                                includePhysical.val = resourceActions.morphPresetIncludePhysicalJSB.val;
                                JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                preState2 = includeAppearance.val;
                                includeAppearance.val = resourceActions.morphPresetIncludeAppJSB.val;
                            }
                            if (resourceType == ResourceType.presetGeneral)
                            {
                                JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                preState1 = includePhysical.val;
                                includePhysical.val = resourceActions.generalPresetIncludePhysicalJSB.val;

                                JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                preState2 = includeAppearance.val;
                                includeAppearance.val = resourceActions.generalPresetIncludeAppJSB.val;

                                JSONStorableBool includePose = js.GetBoolJSONParam("includeOptional");
                                preState3 = includePose.val;
                                includePose.val = resourceActions.generalPresetIncludePoseJSB.val;
                            }

                            JSONStorableUrl presetPathJSON = js.GetUrlJSONParam("presetBrowsePath");
                            presetPathJSON.val = SuperController.singleton.NormalizePath(resourceFullPathReference);
                            if (resourceActions.mergePresetJSB.val && resourceActions.targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag) js.CallAction("MergeLoadPreset");
                            else js.CallAction("LoadPreset");

                            loadOnSelectJSB.val = loadOnSelectPreState;

                            if (resourceType == ResourceType.presetMorphs)
                            {
                                JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                includePhysical.val = preState1;
                                JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                includeAppearance.val = preState2;
                            }
                            if (resourceType == ResourceType.presetGeneral)
                            {
                                JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                includePhysical.val = preState1;

                                JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                includeAppearance.val = preState2;

                                JSONStorableBool includePose = js.GetBoolJSONParam("includeOptional");
                                includePose.val = preState3;
                            }
                        }                           
                    }                        
                    atomPresetLockStore.RestorePresetLocks(atom);

                    if (createNewAtom) resourceActions.targetAtomJSMultiEnum.val = atomName;
                    if (Settings.selectAtomOnLoadJSB.val && resourceActions.isAtomTargetable)
                    {
                        FreeControllerV3 fcV3 = atom.freeControllers.First(fc => fc.name == "control");
                        SuperController.singleton.SelectController(fcV3);
                    }
                }
            }
            else if (resourceCategory == ResourceCategory.plugins && PC.patreonContentEnabled)
            {
                Atom atom = null;
                if (resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
                {
                    if (AtomUtils.GetAtomNames(includeNonPersonAtoms: true).Contains(resourceActions.targetAtomJSMultiEnum.mainVal))
                    {
                        atom = SuperController.singleton.GetAtomByUid(resourceActions.targetAtomJSMultiEnum.mainVal);
                        
                    }
                }
                else if(resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
                {
                    if (resourceActions.targetAtomJSMultiEnum.valTopEnum == PluginTargetTypes.sessionPlugins) atom = Globals.mvrScript.containingAtom;
                    else atom = SuperController.singleton.GetAtomByUid("CoreControl");
                }

                PC.LoadPluginAsPreset(atom, resourceFullPathReference, resourceActions.openPluginUIJSB.val);

                if (Settings.selectAtomOnLoadJSB.val && resourceActions.isAtomTargetable) {
                    FreeControllerV3 fcV3 = atom.freeControllers.First(fc => fc.name == "control");
                    SuperController.singleton.SelectController(fcV3);
                }
            }

            yield return new WaitForSeconds(1f);
            
            resourceActions.parentBrowser.resourceBrowserUI.UnsetButtonLoading(this);
        }





        private void QueuedImageLoadCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi == queuedImage)
            {
                if (resourceSprite != null) GameObject.Destroy( resourceSprite.texture);
                resourceSprite = Sprite.Create(qi.tex, new Rect(0.0f, 0.0f, qi.tex.width, qi.tex.height), new Vector2(0.5f, 0.5f), 100.0f);               
                LoadResourceTexture(queuedButtonLoad);
            }
            else GameObject.Destroy(qi.tex);
            
        }

        private void LoadResourceTexture(UIDynamicButton button)
        {
            if (button != null)
            {
                Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
                UnityEngine.UI.Image image = newImageTrans.GetComponent<UnityEngine.UI.Image>();
                image.sprite = resourceSprite;
            }
        }

        public void CancelButtonImageUpdate()
        {
            queuedButtonLoad = null;
        }

        private void QueuedImageReloadCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi.tex != null)
            {
                resourceSprite = Sprite.Create(qi.tex, new Rect(0.0f, 0.0f, qi.tex.width, qi.tex.height), new Vector2(0.5f, 0.5f), 100.0f);
            }
            

            BrowserAssist.mainBrowserUI.resourceBrowserUI.ReloadSprite(this);
            BrowserAssist.mainBrowserUI.resourceActionsUI.RefreshUI();
        }

        public void ReloadSprite()
        {
            if (resourceSprite != null)
            {
                GameObject.Destroy(resourceSprite.texture);
                GameObject.Destroy(resourceSprite);
            }
            
            resourceSprite = null;

            string url = FileManagerSecure.NormalizePath(resourceThumbnailReference);

            ImageUtils.QueueLoadTexture(url, QueuedImageReloadCallback, true);
        }

        public void LoadSprite(UIDynamicButton button)
        {
            string url = FileManagerSecure.NormalizePath( resourceThumbnailReference);

            Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
            GameObject newImageGO = newImageTrans.gameObject;           

            UnityEngine.UI.Image image = newImageGO.GetComponent<UnityEngine.UI.Image>();

            if (string.IsNullOrEmpty(url) || !FileManagerSecure.FileExists(url))
            {                
                resourceSprite = null;
                image.sprite = null;
                return;
            }
            if (resourceSprite != null)
            {
                LoadResourceTexture(button);
            }
            else 
            {
                image.sprite = null;
                queuedButtonLoad = button;
                queuedImage = ImageUtils.QueueLoadTexture(url, QueuedImageLoadCallback,true);
            }
           
        }

        public virtual bool IsTagged(Tag tag)
        {
            return TagManifest.IsRVGETagged(this, tag);
        }

        public virtual bool IsNotTagged()
        {
            return TagManifest.IsRVGENotTagged(this);
        }

        public virtual bool IsNotUserTagged()
        {
            return TagManifest.IsRVGENotUserTagged(this);
        }

        public virtual void SetTag(CombiTag combiTag)
        {
            SetTag(combiTag, combiTag.GetState(this));
        }

        public virtual void SetTag(Tag tag, bool value)
        {
            bool currentTagValue = TagManifest.IsRVGETagged(this, tag);

            if (value != currentTagValue)
            {
                TagManifest.SetRVGETagged(this, tag, value);
                foreach (CombiTag combiTag in tag.usedByCombiTags) SetTag(combiTag);

                if (tag.hideTaggedResourcesJSB.val)
                {
                    BrowserAssist.mainBrowserUI.resourceBrowserUI.SetHideFavColorTags(this);
                    if (!BrowserAssist.mainBrowserUI.browserFiltersUI.showHiddenJSB.val) BrowserAssist.mainBrowserUI.RefreshFilterResults();
                }
                SetUserDataPersist();
            }
            
        }
    }

    public class LocalRVGE :ResourceVersionGroupEntry
    {
        public override string creatorName { get { return ""; } }
        public override string packageName { get { return ""; } }

        private LocalResourceUserDataFile _userDataFile = null;
        public LocalResourceUserDataFile userDataFile
        {
            get
            {
                if (_userDataFile == null) DataFileManifest.RegisterNewDFUserLocalResource(this);
                return _userDataFile;
            }
            private set { _userDataFile = value; }
        }

        public override bool isCSLReferenced
        {
            get
            {
                return csListReferenced;
            }
        }
        public override DateTime firstCataloguedDateTime
        {
            get { return localFileResourceFirstScan; }
        }

        // For Local resources, first and last dates are the same (unfortunately)
        public override DateTime lastCataloguedDateTime
        {
            get { return localFileResourceFirstScan; }
        }
        public override string resourceThumbnailReference
        {
            get
            {
                return resourceFolderPathFileNameNoExt + ".jpg";
            }
        }
        private DateTime localFileResourceFirstScan;

        private bool isLocalResourceSymLinkAvaliable
        {
            get
            {
                if (resourceCategory == ResourceCategory.preset && Globals.customFolderAccess) return true;
                if (resourceCategory == ResourceCategory.scene && Globals.scenesFolderAccess) return true;
                return false;
            }
        }

        public LocalRVGE(string resourceFileName, int _resourceType) : base()
        {
            resourceType = _resourceType;
            resourceFolderPathFileName = resourceFileName;
            localFileResource = true;
            Init();

            RefreshLocalHideFavFlags();
        }

        protected override void Init()
        {
            base.Init();

            resourceFullPathReference = resourceFolderPathFileName;
            resourceFullPathReferenceWithAliasLowerCaps = (resourceFullPathReference + resourceNameAliasJSS.val).ToLower();
        }

        public void ResetFolderPath()
        {
            resourceFolderPathFileName = containingLocalFolder.GetFolderPath() + "\\" + resourceFileNameNoExt+"."+resourceFileNameExtension;
            Init();
            SetUserDataPersist();
        }

        public void MoveFolderPath(Folder targetFolder)
        {
            containingLocalFolder = targetFolder;
            ResetFolderPath();
        }

        public void RenameRVGEFilename(string newFullPathFileName)
        {
            resourceFolderPathFileName = newFullPathFileName;
            Init();
            SetUserDataPersist();
        }

        protected override void SetVAMUserHidden(bool value)
        {
            bool success = SetVAMUserHideFav(value, ".hide");
            if (success) vamUserHidden = value;
        }
        protected override void SetVAMUserFavourite(bool value)
        {
            bool success = SetVAMUserHideFav(value, ".fav");
            if (success) vamUserFavorite = value;
        }

        private bool csListReferenced = false;

        private bool SetVAMUserHideFav(bool value, string fileExt)
        {
            if (isLocalResourceSymLinkAvaliable)
            {
                string path = Const.SymLinksFolder + "\\" + folderPath;
                string tagFileName = path + "\\" + resourceFileNameNoExt + "." + resourceFileNameExtension + fileExt;
                if (!FileManagerSecure.DirectoryExists(path) && value) FileManagerSecure.CreateDirectory(path);
                if (FileManagerSecure.DirectoryExists(path))
                {
                    if (value) FileManagerSecure.WriteAllBytes(tagFileName, new byte[] { 0 });
                    else FileManagerSecure.DeleteFile(tagFileName);
                    return true;
                }
            }
            return false;
        }

        public void SetLocalFileScanTime(DateTime scanTime)
        {
            localFileResourceFirstScan = scanTime;
            SetUserDataPersist();
        }

        public override void SetUserDataPersist()
        {
            DataFileManifest.RegisterLocalResourceUserSaveRequest(userDataFile);
        }

        public void LoadLegacyLocalUserData(string fileName)
        {
             
            JSONClass resourceJC = SuperController.singleton.LoadJSON(fileName).AsObject;
            if (resourceJC["localFileFirstScanDateTime"] != null) localFileResourceFirstScan = DateTime.Parse(resourceJC["localFileFirstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);

            if (resourceJC["baFavourite"] != null) baFavorite = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baFavourite"]);
            if (resourceJC["baHidden"] != null) baHidden = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baHidden"]);
            if (resourceJC["baNew"] != null) _isNew = resourceJC["baNew"].AsBool;
            else _isNew = false;

            if (resourceJC["Tags"] != null) RestoreTags(resourceJC["Tags"].AsArray);

            // Initial alpha release had a wrong JSON Node Name for the file format
 //           if (resourceJC["VARManifestStoreFormat"] != null) SetUserDataPersist();           
        }
        public void LoadLocalUserData(JSONClass resourceJC)
        {

            if (resourceJC["localFileFirstScanDateTime"] != null) localFileResourceFirstScan = DateTime.Parse(resourceJC["localFileFirstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);

            if (resourceJC["baFavourite"] != null) baFavorite = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baFavourite"]);
            if (resourceJC["baHidden"] != null) baHidden = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baHidden"]);
            if (resourceJC["baNew"] != null) _isNew = resourceJC["baNew"].AsBool;
            else _isNew = false;


            if (resourceJC["Tags"] != null) RestoreTags(resourceJC["Tags"].AsArray);

            if (resourceJC[resourceNameAliasJSS.name] != null) resourceNameAliasJSS.valNoCallback = resourceJC[resourceNameAliasJSS.name];
            if (resourceJC[notesJSS.name] != null) notesJSS.valNoCallback = resourceJC[notesJSS.name];

            Init();
        }

        public override JSONClass GetCoreJSON()
        {
            JSONClass resourceJC = base.GetCoreJSON();

            resourceJC["localFileFirstScanDateTime"] = localFileResourceFirstScan.ToString("O");
            if (baFavorite != TriState.indeterminate) resourceJC["baFavourite"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baFavorite);
            if (baHidden != TriState.indeterminate) resourceJC["baHidden"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baHidden);
            if (isNew != false) resourceJC["baNew"].AsBool = isNew;


            if (TagManifest.GetRVGETags(this).Count>0) resourceJC["Tags"] = GetTagsJSON();

            resourceNameAliasJSS.StoreJSON(resourceJC);
            notesJSS.StoreJSON(resourceJC);

            return resourceJC;
        }

        private void RefreshLocalHideFavFlags()
        {
            if (FileManagerSecure.FileExists(resourceFolderPathFileName + ".hide")) vamUserHidden = true;
            else vamUserHidden = false;
            if (FileManagerSecure.FileExists(resourceFolderPathFileName + ".fav")) vamUserFavorite = true;
            else vamUserFavorite = false;
        }

        public void RegisterUserDataFile(LocalResourceUserDataFile userDataFile)
        {
            this.userDataFile = userDataFile;
        }

        public void DeRegisterUserDataFile()
        {
            this.userDataFile = null;
        }
    }

    public class PackagedRVGE : ResourceVersionGroupEntry
    {
        public VARPackageVersionGroupEntry vpvge { get; private set; }

        private HashSet<int> varVersionsCreatorHidden = new HashSet<int>();
        private HashSet<int> varVersionsUserHidden = new HashSet<int>();
        private HashSet<int> varVersionsCreatorFav = new HashSet<int>();
        private HashSet<int> varVersionsUserFav = new HashSet<int>();
        private HashSet<int> varVersions = new HashSet<int>();
        private HashSet<int> varVersionsCSListReferenced = new HashSet<int>();

        private VARResourcesUserDataFile _userDataFile = null;

        public VARResourcesCoreDataFile coreDataFile { get; private set; }
        public VARResourcesUserDataFile userDataFile
        {
            get
            {
                if (_userDataFile == null) DataFileManifest.RegisterNewDFUserVARResource(this);
                return _userDataFile;
            }
            private set { _userDataFile = value; }
        }


        //       private Dictionary<int, DateTime> varVersionFirstScanTimes = new Dictionary<int, DateTime>();

        public int highestVARVersion { get; private set; }
        private bool packageCreatorHidden { get; set; } = false;
        private bool packageCreatorFav { get; set; } = false;

        public int varLicenseType { get { return vpvge.varVersionLicenses[highestVARVersion]; } }

        public override string creatorName
        {
            get {return vpvge.creatorName; }
        }
        public override string packageName
        {
            get { return vpvge.packageName; }
        }

        public string latestVARName { get { return creatorName + "." + packageName + "." + highestVARVersion.ToString(); } }

        protected override bool vamUserHidden { get
            {
                if (varVersionsUserHidden.Count == 0) return false;
                if (varVersionsUserHidden.Max() == highestVARVersion) return true;
                return false;
            }
        }
        protected override bool vamUserFavorite { get
            {
                if (varVersionsUserFav.Count == 0) return false;
                if (varVersionsUserFav.Max() == highestVARVersion) return true;
                return false;
            }
        }

        public string customThumbnailFolder { get {
            return Const.VARResourceThumbnails + "\\" + vpvge.varGroupName + "\\" + folderPath;
            } }

        public string customThumbnailFile
        {
            get
            {
                return customThumbnailFolder+"\\"+resourceFileNameNoExt+".jpg";
            }
        }

        public override bool isCSLReferenced
        {
            get
            {
                if (varVersionsCSListReferenced.Contains(highestVARVersion)) return true;
                return false;
            }
        }

        public override DateTime firstCataloguedDateTime
        {
            get {
               return vpvge.firstCataloguedDateTime; }
        }

        public override DateTime lastCataloguedDateTime
        {
            get
            {
                return vpvge.lastCataloguedDateTime;
            }
        }

        public override bool resourcePresent
        {
            get
            {
                if (vpvge.highestVARVersion == highestVARVersion) return true;
                return false;
            }
        }

        public override bool isMultiVARVersion { get { return (varVersions.Count>1); } }
        public int versionsCount { get { return varVersions.Count(); } }

        public override bool isHidden
        {
            get
            {
                return (vamUserHidden || packageCreatorHidden);
            }
        }

        public override bool isFavorite
        {
            get
            {
                return (vamUserFavorite || packageCreatorFav );
            }
        }

        public override void SetUserDataPersist()
        {
            DataFileManifest.RegisterVARResourceUserSaveRequest(userDataFile);
        }
        public void SetCoreDataPersist()
        {
            DataFileManifest.RegisterVARResourceCoreSaveRequest(coreDataFile);
        }

        
        protected override void SetVAMUserFavourite(bool value)
        {
            bool success=SetVAMUserHideFav(value, ".fav");
            if (success) {
                if (value) varVersionsUserFav.Add(highestVARVersion);
                else varVersionsUserFav.Remove(highestVARVersion);
            }
        }
        protected override void SetVAMUserHidden(bool value)
        {
            bool success = SetVAMUserHideFav(value, ".hide");
            if (success)
            {
                if (value) varVersionsUserHidden.Add(highestVARVersion);
                else varVersionsUserHidden.Remove(highestVARVersion);
            }
        }

        private bool SetVAMUserHideFav(bool value, string fileExt)
        {
            if (Globals.addonPackageFilePrefsFolderAccess)
            {
                string path = Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + latestVARName + "\\" + folderPath;
                string tagFileName = path + "\\" + resourceFileNameNoExt + "." + resourceFileNameExtension + fileExt;
                if (!FileManagerSecure.DirectoryExists(path) && value) FileManagerSecure.CreateDirectory(path);
                if (FileManagerSecure.DirectoryExists(path))
                {
                    if (value) FileManagerSecure.WriteAllBytes(tagFileName, new byte[] { 0 });
                    else FileManagerSecure.DeleteFile(tagFileName);
                }
                return true;
            }
            return false;
        }

        protected override void Init()
        {
            base.Init();

            resourceFullPathReference = latestVARName + ":\\" + resourceFolderPathFileName;
            resourceFullPathReferenceWithAliasLowerCaps = (resourceFullPathReference+resourceNameAliasJSS.val).ToLower();
            
        }
        public override string resourceThumbnailReference
        {
            get
            {
                if (FileManagerSecure.FileExists(customThumbnailFile)) return customThumbnailFile;
                string thumbailFileRef=latestVARName.ToString() + ":\\" + resourceFolderPathFileNameNoExt + ".jpg";

                if (!FileManagerSecure.FileExists(thumbailFileRef) && resourceType == ResourceType.plugins)
                {
                    string pluginThumnailsVARRef = Utils.GetLatestVARPath(Const.resourceThumbailsVAR + ".latest:\\Saves\\PluginData\\JayJayWon\\BrowserAssist\\VARPluginThumbnails\\" + creatorName + "." + packageName+".jpg");
                    
                    if (FileManagerSecure.FileExists(pluginThumnailsVARRef)) return pluginThumnailsVARRef;
                }

                return thumbailFileRef;
            }
        }

        public PackagedRVGE(VARPackageManifestEntry vpme, string resourceFileName, int _resourceType, bool setNew) : base()
        {
            vpvge = vpme.vpvge;
            resourceType = _resourceType;
            resourceFolderPathFileName = resourceFileName;

            if (setNew) isNew = true;

            highestVARVersion = vpme.varVersion;

            varVersions.Add(vpme.varVersion);

            Init();
        }

        public PackagedRVGE(JSONClass jc, VARPackageVersionGroupEntry _vpvge) : base()
        {
            vpvge = _vpvge;

            if (jc["resourceFullFileName"] != null) resourceFolderPathFileName = jc["resourceFullFileName"];
            if (jc["resourceType"] != null) resourceType = EnumManifest.GetEnumValFromStoreVal(ResourceType.enumManifestName, jc["resourceType"]);

            if (jc["varVersions"] != null)
            {
                JSONArray varVersionsJA = jc["varVersions"].AsArray;
                foreach (JSONNode jn in varVersionsJA)
                {
                    int varVersion = jn.AsInt;
                    if (vpvge.varVersionLicenses.ContainsKey(varVersion)) AddVARVersion(varVersion,false);

                }
            }

            if (jc["varCreatorHidVersions"]!=null)
            {
                JSONArray varVersionsCreatorHidJA = jc["varCreatorHidVersions"].AsArray;
                foreach (JSONNode jn in varVersionsCreatorHidJA) varVersionsCreatorHidden.Add(jn.AsInt);
            }

            if (jc["varCreatorFavVersions"] != null)
            {
                JSONArray varVersionsCreatorFavJA = jc["varCreatorFavVersions"].AsArray;
                foreach (JSONNode jn in varVersionsCreatorFavJA) varVersionsCreatorFav.Add(jn.AsInt);
            }               

            if (jc["varVersionsCSLReferenced"] != null)
            {
                JSONArray varVersionsCSLReferencedJA = jc["varVersionsCSLReferenced"].AsArray;
                foreach (JSONNode jn in varVersionsCSLReferencedJA) varVersionsCSListReferenced.Add(jn.AsInt);
            }           

            Init();
        }

        public void RegisterCoreDataFile(VARResourcesCoreDataFile coreDataFile)
        {
            this.coreDataFile = coreDataFile;
        }
        public void RegisterUserDataFile(VARResourcesUserDataFile userDataFile)
        {
            this.userDataFile = userDataFile;
        }

        public void RestoreVARUserData(JSONClass resourceJC)
        {

            if (resourceJC["baFavourite"] != null) baFavorite = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baFavourite"]);
            if (resourceJC["baHidden"] != null) baHidden = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baHidden"]);
            if (resourceJC["baNew"] != null) _isNew = resourceJC["baNew"].AsBool;

            if (resourceJC["Tags"] != null) RestoreTags(resourceJC["Tags"].AsArray);

            resourceNameAliasJSS.RestoreFromJSON(resourceJC);
        }


        private bool isUserDataDefault { get
            {
                if (baFavorite != TriState.indeterminate) return false;
                if (baHidden != TriState.indeterminate) return false;
                if (isNew) return false;
                if (TagManifest.GetRVGETags(this).Count > 0) return false;

                return true;
            }
        }
        public JSONClass GetVARUserDataJSON()
        {
//            if (isUserDataDefault) return null;

            JSONClass resourceJC = new JSONClass();

            resourceJC["creatorName"] = creatorName;
            resourceJC["packageName"] = packageName;
            resourceJC["resourceFullFileName"] = resourceFolderPathFileName;

            if (baFavorite != TriState.indeterminate)  resourceJC["baFavourite"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baFavorite);
            if (baHidden != TriState.indeterminate) resourceJC["baHidden"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baHidden);
            if (isNew) resourceJC["baNew"].AsBool = isNew;

            if (TagManifest.GetRVGETags(this).Count > 0) resourceJC["Tags"] = GetTagsJSON();

            resourceNameAliasJSS.StoreJSON(resourceJC);

            return resourceJC;
        }

        public override JSONClass GetCoreJSON()
        {
            JSONClass jc = base.GetCoreJSON();
           
            jc["creatorName"] = creatorName;
            jc["packageName"] = packageName;

            JSONArray varVersionsJA = new JSONArray();
            jc["varVersions"] = varVersionsJA;
            foreach (int varVersion in varVersions) varVersionsJA[-1].AsInt = varVersion;

            if (varVersionsCreatorHidden.Count > 0)
            {
                JSONArray varVersionsCreatorHidJA = new JSONArray();
                jc["varCreatorHidVersions"] = varVersionsCreatorHidJA;
                foreach (int varVersion in varVersionsCreatorHidden) varVersionsCreatorHidJA[-1].AsInt = varVersion;
            }
            
            if (varVersionsCreatorFav.Count > 0)
            {
                JSONArray varVersionsCreatorFavJA = new JSONArray();
                jc["varCreatorFavVersions"] = varVersionsCreatorFavJA;
                foreach (int varVersion in varVersionsCreatorFav) varVersionsCreatorFavJA[-1].AsInt = varVersion;
            }
            
            if (varVersionsCSListReferenced.Count > 0)
            {
                JSONArray varVersionsCSLReferencedJA = new JSONArray();
                jc["varVersionsCSLReferenced"] = varVersionsCSLReferencedJA;
                foreach (int varVersion in varVersionsCSListReferenced) varVersionsCSLReferencedJA[-1].AsInt = varVersion;
            }            

            return (jc);
        }
        public void PurgeVARVersion(int varVersion)
        {
            varVersions.Remove(varVersion);
            varVersionsCreatorFav.Remove(varVersion);
            varVersionsCreatorHidden.Remove(varVersion);
            varVersionsUserFav.Remove(varVersion);
            varVersionsUserHidden.Remove(varVersion);
            varVersionsCSListReferenced.Remove(varVersion);

            if (highestVARVersion == varVersion)
            {
                highestVARVersion = -1;
                foreach(int remainingVARVersion in varVersions)
                {
                    if (remainingVARVersion > highestVARVersion) highestVARVersion = remainingVARVersion;
                }
                if (varVersionsCreatorFav.Contains(highestVARVersion)) packageCreatorFav = true;
                else packageCreatorFav = false;
                if (varVersionsCreatorHidden.Contains(highestVARVersion)) packageCreatorHidden = true;
                else packageCreatorHidden = false;

                Init();
            }
        }
        public void AddVARVersion(int varVersion, bool setNew)
        {
            varVersions.Add(varVersion);

            if (varVersion > highestVARVersion)
            {
                highestVARVersion = varVersion;
                if (setNew) isNew = true;
                if (varVersionsCreatorFav.Contains(varVersion)) packageCreatorFav = true;
                else packageCreatorFav = false;
                if (varVersionsCreatorHidden.Contains(varVersion)) packageCreatorHidden = true;
                else packageCreatorHidden = false;
                Init();
            }    
        }

        public void NewVARVersions(HashSet<int> newVARVersions)
        {            
            foreach (int varVersion in newVARVersions.Union(varVersions))
            {
                if (baHidden == TriState.active) SetVAMUserHidden(true);
                if (baFavorite == TriState.active) SetVAMUserFavourite(true);
            }
        }

        public void SetHideCreatorVARVersion(int varVersion)
        {
            varVersionsCreatorHidden.Add(varVersion);
            if (varVersion == highestVARVersion) packageCreatorHidden = true;
        }

        public void SetFavCreatorVARVersion(int varVersion)
        {
            varVersionsCreatorFav.Add(varVersion);
            if (varVersion == highestVARVersion) packageCreatorFav = true;
        }

        public void SetHideUserVARVersion(int varVersion, bool hide)
        {
            if (hide) varVersionsUserHidden.Add(varVersion);
            else varVersionsUserHidden.Remove(varVersion);
            if (varVersion == highestVARVersion)
            {
                if (hide)
                {
                    if (baHidden != TriState.active)
                    {
                        baHidden = TriState.active;
                        vpvge.SetPersistVARGroupUserData();
                    }
                    
                }
                else
                {
                    if (baHidden != TriState.inactive)
                    {
                        baHidden = TriState.inactive;
                        vpvge.SetPersistVARGroupUserData();
                    }
                }
            }
        }
        
        public void SetFavUserVARVersion(int varVersion, bool fav)
        {
            if (fav) varVersionsUserFav.Add(varVersion);
            else varVersionsUserFav.Remove(varVersion);
            if (varVersion == highestVARVersion)
            {
                if (fav)
                {
                    if (baFavorite != TriState.active)
                    {
                        baFavorite = TriState.active;
                        vpvge.SetPersistVARGroupUserData();
                    }
                    
                }
                else
                {
                    if (baFavorite != TriState.inactive)
                    {
                        baFavorite = TriState.inactive;
                        vpvge.SetPersistVARGroupUserData();
                    }
                }
            }
        }

        public void ClearAllUserFilePrefs()
        {
            varVersionsUserHidden.Clear();
            varVersionsUserFav.Clear();
        }

        public void SetCSLFileReferenced(int varVersion)
        {
            varVersionsCSListReferenced.Add(varVersion);
        }
    }

    public class CreatorManifest
    {
        private static Dictionary<string, CreatorManifestEntry> creatorManifestDict = new Dictionary<string, CreatorManifestEntry>();

        public static CreatorManifestEntry RegisterCreator (string creatorName)
        {
            if (creatorManifestDict.ContainsKey(creatorName)) return creatorManifestDict[creatorName];
            CreatorManifestEntry cme = new CreatorManifestEntry(creatorName);
            creatorManifestDict.Add(creatorName,cme);
            return cme;
        }

        public static void RegisterResource (ResourceVersionGroupEntry rvge)
        {
            creatorManifestDict[rvge.creatorName].RegisterResource(rvge);
        }

        public static void RemoveRVGE(ResourceVersionGroupEntry rvge)
        {
            creatorManifestDict[rvge.creatorName].RemoveRVGE(rvge);
        }


        public static HashSet<string> GetAllCreatorNames()
        {
            return new HashSet<string>(creatorManifestDict.Keys);
        }
        public static List<string> GetAllCreatorNamesList()
        {
            return new List<string>(creatorManifestDict.Keys);
        }


        public static void UpdateCreatorChoices(JSONStorableMultiEnumStringChooser creatorJSMultiEnum, HashSet<ResourceVersionGroupEntry> filteredRVGEs)
        {
            HashSet<JSSCDisplayChoice> jSSCDisplayChoices = new HashSet<JSSCDisplayChoice>();

            foreach (var creatorKVP in creatorManifestDict)
            {
                var creator = creatorKVP.Value;
                int rvgeCount = creator.GetResourceCount(filteredRVGEs);
                if (rvgeCount> 0)
                {
                    var jsscDisplayChoice = new JSSCDisplayChoice();
                    jsscDisplayChoice.choice = creator.creatorName;
                    jsscDisplayChoice.displayChoice = creator.creatorName + " (" + rvgeCount.ToString() + ")";
                    jsscDisplayChoice.count = rvgeCount;
                    jSSCDisplayChoices.Add(jsscDisplayChoice);
                }
            }

            List<string> creatorChoices ;
            List<string> creatorDisplayChoices ;
            if (Settings.creatorFilterOrderJSEC.val == CreatorFilterSortOrder.alphabetically)
            {
                creatorChoices = jSSCDisplayChoices.OrderBy(o => o.choice).Select(o=>o.choice).ToList();
                creatorDisplayChoices = jSSCDisplayChoices.OrderBy(o => o.choice).Select(o => o.displayChoice).ToList();

            }
            else
            {
                creatorChoices = jSSCDisplayChoices.OrderByDescending(o => o.count).ThenBy(o => o.choice).Select(o => o.choice).ToList();
                creatorDisplayChoices = jSSCDisplayChoices.OrderByDescending(o => o.count).ThenBy(o=>o.choice).Select(o => o.displayChoice).ToList();
            }

            creatorJSMultiEnum.SetMainChoices(creatorChoices, creatorDisplayChoices);

            if (creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                if (!creatorChoices.Contains(creatorJSMultiEnum.mainVal)) creatorJSMultiEnum.valTopEnum = CreatorFilterChoices.allVARs;
            }
        }
        
    }

    public class JSSCDisplayChoice
    {
        public string choice { get; set; }
        public string displayChoice { get; set; }

        public int count =0;
    }
    public class CreatorManifestEntry
    {
        public string creatorName { get; private set; }

        public HashSet<int> resourceTypes { get; private set; }

        public Dictionary<int, HashSet<ResourceVersionGroupEntry>> resourcesByTypeDict = new Dictionary<int, HashSet<ResourceVersionGroupEntry>>();
        public CreatorManifestEntry(string _creatorName)
        {
            creatorName = _creatorName;
            resourceTypes = new HashSet<int>();
        }


        public int GetResourceCount(HashSet<ResourceVersionGroupEntry> rvges) 
        {
            if (rvges.Count() == 0 || !resourcesByTypeDict.ContainsKey(rvges.First().resourceType)) return 0;

            return rvges.Intersect(resourcesByTypeDict[rvges.First().resourceType]).Count();
        }

        public void RegisterResource (ResourceVersionGroupEntry rvge)
        {
            resourceTypes.Add(rvge.resourceType);

            if (!resourcesByTypeDict.ContainsKey(rvge.resourceType)) resourcesByTypeDict.Add(rvge.resourceType, new HashSet<ResourceVersionGroupEntry>());
            resourcesByTypeDict[rvge.resourceType].Add(rvge);
        }

        public void RemoveRVGE(ResourceVersionGroupEntry rvge)
        {
            resourcesByTypeDict[rvge.resourceType].Remove(rvge);
            if (resourcesByTypeDict[rvge.resourceType].Count ==0) resourceTypes.Remove(rvge.resourceType);
        }
    }
}
;
