﻿
using Battlehub.RTCommon;
using SimpleJSON;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Security.AccessControl;
using static TabbedUIBuilder;


namespace JayJayWon
{
    public class Tag
    {
        protected string _tagName;
        public JSONStorableEnumStringChooser tagCategoryJSEnum;
        protected int _resourceType;

        protected HashSet<CombiTag> _usedByCombiTags = new HashSet<CombiTag>();

        public JSONStorableBool hideTaggedResourcesJSB;


        protected void CopyFrom (Tag tag)
        {
            tagCategoryJSEnum.valNoCallback = tag.tagCategoryJSEnum.val;
            _tagName = tag._tagName;
            _resourceType = tag._resourceType;
            hideTaggedResourcesJSB.valNoCallback = tag.hideTaggedResourcesJSB.val;

            _usedByCombiTags.Clear();
            foreach(CombiTag combiTag in tag._usedByCombiTags)
            {
                _usedByCombiTags.Add(combiTag);
            }
        }
        public string tagName { get
            {
                return _tagName;
            }
        }

        public int tagCategory
        {
            get
            {
                return tagCategoryJSEnum.val;
            }
        }

        public int resourceType
        {
            get
            {
                return _resourceType;
            }
        }

        public void AddUsedBy(CombiTag combiTag)
        {
            _usedByCombiTags.Add(combiTag);
        }

        public void RemoveUsedBy(CombiTag combiTag)
        {
            _usedByCombiTags.Remove(combiTag);
        }

        public HashSet<CombiTag> usedByCombiTagsHieararchy
        {
            get
            {
                HashSet<CombiTag> recursiveUsedByCombiTags = new HashSet<CombiTag>(_usedByCombiTags);

                foreach (var usedCombiTag in _usedByCombiTags)
                {
                    recursiveUsedByCombiTags.UnionWith(usedCombiTag.usedByCombiTagsHieararchy);
                }

                return recursiveUsedByCombiTags;
            }
        }

        public HashSet<CombiTag> usedByCombiTags
        {
            get
            {
                return _usedByCombiTags;
            }
        }


        public Tag(string newTagName, int tagCategory, int newTagResourceType)
        {
            tagCategoryJSEnum = new JSONStorableEnumStringChooser("tagCategory", TagCategory.enumManifestName, tagCategory, "Tag Category");
            hideTaggedResourcesJSB = new JSONStorableBool("hideTaggedResources", false, HideTaggedResourceCallback);

            _tagName = newTagName;
            _resourceType = newTagResourceType;
        }

        public void RenameTag(string newName)
        {
            _tagName = newName;
        }

        protected void HideTaggedResourceCallback(bool hide)
        {
            Settings.SaveSettings();
            BrowserAssist.mainBrowserUI.RefreshFilterResults();
        }
    }

    public class CombiTag : Tag
    {
        public HashSet<Tag> usesTags
        {
            get
            {
                HashSet<Tag> result = new HashSet<Tag>();
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.isTaggableCondition) result.UnionWith(condition.usesTags);
                }
                return result;
            }
        }
        public bool usesNew
        {
            get
            {
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.usesNew) return true;
                }
                return false;
            }
        }

        public bool usesHidden
        {
            get
            {
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.usesHidden) return true;
                }
                return false;
            }
        }

        public bool usesFavorite
        {
            get
            {
                foreach (var condition in _combiLogicConditions)
                {
                    if (condition.usesFavorite) return true;
                }
                return false;
            }
        }
/*
        public HashSet<CombiTag> usesCombiTagsHierarchy
        {
            get
            {
                HashSet<CombiTag> result = new HashSet<CombiTag>();
                foreach(var tag in usesTags)
                {
                    if(tag.tagCategory == TagCategory.combi)
                    {
                        CombiTag combiTag = tag as CombiTag;
                        result.Add(combiTag);
                        result.UnionWith(combiTag.usesCombiTagsHierarchy);
                    }
                }


                return result;
            }
        }
*/
        public JSONStorableEnumStringChooser logicOperandJSEnum;

        protected bool _modified = false;
        public bool isModfied { get { return _modified; }
            set { _modified = value; } }

        protected List<CombiLogicCondition> _combiLogicConditions = new List<CombiLogicCondition>();

        public IEnumerable<CombiLogicCondition> combiLogicConditions { get { return _combiLogicConditions.Select(o => o); } }

        public void AddCombiLogicCondition()
        {
            _combiLogicConditions.Add(new CombiLogicCondition(this));
            _modified=true;
        }

        public void RemoveCombiLogicConditions(CombiLogicCondition clc)
        {
            _combiLogicConditions.Remove(clc);
            _modified=true;
        }

        public CombiTag(string newTagName, int newTagResourceType) : base(newTagName, TagCategory.combi, newTagResourceType)
        {
            logicOperandJSEnum = new JSONStorableEnumStringChooser("logicOperand", LogicOperands.enumManifestName, LogicOperands.and, "",LogicOperandCallback);

            AddCombiLogicCondition();
        }

        public CombiTag(int newTagResourceType, JSONClass jc) : base("", TagCategory.combi, newTagResourceType)
        {
            logicOperandJSEnum = new JSONStorableEnumStringChooser("logicOperand", LogicOperands.enumManifestName, LogicOperands.and, "", LogicOperandCallback);

            _tagName = jc["tagName"];
            logicOperandJSEnum.RestoreFromJSON(jc);

            JSONArray clcJA = jc["combiLogicCondition"].AsArray;

            foreach (JSONClass clcJC  in clcJA)
            {
                _combiLogicConditions.Add(new CombiLogicCondition(this, clcJC));
            }

        }

        public void Init()
        {
            foreach (var clc in _combiLogicConditions)
            {
                clc.Init();
            }
        }

        public JSONClass GetJSON()
        {
            var jc = new JSONClass();
            jc["tagName"] = tagName;
            logicOperandJSEnum.StoreJSON(jc,forceStore:true);

            var clcJA = new JSONArray();

            foreach (var clc in combiLogicConditions)
            {
                clcJA.Add(clc.GetJSON());
            }

            jc["combiLogicCondition"] = clcJA;

            return jc;
        }

        public void LogicOperandCallback(int val)
        {
            isModfied = true;
        }

        public void CopyFrom(CombiTag sourceTag)
        {
            base.CopyFrom(sourceTag);
            logicOperandJSEnum.val = sourceTag.logicOperandJSEnum.val;

            _combiLogicConditions.Clear();
            foreach (var combiLogicCondition in sourceTag.combiLogicConditions)
            {
                _combiLogicConditions.Add(new CombiLogicCondition(this, combiLogicCondition));
            }

            _modified = false; 
        }

        public bool GetState(ResourceVersionGroupEntry rvge)
        {
            int validCLCcount = 0;
            foreach(CombiLogicCondition combiLogicCondition in _combiLogicConditions)
            {
                if (combiLogicCondition.isValidCondition)
                {
                    bool clcState = combiLogicCondition.GetState(rvge);
                    validCLCcount++;
                    if (logicOperandJSEnum.val == LogicOperands.or && clcState) return true;
                    if (logicOperandJSEnum.val == LogicOperands.and && !clcState) return false;
                }


            }

            if (validCLCcount==0) return true;

            return logicOperandJSEnum.val == LogicOperands.and;
        }

        public string description
        {
            get
            {
                string descriptionBuild = "";
                int validCLCcount = 0;
                foreach (CombiLogicCondition combiLogicCondition in _combiLogicConditions)
                {
                    if (combiLogicCondition.isValidCondition)
                    {
                        validCLCcount++;
                        if (validCLCcount > 1)
                        {
                            if (logicOperandJSEnum.val == LogicOperands.or) descriptionBuild += " or";
                            if (logicOperandJSEnum.val == LogicOperands.and) descriptionBuild += " and";
                        }
                        descriptionBuild += combiLogicCondition.description;
                    }
                }

                if (validCLCcount == 0) return "Select Tag Filter conditions...";

                return descriptionBuild;
            }

        }


        public void DeleteTagUsage(Tag tag)
        {
            foreach (var clc in combiLogicConditions)
            {
                clc.DeleteTagUsage(tag);
            }
        }
    }

    public class OrderedHashSet<T> : KeyedCollection<T, T>
    {
        protected override T GetKeyForItem(T item)
        {
            return item;
        }
    }
    public class CombiLogicCondition
    {
        protected HashSet<Tag> _usesTags = new HashSet<Tag>();

        private HashSet<string> usesTagsToInit = new HashSet<string>();

        public ReadOnlyCollection<Tag> usesTags
        {
            get
            {
                return _usesTags.ToList().AsReadOnly();
            }
        }

        public bool usesHidden
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.hidden || logicConditionJSEnum.val == CombiLogicConditionOptions.notHidden);
            }
        }

        public bool usesFavorite
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.favorite || logicConditionJSEnum.val == CombiLogicConditionOptions.notFavorite);
            }
        }

        public bool usesNew
        {
            get
            {
                return (logicConditionJSEnum.val == CombiLogicConditionOptions.New || logicConditionJSEnum.val == CombiLogicConditionOptions.notNew);
            }
        }

        public CombiTag parentCombiTag { get; protected set; }

        public JSONStorableEnumStringChooser logicConditionJSEnum;

        public bool isTaggableCondition { get 
            {
                return !(logicConditionJSEnum.val == CombiLogicConditionOptions.everything || logicConditionJSEnum.val == CombiLogicConditionOptions.notHidden || logicConditionJSEnum.val == CombiLogicConditionOptions.hidden || logicConditionJSEnum.val == CombiLogicConditionOptions.notTagged || logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnything || logicConditionJSEnum.val == CombiLogicConditionOptions.favorite || logicConditionJSEnum.val == CombiLogicConditionOptions.notFavorite || logicConditionJSEnum.val == CombiLogicConditionOptions.New || logicConditionJSEnum.val == CombiLogicConditionOptions.notNew);
            }
        }

        public string GetTagNameList()
        {
            string usesTagsNameList="";

            if (isTaggableCondition)
            {
                foreach (Tag tag in usesTags)
                {
                    string tagName = tag.tagName;
                    if (tag.tagCategory == TagCategory.combi) tagName = "<i>" + tag.tagName + "</i>";
                    if (usesTagsNameList == "") usesTagsNameList = tagName;
                    else usesTagsNameList = usesTagsNameList + ", " + tagName;
                }
            }
            

            return usesTagsNameList;
        }

        public CombiLogicCondition(CombiTag parentCombiTag)
        {
            this.parentCombiTag = parentCombiTag;
            List<int> exlcusions = new List<int>();
            exlcusions.Add(CombiLogicConditionOptions.everything);
            logicConditionJSEnum = new JSONStorableEnumStringChooser("logicCondition", CombiLogicConditionOptions.enumManifestName, CombiLogicConditionOptions.taggedWith, "", LogicConditionCallback, exlcusions);
        }

        public CombiLogicCondition(CombiTag parentCombiTag, JSONClass jc)
        {
            this.parentCombiTag = parentCombiTag;
            List<int> exlcusions = new List<int>();
            exlcusions.Add(CombiLogicConditionOptions.everything);
            logicConditionJSEnum = new JSONStorableEnumStringChooser("logicCondition", CombiLogicConditionOptions.enumManifestName, CombiLogicConditionOptions.taggedWith, "", LogicConditionCallback, exlcusions);


            logicConditionJSEnum.valNoCallback = EnumManifest.GetEnumValFromStoreVal(CombiLogicConditionOptions.enumManifestName, jc["logicCondition"]);

            if (isTaggableCondition)
            {
                var usesTagsJA = jc["usesTags"].AsArray;
                foreach (JSONClass usesTagJC in usesTagsJA)
                {
                    int tagCategory = EnumManifest.GetEnumValFromStoreVal(TagCategory.enumManifestName, usesTagJC["tagCategory"]);
                    string tagName = usesTagJC["tagName"];

                    if (tagCategory != TagCategory.combi && TagManifest.GetTagNameCount(parentCombiTag.resourceType,tagName) > 0)
                    {
                        Tag tag = TagManifest.GetTag(parentCombiTag.resourceType, tagName);
                        UseTag(tag);
                        tag.AddUsedBy(parentCombiTag);
                        
                    }
                    else  usesTagsToInit.Add(tagName);

                }

            }

        }

        public void Init()
        {
            foreach (string usesTagName in usesTagsToInit)
            {
                if (TagManifest.GetTagNameCount(parentCombiTag.resourceType, usesTagName) >0)
                {
                    Tag tag = TagManifest.GetTag(parentCombiTag.resourceType, usesTagName);
                    UseTag(tag);
                    tag.AddUsedBy(parentCombiTag);
                }
                
            }
            usesTagsToInit.Clear();
        }

        public bool isValidCondition
        {
            get
            {
                if (isTaggableCondition && _usesTags.Count == 0) return false;
                return true;
            }
        }

        public string description
        {
            get
            {
                string descriptionBuild="";
                string usesTagsDescription = " (";
                if (isTaggableCondition && _usesTags.Count == 0) return "";
                if (isTaggableCondition && _usesTags.Count > 0)
                {
                    foreach(Tag tag in _usesTags)
                    {
                        if (tag==_usesTags.First()) usesTagsDescription+=tag.tagName;
                        else
                        {
                            usesTagsDescription += ", "+tag.tagName;
                        }
                    }
                    
                }
                usesTagsDescription += ")";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWith)
                    descriptionBuild = " tagged with" + usesTagsDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAll) descriptionBuild = " tagged with all of"+usesTagsDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAny) descriptionBuild = " tagged with any of" + usesTagsDescription;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTagged) descriptionBuild = " not tagged";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTaggedWithAny) descriptionBuild = " not tagged with any of" + usesTagsDescription; ;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnything) descriptionBuild = " tagged with anything"; ;

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.hidden) descriptionBuild = " hidden";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notHidden) descriptionBuild = " not hidden";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.favorite) descriptionBuild = " favorite";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notFavorite) descriptionBuild = " not favorite";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.New) descriptionBuild = " new";

                if (logicConditionJSEnum.val == CombiLogicConditionOptions.notNew) descriptionBuild = " not new";

                return descriptionBuild;
            }
        }

        public bool GetState(ResourceVersionGroupEntry rvge)
        {
            if (isTaggableCondition && _usesTags.Count==0) return false;

            var rvgeTags = new HashSet<Tag>(TagManifest.GetRVGETags(rvge));
            rvgeTags.Remove(parentCombiTag);        

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWith)  return (rvgeTags.Contains(_usesTags.First()));

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAll) return (_usesTags.All(s => rvgeTags.Contains(s)));

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAny) return (_usesTags.Intersect(rvgeTags).Count() >0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTagged) return (rvgeTags.Count() ==0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notTaggedWithAny) return (_usesTags.Intersect(rvgeTags).Count() == 0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.taggedWithAnything) return (rvgeTags.Count() > 0);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.hidden) return (rvge.isHidden || rvge.isTagHidden);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notHidden) return (!rvge.isHidden && !rvge.isTagHidden);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.favorite) return (rvge.isFavorite);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notFavorite) return (!rvge.isFavorite);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.New) return (rvge.isNew);

            if (logicConditionJSEnum.val == CombiLogicConditionOptions.notNew) return (!rvge.isNew);

            return false;
        }

        public JSONClass GetJSON()
        {
            var jc = new JSONClass();
            jc["logicCondition"] = logicConditionJSEnum.storeVal;

            if (isTaggableCondition)
            {
                var usesTagsJA = new JSONArray();
                foreach (Tag usesTag in _usesTags)
                {
                    var usesTagJC = new JSONClass();
                    usesTagJC["tagCategory"] = usesTag.tagCategoryJSEnum.storeVal;
                    usesTagJC["tagName"] = usesTag.tagName;

                    usesTagsJA.Add(usesTagJC);
                }
                jc["usesTags"] = usesTagsJA;

            }
            return jc;
        }



        public CombiLogicCondition(CombiTag parentCombiTag, CombiLogicCondition copyFrom) : this (parentCombiTag)
        {
            logicConditionJSEnum.valNoCallback = copyFrom.logicConditionJSEnum.val;
            foreach (Tag tag in copyFrom._usesTags)
            {
                _usesTags.Add(tag);
            }
        }
        private void LogicConditionCallback(int value)
        {
            parentCombiTag.isModfied = true;
        }

        public void UseOnlyOneTag()
        {
            if (_usesTags.Count > 1)
            {
                Tag keepTag = _usesTags.First();
                _usesTags.Clear();
                UseTag(keepTag);
            }
        }

        public void UseTag(Tag tag, bool replace=false)
        {
            if (replace)
            {
                _usesTags.Clear();
            }
            
            _usesTags.Add(tag);
//            tag.UsedBy(parentCombiTag);
            parentCombiTag.isModfied = true;
        }

        public void RemoveUseTag(Tag tag)
        {
            _usesTags.Remove(tag);
            parentCombiTag.isModfied = true;

//            if (!parentCombiTag.usesTags.Contains(tag)) tag.RemoveUsedBy(parentCombiTag);
        }

        public void DeleteTagUsage(Tag tag)
        {
            _usesTags.Remove(tag);
        }
    }


    public static class TagManifest
    {
        private static HashSet<Tag> tags = new HashSet<Tag>();

        private static Dictionary<ResourceVersionGroupEntry, HashSet<Tag>> rvgeTags = new Dictionary<ResourceVersionGroupEntry, HashSet<Tag>>();
        private static Dictionary<Tag, HashSet<ResourceVersionGroupEntry>> tagsRVGEs = new Dictionary<Tag, HashSet<ResourceVersionGroupEntry>>();

        private static Dictionary<int, Dictionary<string, Tag>> tagNamesToTagByResourceType = new Dictionary<int, Dictionary<string, Tag>>();

        private static Dictionary<int, HashSet<Tag>> categoryTagMap = new Dictionary<int, HashSet<Tag>>();

        private static Dictionary<int, HashSet<Tag>> resourceTypeTagMap = new Dictionary<int, HashSet<Tag>>();

        private static Dictionary<int, HashSet<CombiTag>> resourceTypeCombiTagMap = new Dictionary<int, HashSet<CombiTag>>();

        private static Dictionary<int, Dictionary<int,HashSet<ResourceVersionGroupEntry>>> tagCategoryToResourceByResourceType = new Dictionary<int, Dictionary<int, HashSet<ResourceVersionGroupEntry>>> ();

        static TagManifest()
        {
            foreach (int resourceType in ResourceType.GetResourceTypesByTargetCategory(ResourceTargetCategory.personAtomTargetable))
            {
                CreateTag("male", TagCategory.gender, resourceType);
                CreateTag("female", TagCategory.gender, resourceType);
            }
            CreateTag("session", TagCategory.pluginType, ResourceType.plugins);
            CreateTag("scene", TagCategory.pluginType, ResourceType.plugins);

        }

        public static int GetTaggedResourcesCount(Tag tag)
        {
            return tagsRVGEs[tag].Count();
        }
        public static string NormalizeTagName(string tagName)
        {
            string tagNameMod = tagName.ToLower();
            if (tagName.Length > 0)
            {
                char lastCharacter = tagNameMod.Last();
                if (tagNameMod.Length > 25 || tagNameMod.EndsWith("  ") || (!char.IsLetterOrDigit(lastCharacter) && !tagNameMod.EndsWith(" ")))
                {
                    string tagNameLessOneChar = tagName.Remove(tagName.Length - 1, 1);
                    tagNameMod = tagNameLessOneChar.ToLower();

                }
            }

            return tagNameMod;
        }
/*
        private static HashSet<CombiTag> assesedDependentCombiTabs = new HashSet<CombiTag>();
        public static void AssessCombiTagState(ResourceVersionGroupEntry rvge, CombiTag combiTag)
        {
            
            assesedDependentCombiTabs.Clear();

            bool combiTagState = GetCombiTagState(rvge, combiTag);
            rvge.SetTag(combiTag, combiTagState);
            assesedDependentCombiTabs.Clear();
        }

        private static bool GetCombiTagState(ResourceVersionGroupEntry rvge, CombiTag combiTag)
        {
            var dependentCombiTags = combiTag.usesCombiTagsHierarchy;
            foreach( var dependentCombiTag in dependentCombiTags)
            {
                if (!assesedDependentCombiTabs.Contains(dependentCombiTag))
                {
                    bool dependentCombiTagState = GetCombiTagState(rvge, dependentCombiTag);
                    rvge.SetTag(dependentCombiTag, dependentCombiTagState);
                    assesedDependentCombiTabs.Add(dependentCombiTag);
                }
            }

            return combiTag.GetState(rvge);

        }
*/
        public static bool RenameTag(Tag tag, string newName)
        {
            if (newName != "hide" && newName != "favorite" && !newName.EndsWith(" ") && newName != "" && !GetTags(tag.resourceType).Select(x => x.tagName).Contains(newName))
            {
                
                tagNamesToTagByResourceType[tag.resourceType].Remove(tag.tagName);
                tagNamesToTagByResourceType[tag.resourceType].Add(newName, tag);

                tag.RenameTag(newName);

                if (tagsRVGEs.ContainsKey(tag))
                {
                    foreach (var rvge in tagsRVGEs[tag])
                    {
                        rvge.SetUserDataPersist();
                    }
                }

                if (tag.usedByCombiTags.Count > 0) Settings.SaveSettings();

                return true;
            }
            return false;
        }

        public static int GetTagNameCount(int resourceType, string tagName)
        {
            // TBD Will need to account for Combi tags also

            return tagNamesToTagByResourceType[resourceType].Where(x => x.Key == tagName).ToList().Count();
        }

        public static CombiTag CreateCombiTag(string name, int resourceType)
        {
            if (tags.Any(t => t.tagName == name && t.resourceType == resourceType)) return null;

            CombiTag newTag = new CombiTag(name, resourceType);
            RegisterNewTag(newTag);

            if (!resourceTypeCombiTagMap.ContainsKey(resourceType)) resourceTypeCombiTagMap.Add(resourceType, new HashSet<CombiTag>());
            resourceTypeCombiTagMap[resourceType].Add(newTag);

            return newTag;
        }

        public static JSONArray GetCombiTagsJSON(int resourceType)
        {
            var combiTagsJA = new JSONArray();

            if (resourceTypeCombiTagMap.ContainsKey(resourceType))
            {
                foreach (var combiTag in resourceTypeCombiTagMap[resourceType])
                {
                    var combiTagJC = combiTag.GetJSON();
                    combiTagsJA.Add(combiTagJC);
                }
            }
            return combiTagsJA;
        }
        public static JSONArray GetHiddenTagsJSON(int resourceType)
        {
            var hiddenTagsJA = new JSONArray();

            if (resourceTypeTagMap.ContainsKey(resourceType))
            {
                foreach (var tag in resourceTypeTagMap[resourceType])
                {
                    if (tag.hideTaggedResourcesJSB.val) hiddenTagsJA.Add(tag.tagName);

                }
            }
            return hiddenTagsJA;
        }

        public static void RestoreHiddenTagsJSON(int resourceType, JSONArray tagsJA)
        {
            foreach (JSONNode node in tagsJA)
            {
                if (tagNamesToTagByResourceType.ContainsKey(resourceType) && tagNamesToTagByResourceType[resourceType].ContainsKey(node))
                {
                    var tag = GetTag(resourceType, node);
                    tag.hideTaggedResourcesJSB.valNoCallback = true;
                }
                
            }
        }

        public static void RestoreCombiTagsJSON(int resourceType, JSONArray combiTagsJA)
        {
            foreach (JSONClass jc in combiTagsJA)
            {
                CombiTag newTag = new CombiTag(resourceType, jc);
                RegisterNewTag(newTag);

                if (!resourceTypeCombiTagMap.ContainsKey(resourceType)) resourceTypeCombiTagMap.Add(resourceType, new HashSet<CombiTag>());
                resourceTypeCombiTagMap[resourceType].Add(newTag);
            }
            if (resourceTypeCombiTagMap.ContainsKey(resourceType))
            {
                foreach (CombiTag combiTag in resourceTypeCombiTagMap[resourceType])
                {
                    combiTag.Init();
                }
            }

        }



        public static void DeleteTag(Tag tag)
        {
            if (tag.tagCategory == TagCategory.userDefined || tag.tagCategory == TagCategory.combi)
            {
                if (tagsRVGEs.ContainsKey(tag))
                {
                    var taggedRVGEs = new HashSet<ResourceVersionGroupEntry> (tagsRVGEs[tag]);
                    foreach (var rvge in taggedRVGEs)
                    {
                        rvge.SetTag(tag, false);
                        
                    }
                }

                if (tag.usedByCombiTags.Count > 0)
                {
                    foreach (var combiTag in tag.usedByCombiTags)
                    {
                        combiTag.DeleteTagUsage(tag);
                    }
                    Settings.SaveSettings();
                }


                categoryTagMap[tag.tagCategory].Remove(tag);
                resourceTypeTagMap[tag.resourceType].Remove(tag);
                if (tag.tagCategory == TagCategory.combi) resourceTypeCombiTagMap[tag.resourceType].Remove(tag as CombiTag);
                var tagNamesToTag = tagNamesToTagByResourceType[tag.resourceType];
                tagNamesToTag.Remove(tag.tagName);
                tags.Remove(tag);

            }
            else SuperController.LogError("BA.TagManifest.DeleteTag: Attempt to delete tag that is not a user defined tag.");
        }
        public static void DeleteTag(string tagName, int resourceType)
        {
            Tag tag = GetTag(resourceType, tagName);

            DeleteTag(tag);
        }

        private static void RegisterNewTag(Tag newTag)
        {
            tags.Add(newTag);
            tagsRVGEs.Add(newTag, new HashSet<ResourceVersionGroupEntry> ());

            if (!categoryTagMap.ContainsKey(newTag.tagCategory)) categoryTagMap.Add(newTag.tagCategory, new HashSet<Tag>());
            categoryTagMap[newTag.tagCategory].Add(newTag);

            if (!resourceTypeTagMap.ContainsKey(newTag.resourceType)) resourceTypeTagMap.Add(newTag.resourceType, new HashSet<Tag>());
            resourceTypeTagMap[newTag.resourceType].Add(newTag);

            if (!tagCategoryToResourceByResourceType.ContainsKey(newTag.resourceType)) tagCategoryToResourceByResourceType.Add(newTag.resourceType, new Dictionary<int, HashSet<ResourceVersionGroupEntry>>());
            var tagCategoryToResource = tagCategoryToResourceByResourceType[newTag.resourceType];
            if (!tagCategoryToResource.ContainsKey(newTag.tagCategory)) tagCategoryToResource.Add(newTag.tagCategory, new HashSet<ResourceVersionGroupEntry>());

            if (!tagNamesToTagByResourceType.ContainsKey(newTag.resourceType)) tagNamesToTagByResourceType.Add(newTag.resourceType, new Dictionary<string, Tag>());

            var tagNamesToTag = tagNamesToTagByResourceType[newTag.resourceType];
            if (!tagNamesToTag.ContainsKey(newTag.tagName)) tagNamesToTag.Add(newTag.tagName, newTag);
            else tagNamesToTag[newTag.tagName] = newTag;
        }

        public static Tag CreateTag(string name, int tagCategory, int resourceType)
        {
            if (tags.Any(t => t.tagName == name && t.resourceType == resourceType)) return null;

            Tag newTag = new Tag(name, tagCategory, resourceType);

            RegisterNewTag(newTag);
            return newTag;
        }

        public static HashSet<Tag> GetRVGETags(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return new HashSet<Tag>();
            return rvgeTags[rvge];
        }

        public static List<Tag> GetTags(int resourceType)
        {
            if (!resourceTypeTagMap.ContainsKey(resourceType)) resourceTypeTagMap.Add(resourceType, new HashSet<Tag>());
            return resourceTypeTagMap[resourceType].OrderBy(t=>t.tagCategory).ThenBy(t=>t.tagName).ToList();
        }

        public static HashSet<CombiTag> GetCombiTags(int resourceType)
        {
            if (!resourceTypeCombiTagMap.ContainsKey(resourceType)) resourceTypeCombiTagMap.Add(resourceType, new HashSet<CombiTag>());
            return resourceTypeCombiTagMap[resourceType];
        }
        public static List<Tag> GetTags(int resourceType, int tagCategory)
        {
            if (!resourceTypeTagMap.ContainsKey(resourceType)) resourceTypeTagMap.Add(resourceType, new HashSet<Tag>());
            return resourceTypeTagMap[resourceType].Where(x => x.tagCategory==tagCategory).OrderBy(t => t.tagName).ToList();
        }


        public static List<string> GetTagNames(int resourceType)
        {
            if (!tagNamesToTagByResourceType.ContainsKey(resourceType)) tagNamesToTagByResourceType.Add(resourceType, new Dictionary<string, Tag>());
            List<string> tagNames = new List<string>(tagNamesToTagByResourceType[resourceType].Keys);
            tagNames.Sort();
            return tagNames;
        }

        public static HashSet<int> GetTagCategories(int resourceType)
        {
            if (!tagCategoryToResourceByResourceType.ContainsKey(resourceType)) tagCategoryToResourceByResourceType.Add(resourceType, new Dictionary<int, HashSet<ResourceVersionGroupEntry>>());

            var tagCategories = new HashSet<int>();

            foreach (var kvp in tagCategoryToResourceByResourceType[resourceType])
            {
                if (kvp.Value.Count>0) tagCategories.Add(kvp.Key);
            }

            return tagCategories;
        }

        public static Tag GetTag(int resourceType, string tagName)
        {
            return tagNamesToTagByResourceType[resourceType][tagName];
        }

        public static bool IsRVGETagged(ResourceVersionGroupEntry rvge, Tag tag)
        {
            if (!rvgeTags.ContainsKey(rvge)) return false;

            if (rvgeTags[rvge].Contains(tag)) return true;

            return false;
        }

        public static bool IsRVGENotTagged(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return true;

            if (rvgeTags[rvge].Count==0) return true;

            return false;
        }

        public static bool IsRVGENotUserTagged(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return true;

            if (rvgeTags[rvge].Where(t=>t.tagCategory==TagCategory.userDefined).Count() == 0) return true;

            return false;
        }


        public static void SetRVGETagged(ResourceVersionGroupEntry rvge, Tag tag, bool state)
        {           

            if (state)
            {
                if (!rvgeTags.ContainsKey(rvge)) rvgeTags.Add(rvge, new HashSet<Tag>());
                rvgeTags[rvge].Add(tag);
                if (!tagsRVGEs.ContainsKey(tag)) tagsRVGEs.Add(tag, new HashSet<ResourceVersionGroupEntry>());
                tagsRVGEs[tag].Add(rvge);


                tagCategoryToResourceByResourceType[tag.resourceType][tag.tagCategory].Add(rvge);
            }
            else
            {
                if (!rvgeTags.ContainsKey(rvge)) return;
                rvgeTags[rvge].Remove(tag);
                tagsRVGEs[tag].Remove(rvge);

                tagCategoryToResourceByResourceType[tag.resourceType][tag.tagCategory].Remove(rvge);
            }
        }

        public static void SetRVGETagged(ResourceVersionGroupEntry rvge, string tagName, bool state)
        {
            if (!tagNamesToTagByResourceType[rvge.resourceType].ContainsKey(tagName)) CreateTag(tagName,TagCategory.userDefined,rvge.resourceType);
            Tag tag = tagNamesToTagByResourceType[rvge.resourceType][tagName];
            SetRVGETagged(rvge, tag, state);
        }

    }
}
