﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using Leap.Unity.Query;

namespace JayJayWon
{

    public class BrowserUI
    {
        public static Color darkBlueGrey = new Color(0.4f, 0.4f, 0.6f, 1f);
        public static Color lightBlueGrey = new Color(0.5f, 0.5f, 0.7f, 1f);
        public static Color lightGrey = new Color(0.9f, 0.9f, 0.9f, 1f);
        public static Color mediumGrey = new Color(0.5f, 0.5f, 0.5f, 1f);
        public static Color darkGrey = new Color(0.2f, 0.2f, 0.2f, 1f);

        private Image fileBrowserImage;
        private Image hubBrowserImage;
        private CanvasScaler mainCanvasScalar;
        public BrowserUI()
        {
            GameObject canvasObject = new GameObject();
            mainCanvas = canvasObject.AddComponent<Canvas>();
            mainCanvas.pixelPerfect = false;

            if (isVREnabled) mainCanvas.renderMode = RenderMode.WorldSpace;
            else mainCanvas.renderMode = RenderMode.ScreenSpaceOverlay;

            mainCanvasScalar = canvasObject.AddComponent<CanvasScaler>();
            GraphicRaycaster rayCaster = canvasObject.AddComponent<GraphicRaycaster>();

            mainCanvas.sortingOrder = -3;

            SuperController.singleton.AddCanvas(mainCanvas);


            if (isVREnabled)
            {
                mainCanvasScalar.dynamicPixelsPerUnit = 4f;

                RefreshVRScale();
            }
            else
            {
                mainCanvasScalar.dynamicPixelsPerUnit = 2f;
                mainCanvasScalar.uiScaleMode = CanvasScaler.ScaleMode.ScaleWithScreenSize;
                RefreshDesktopScale();

            }
            browserFiltersUI = new BrowserFiltersUI(this);
            resourceSelectionsUI = new ResourceSelections(this);


            helpTextJSS = new JSONStorableString("helpText", "");

            InitBrowserUI();

            var fileBrowserTransform = SuperController.singleton.worldUI.Find("Canvas/FileBrowserWorld/MainPanel");
            var hubBrowserTransform = SuperController.singleton.worldUI.Find("Canvas/HubBrowsePanel");
            fileBrowserImage = fileBrowserTransform.GetComponent<Image>();
            hubBrowserImage = hubBrowserTransform.GetComponent<Image>();

            SetInactive();
        }

        public float worldScaleAdjust
        {
            get
            {
                if (isVREnabled) return SuperController.singleton.worldScale;
                else return 1f;
            }
        }

        public void SetLocalScale(bool attachedToVAMUI)
        {            
            if (attachedToVAMUI && isVREnabled)
            {
                mainCanvas.transform.localScale = new Vector3(Const._localCanvasScale * .6f * worldScaleAdjust, Const._localCanvasScale * .6f * worldScaleAdjust, Const._localCanvasScale * .6f * worldScaleAdjust);
            }
            else mainCanvas.transform.localScale = new Vector3(Const._localCanvasScale * worldScaleAdjust, Const._localCanvasScale * worldScaleAdjust, Const._localCanvasScale * worldScaleAdjust);
        }
        public void RefreshDesktopScale()
        {
            if (!isVREnabled)  mainCanvasScalar.referenceResolution = new Vector2(1440 / Settings.uiScaleJSF.val, 900 / Settings.uiScaleJSF.val);
        }

        private void RefreshVRScale()
        {
            SetLocalScale(attachedVAMUI);

        }


        const int filterBoxHeight = 50;
        const int boxSpacerHeight = 3;
        public static int edgeSpacerSize = 2;
        const int infoTextBoxHeight = 22;
        public const int resourceActionBoxHeight = 90;



        private int screenHeight { get { if (isVREnabled) return 600;
                else return Screen.height; } }

        private int resourceBrowserBottomOffset { get
            {
                return edgeSpacerSize + infoTextBoxHeight + resourceActionBoxHeight + resourceSelectionsUI.resourceSelectionBoxHeight + (boxSpacerHeight * 3);
            }
        }
        private int resourceBrowserTopOffset
        {
            get
            {
                return edgeSpacerSize + filterBoxHeight + boxSpacerHeight;
            }
        }

        private int resourceBrowserHeight { get
            {
                return screenHeight - resourceBrowserBottomOffset - resourceBrowserTopOffset;
            } }

        public int selectedResourcesVOffset
        {
            get
            {
                return edgeSpacerSize + infoTextBoxHeight + resourceActionBoxHeight + (boxSpacerHeight * 2);
            }
        }

        public float browserScreenWidth { get
            {
                if (isVREnabled) return 650f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked || currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return 650f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return 1440 / Settings.uiScaleJSF.val*0.9f;
                    return -1f;
            } }

        public int backgroundBoxAnchorPreset { get
            {
                if (isVREnabled) return AnchorPresets.centre;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked) return AnchorPresets.vStretchRight;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return AnchorPresets.vStretchMiddle;
                if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return AnchorPresets.vStretchLeft;
                return -1;
            } }

        private void InitBrowserUI()
        {
            float vSize = 0f;
            if (isVREnabled) vSize = screenHeight;

            backgroundBoxGO = UI.AddChildImage(parentTransform: mainCanvas.transform,
                color: new Color(0.5f, 0.5f, 0.5f, 0.95f),
                anchorPreset: backgroundBoxAnchorPreset,
                horizontalSize: browserScreenWidth,
                verticalSize: vSize,
                anchoredPositionOffset: new Vector2(-5f, 0f));
           
            resourceBrowserGO = UI.CreateVScrollableContent(backgroundBoxGO, new Color(0.8f, 0.8f, 0.8f, 0.8f), AnchorPresets.vStretchMiddle, browserScreenWidth-10f, resourceBrowserHeight, new Vector2(0f, -filterBoxHeight- edgeSpacerSize- boxSpacerHeight));


           selectedResourceGO = UI.AddChildGOImage(parentGO:backgroundBoxGO,
                color: darkBlueGrey,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: resourceSelectionsUI.resourceSelectionBoxWidth,
                verticalSize: resourceSelectionsUI.resourceSelectionBoxHeight,
                anchoredPositionOffset: new Vector2((resourceSelectionsUI.resourceSelectionBoxWidth -640f)/2, selectedResourcesVOffset));

            resourceActionsGO = UI.AddChildGOImage(parentGO: backgroundBoxGO,
                color: darkBlueGrey,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: 640f,
                verticalSize: resourceActionBoxHeight,
                anchoredPositionOffset: new Vector2(0f, infoTextBoxHeight+edgeSpacerSize+boxSpacerHeight));

            filtersBoxGO = UI.AddChildGOImage(parentGO: backgroundBoxGO,
                color: darkBlueGrey,
                anchorPreset: AnchorPresets.topMiddle,
                horizontalSize: 640f,
                verticalSize: filterBoxHeight,
                anchoredPositionOffset: new Vector2(0f, -edgeSpacerSize));

            cancelButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b>X</b>", 28, (650f / 2) - 15f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle) ;
            cancelButtonUID.buttonImage.sprite = null;
            cancelButtonUID.button.onClick.AddListener(delegate {
                SetInactive();
            });
            ApplyHelp(cancelButtonUID, "Close the Browser Assist window.");

            if (isVREnabled)
            {
                repositionUI = UI.CreateUIButton(backgroundBoxGO, 25f, 25f, "<b>O</b>", 34, 1, 2, anchorPreset: AnchorPresets.bottomLeft);
                repositionUI.button.onClick.AddListener(delegate {                   
                    if (Settings.dockToVAMUIJSB.val && Globals.isVREnabled)
                    {
                        attachedVAMUI = false;
                        SuperController.singleton.HideMainHUD();
                        SetActive();
                    }
                    });
                var component = repositionUI.button.gameObject.AddComponent<ButtonPointerDrag>();
                component.Init(mainCanvas);
            }
            else
            {
                leftExpandContractButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b>|<</b>", 28, -(640f / 2) + 10f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle);
                leftExpandContractButtonUID.buttonImage.sprite = null;
                leftExpandContractButtonUID.button.onClick.AddListener(delegate {
                    if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked)
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.fullWidth;
                        leftExpandContractButtonUID.label = "<b>|<</b>";
                        ApplyHelp(leftExpandContractButtonUID, "Dock the Browser Assist window on the left of the screen.");
                    }
                    else
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.leftDocked;
                        leftExpandContractButtonUID.label = "<b>>></b>";
                        
                        ApplyHelp(leftExpandContractButtonUID, "Expand the Browser Assist window across the screen width.");
                    }
                    rightExpandContractButtonUID.label = "<b>>|</b>";
                    ApplyHelp(rightExpandContractButtonUID, "Dock the Browser Assist window on the right of the screen.");
                    RefreshBrowserUIDimensions();
                    RepositionBrowserBottomUIElements();
                    browserFiltersUI.SetFolderBrowserInActive();
                });
                rightExpandContractButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b><<</b>", 28, -(640f / 2) + 32f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle);
                rightExpandContractButtonUID.buttonImage.sprite = null;
                rightExpandContractButtonUID.button.onClick.AddListener(delegate {
                    if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked)
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.fullWidth;
                        
                        rightExpandContractButtonUID.label = "<b>>|</b>";
                        ApplyHelp(rightExpandContractButtonUID, "Dock the Browser Assist window on the right of the screen.");

                    }
                    else
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.rightDocked;
                        rightExpandContractButtonUID.label = "<b><<</b>";
                        ApplyHelp(rightExpandContractButtonUID, "Expand the Browser Assist window across the screen width.");

                    }
                    leftExpandContractButtonUID.label = "<b>|<</b>";
                    ApplyHelp(leftExpandContractButtonUID, "Dock the Browser Assist window on the left of the screen.");
                    RefreshBrowserUIDimensions();
                    RepositionBrowserBottomUIElements();
                    browserFiltersUI.SetFolderBrowserInActive();
                });
            }
            infoTextBoxGO = UI.CreateBasicTextBox(backgroundBoxGO, helpTextJSS, 574f, infoTextBoxHeight, 22, 12f, edgeSpacerSize, AnchorPresets.bottomMiddle, true, 18);

            browserFiltersUI.InitUI(filtersBoxGO);
            resourceSelectionsUI.InitUI(selectedResourceGO);
            resourceActionsUI = new ResourceActionsUI(this,resourceActionsGO);
            resourceActionsUI.RefreshUI();

            resourceBrowserUI = new ResourceBrowserUI(this, resourceBrowserGO);
        }

        private void RepositionBrowserBottomUIElements()
        {
            if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked)
            {
                UI.ResetAnchorPosition(rightExpandContractButtonUID.gameObject, (640f / 2) - 10f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(leftExpandContractButtonUID.gameObject, (640f / 2) - 32f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(cancelButtonUID.gameObject, -(650f / 2) + 15f, edgeSpacerSize + 1f);

                UI.ResetAnchorPosition(infoTextBoxGO, -12f, edgeSpacerSize);
            }
            else
            {
                UI.ResetAnchorPosition(rightExpandContractButtonUID.gameObject, -(640f / 2) +32f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(leftExpandContractButtonUID.gameObject, -(640f / 2) + 10f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(cancelButtonUID.gameObject, (650f / 2) - 15f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(infoTextBoxGO, 12f, edgeSpacerSize);
            }
        }
        private void RefreshBrowserUIDimensions()
        {
            UI.ResizeWidth(backgroundBoxGO, browserScreenWidth);
            UI.ResetAnchorType(backgroundBoxGO, backgroundBoxAnchorPreset);
            UI.ResizeWidth(resourceBrowserGO, browserScreenWidth - 10f);
            GameObject buttonListViewportGO = resourceBrowserGO.transform.Find("scrollableContentViewportGO").gameObject;
            UI.ResizeWidth(buttonListViewportGO, browserScreenWidth - 20f);
            resourceBrowserUI.RepopulateButtons();
        }
        public void RefreshBrowserUILayout()
        {       
            UI.ResizeHeight(selectedResourceGO, resourceSelectionsUI.resourceSelectionBoxHeight);
            UI.ResizeWidth(selectedResourceGO, resourceSelectionsUI.resourceSelectionBoxWidth);
            UI.ResetAnchorPosition(selectedResourceGO, (resourceSelectionsUI.resourceSelectionBoxWidth - 640f) / 2, selectedResourcesVOffset);
            resourceBrowserUI.RepositionNavButtons();

            UI.ResetBottomOffset(resourceBrowserGO, (edgeSpacerSize * 1) + infoTextBoxHeight + resourceActionBoxHeight + resourceSelectionsUI.resourceSelectionBoxHeight + (boxSpacerHeight * 3));
        }
        public bool isVisible { get { return mainCanvas.enabled; } set { mainCanvas.enabled = value; } }

        public bool isActive { get; private set; }


        private bool _attachedVAMUI = false;
        public bool attachedVAMUI { get { return _attachedVAMUI; } set {
                _attachedVAMUI = value;
                SetLocalScale(value);
                
            } }    
        public void ToggleActive()
        {
            if (mainCanvas.enabled) SetInactive();
            else SetActive();
        }
        public void SetActive()
        {
            RefreshFilterResults();

            bool mainHUDActive = SuperController.singleton.mainHUD.gameObject.activeSelf;

            if (!mainHUDActive || !Settings.dockToVAMUIJSB.val || !isVREnabled) attachedVAMUI = false;
            else attachedVAMUI = true;

            if (isVREnabled )
            {
                if (!attachedVAMUI)
                {
                    Vector3 headPosition = SuperController.singleton.lookCamera.transform.position;
                    if (SuperController.singleton.mainHUD == null || !SuperController.singleton.mainHUD.gameObject.activeSelf)
                    {

                        Quaternion headFacingDirection = SuperController.singleton.lookCamera.transform.rotation;
                        Vector3 offsetPosition = new Vector3(0f, 0f, 2f*worldScaleAdjust);

                        Vector3 newPosition = headPosition + headFacingDirection * offsetPosition;
                        mainCanvas.transform.position = new Vector3(newPosition.x, headPosition.y, newPosition.z);
                    }
                    else
                    {

                        Vector3 newTargetPos = SuperController.singleton.mainHUD.position + (SuperController.singleton.navigationRig.up * 1f) - (SuperController.singleton.mainHUD.forward * 0.6f);

                        Vector3 newPos = new Vector3(newTargetPos.x, headPosition.y, newTargetPos.z);
                        mainCanvas.transform.position = newPos;

                        // SuperController.singleton.activeUI = SuperController.ActiveUI.None;

                        SuperController.singleton.HideMainHUD();
                    }
                }
                else
                {
                    SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                }

            }
            else resourceBrowserUI.NudgeScrollBar();
            isVisible = true;
            isActive = true;
        }
        public void SetInactive()
        {
            isVisible = false;
            isActive = false;
            //            if (isVREnabled && SuperController.singleton.activeUI==SuperController.ActiveUI.None) SuperController.singleton.activeUI = SuperController.singleton.lastActiveUI;

        }
        public void RefreshButtonLayout(bool forceSelectedResourceRefresh=false)
        {
 //           currentFilterResultsPage = 1;
            if (resourceBrowserUI!=null) resourceBrowserUI.RepopulateButtons(forceSelectedResourceRefresh);
        }

        public void RefreshFilterResults(bool forceSelectedResourceRefresh = false)
        {
            currentFilterResults = ResourceManifest.GetResourceGroups(browserFiltersUI.currentFilterSettings);
            randomCurrentFilterResults.Clear();
            RefreshButtonLayout(forceSelectedResourceRefresh);
        }

        public void Update()
        {
            resourceBrowserUI.Update();

            if (isVREnabled)
            {
                RefreshVRScale();

                if (Settings.dockToVAMUIJSB.val)
                {
                    bool mainHUDActive = SuperController.singleton.mainHUD.gameObject.activeSelf;

                    if (mainHUDActive && isActive)
                    {
                        if (SuperController.singleton.activeUI == SuperController.ActiveUI.None)
                        {
                            attachedVAMUI = true;
                            isVisible = true;
                            Transform mainHUDTransform = SuperController.singleton.mainHUD.transform;

                            Vector3 mainHUDRotationEuler = SuperController.singleton.mainHUD.transform.rotation.eulerAngles;
                            mainCanvas.transform.rotation = Quaternion.Euler(mainHUDRotationEuler.x, mainHUDRotationEuler.y, mainHUDRotationEuler.z);
                            mainCanvas.transform.Rotate(0f, -180f, 0f);

                            mainCanvas.transform.position = mainHUDTransform.position + (mainHUDTransform.up * 0.55f*worldScaleAdjust) + (mainHUDTransform.forward * -0f);
                        }
                        else
                        {
                            SetInactive();
                        }

                    }
                    else if (attachedVAMUI)
                    {
                        isVisible = false;
                    }
                }

                if (!Settings.dockToVAMUIJSB.val || !attachedVAMUI)
                {
                    Vector3 headPosition = SuperController.singleton.lookCamera.transform.position;
                    mainCanvas.transform.LookAt(headPosition, SuperController.singleton.navigationRig.up);
                    mainCanvas.transform.rotation *= Quaternion.Euler(0f, 180f, 0f);
                    
                }
                    
            }

            if (SuperController.singleton.screenshotCamera.enabled || SuperController.singleton.hiResScreenshotCamera.enabled)
            {
                SetInactive();
            }

            if (fileBrowserImage.gameObject.activeInHierarchy || hubBrowserImage.gameObject.activeInHierarchy)
            {
                if (!isHidden && mainCanvas!=null && mainCanvas.enabled)
                {
                    SetInactive();
                    isHidden = true;
                }
                
            }
            else
            {
                if (isHidden)
                {
                    SetActive();
                    isHidden = false;
                }
            }
        }

        public void OnDestroy()
        {
            if (SuperController.singleton != null)  SuperController.singleton.RemoveCanvas(mainCanvas);
            if (mainCanvas != null) mainCanvas.transform.SetParent(null, false);

            if (mainCanvas.gameObject != null) GameObject.Destroy(mainCanvas.gameObject);

            if (resourceBrowserUI!=null) resourceBrowserUI.OnDestory();
        }

        public void ApplyHelp(UIDynamicSlider sliderUID, string helpText)
        {
            ApplyHelp(sliderUID.slider.gameObject, helpText);
            ApplyHelp(sliderUID.sliderControl.gameObject, helpText);
            ApplyHelp(sliderUID.rangeAdjustGroup.gameObject, helpText);

            var inputTrans = sliderUID.transform.Find("ValueInputField");
            if (inputTrans != null) ApplyHelp(inputTrans.gameObject, helpText);
        }

        public void ApplyHelp(UIDynamicPopup popupUID, string helpText)
        {
            ApplyHelp(popupUID.popup.gameObject, helpText);
            ApplyHelp(popupUID.popup.popupPanel.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicToggle toggleUID, string helpText)
        {
            ApplyHelp(toggleUID.toggle.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicTextField textFieldUID, string helpText)
        {
            GameObject gameObject = textFieldUID.gameObject;
            if (textFieldUID.transform.GetComponent<InputField>() != null)
            {
                gameObject = textFieldUID.transform.GetComponent<InputField>().gameObject;
            }
            ApplyHelp(gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicButton buttonUID, string helpText, bool warning = false)
        {
            ApplyHelp(buttonUID.button.gameObject, helpText, warning);
        }

        public void ClearHelp()
        {
            helpTextJSS.val = "";
            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.gray;
        }

        public void ApplyHelp(GameObject gameObject, string helpText, bool warning = false)
        {
            GeneralPointerEnterExitHelp contextHelp = gameObject.AddComponent<GeneralPointerEnterExitHelp>();
            contextHelp.Init(helpText, helpTextJSS, warning);
        }

        public void ApplyTagHelp(UIDynamicToggle toggleUID, string helpText)
        {
            ApplyHelp(toggleUID.toggle.gameObject, helpText);
        }

        public ResourceVersionGroupEntry GetNextRandomRVGEfromFilterResults()
        {
            if (randomCurrentFilterResults.Count == 0)
            {
                randomCurrentFilterResults = currentFilterResults.OrderBy(item => UnityEngine.Random.Range(0,1f)).ToList();
            }
            if (randomCurrentFilterResults.Count > 0)
            {
                ResourceVersionGroupEntry nextRVGE=null;
                if (randomCurrentFilterResults.Count ==1) nextRVGE = randomCurrentFilterResults.First();
                else
                {
                    while (nextRVGE==null || nextRVGE == lastRandomRVGE)
                    {
                        nextRVGE = randomCurrentFilterResults.First();
                    }
                }
                lastRandomRVGE = nextRVGE;
                randomCurrentFilterResults.Remove(nextRVGE);
                return nextRVGE;
            }
            return null;
        }

        public void SaveResourceScrollBarPosition()
        {

        }

        public int currentDesktopDisplayMode { get; private set; } = DesktopDisplayMode.rightDocked;

        private GameObject backgroundBoxGO;
        private GameObject filtersBoxGO;
        private GameObject resourceBrowserGO;
        private GameObject selectedResourceGO;
        private GameObject resourceActionsGO;

        private UIDynamicButton repositionUI;
        private UIDynamicButton leftExpandContractButtonUID;
        private UIDynamicButton rightExpandContractButtonUID;
        private UIDynamicButton cancelButtonUID;
        private GameObject infoTextBoxGO;

        public JSONStorableString helpTextJSS;
 

        public BrowserFiltersUI browserFiltersUI;
        public ResourceActionsUI resourceActionsUI;
        public ResourceBrowserUI resourceBrowserUI;

        public List<ResourceVersionGroupEntry> currentFilterResults;
        private List<ResourceVersionGroupEntry> randomCurrentFilterResults = new List<ResourceVersionGroupEntry>();
        private ResourceVersionGroupEntry lastRandomRVGE = null;

        public int currentFilterResultsPage
        {
            get
            {
                int currentResourceType = browserFiltersUI.resourceType;
                if (!filterResultsPageByResourceType.ContainsKey(currentResourceType)) filterResultsPageByResourceType.Add(currentResourceType, 1);
                return filterResultsPageByResourceType[currentResourceType];
            }
            set
            {
                int currentResourceType = browserFiltersUI.resourceType;
                if (!filterResultsPageByResourceType.ContainsKey(currentResourceType)) filterResultsPageByResourceType.Add(currentResourceType, 1);
                filterResultsPageByResourceType[currentResourceType] = value;
            }
        }
        private Dictionary<int,int> filterResultsPageByResourceType = new Dictionary<int,int>();

        public ResourceSelections resourceSelectionsUI ;

        public Canvas mainCanvas;

        public bool isVREnabled { get { return Globals.isVREnabled; } }

        private bool isHidden = false;
    }

    public class ResourceSelections
    {
        private Dictionary<int, ResourceTypeSelections> resourceTypeSelection = new Dictionary<int, ResourceTypeSelections>();

        private GameObject resourceSelectionsScrollableBoxGO;
        private GameObject resourceSelectionsBoxGO;
        private GameObject resourceSelectionScrollableViewportContentGO;

        private BrowserFiltersUI browserFiltersUI;
        private BrowserUI browserUI;

        private List<GameObject> resourceSelectionRows = new List<GameObject>();

        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> selectedRVGEButtons = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        const int resourceSelectionBoxHeightMin = 20;
        const int resourceSelectionBoxHeightMax = 55;
        const int resourceSelectionBoxWidthMin = 441;
        const int resourceSelectionBoxWidthMax = 530;

        public ResourceTypeSelections GetResourceTypeSelections(int resourceType)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) return resourceTypeSelection[resourceType];
            return null;
        }

        public int resourceSelectionBoxHeight
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return resourceSelectionBoxHeightMin;
                return resourceSelectionBoxHeightMax;
            }
        }
        public int resourceSelectionBoxWidth
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return resourceSelectionBoxWidthMin;
                return resourceSelectionBoxWidthMax;
            }
        }

        public ResourceSelections(BrowserUI _browserUI)
        {
            browserUI = _browserUI;
            browserFiltersUI = _browserUI.browserFiltersUI;
            
        }
        public void InitUI(GameObject _resourceSelectionsBoxGO)
        {
            resourceSelectionsBoxGO = _resourceSelectionsBoxGO;
            resourceSelectionsScrollableBoxGO = UI.CreateVScrollableContent(resourceSelectionsBoxGO, BrowserUI.darkBlueGrey, AnchorPresets.vhStretch, resourceSelectionBoxWidth, resourceSelectionBoxHeight, new Vector2(0f, 0f));
            resourceSelectionScrollableViewportContentGO = resourceSelectionsScrollableBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            RePopulateSelectionsBox();

        }

        private void ClearResourceSelectionRows()
        {
            foreach (var rowGO in resourceSelectionRows) GameObject.Destroy(rowGO);
            resourceSelectionRows.Clear();
        }
        public void RePopulateSelectionsBox()
        {
            int resourceType = browserFiltersUI.resourceType;
            ClearResourceSelectionRows();
            selectedRVGEButtons.Clear();

            if (browserUI!=null && browserUI.resourceBrowserUI!=null) browserUI.resourceBrowserUI.SetMultiSelectClear();

            if (resourceTypeSelection.ContainsKey(resourceType))
            {

                GameObject currentRowGO = CreateSelectResourceRow();

                float currentRowResourcesWidth = 0f;

                List<ResourceVersionGroupEntry> selectedRVGEs;
                if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) selectedRVGEs = resourceTypeSelection[resourceType].selectedRVGEs;
                else 
                {
                    selectedRVGEs = new List<ResourceVersionGroupEntry>();
                    var firstSelectedRVGE = resourceTypeSelection[resourceType].GetFirstSelectedRVGE();
                    if (firstSelectedRVGE!=null) selectedRVGEs.Add(firstSelectedRVGE);
                }

                foreach (var resourceVersionGroupEntry in selectedRVGEs)
                {
                    float nextButtonWidth = GetRVGEButtonWidth(resourceVersionGroupEntry);
                    float cancelButtonWidth = 0f;
                    if (selectedRVGEs.Count > 1 && browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) cancelButtonWidth = 18f;

                    if (currentRowResourcesWidth + nextButtonWidth + cancelButtonWidth > resourceSelectionBoxWidth)
                    {
                        currentRowGO = CreateSelectResourceRow();
                        currentRowResourcesWidth = 0f;
                    }
                    var resourceButton = UI.CreateUIButton(currentRowGO, nextButtonWidth, 16f, resourceVersionGroupEntry.resourceDisplayName, 20, currentRowResourcesWidth, 0, AnchorPresets.middleLeft);
                    selectedRVGEButtons.Add(resourceVersionGroupEntry, resourceButton);
                    resourceButton.buttonImage.sprite = null;
                    resourceButton.button.interactable = false;
                    resourceButton.buttonText.resizeTextForBestFit = true;
                    resourceButton.buttonText.resizeTextMinSize = 15;
                    resourceButton.buttonText.resizeTextMaxSize = 21;

                    if (resourceVersionGroupEntry == selectedRVGEs.First()) resourceButton.buttonColor = Color.cyan;
                    ColorBlock colors = resourceButton.button.colors;
                    colors.disabledColor = BrowserUI.lightGrey;
                    resourceButton.button.colors = colors ;
                    SelectedResourceButtonPointerEnterExit click = resourceButton.button.gameObject.AddComponent<SelectedResourceButtonPointerEnterExit>();
                    click.Init(resourceVersionGroupEntry, browserUI.helpTextJSS);

                    currentRowResourcesWidth += nextButtonWidth ;

                    if (cancelButtonWidth>0f)
                    {
                        var cancelButton = UI.CreateUIButton(currentRowGO, 16f, 16f, "X", 22, currentRowResourcesWidth, 0, AnchorPresets.middleLeft);
                        cancelButton.buttonImage.sprite = null;
                        cancelButton.buttonColor = Color.cyan;
                        currentRowResourcesWidth += 18f;
                        SelectedResourceButtonPointerEnterExit click2 = cancelButton.button.gameObject.AddComponent<SelectedResourceButtonPointerEnterExit>();
                        click2.Init(resourceVersionGroupEntry, browserUI.helpTextJSS);
                        cancelButton.button.onClick.AddListener(delegate
                        {
                            DeSelectRVGE(resourceVersionGroupEntry);
                        });
                    }
                    currentRowResourcesWidth += 3f;
                }
                if (selectedRVGEs.Count < 2 && !browserUI.isVREnabled && browserUI.browserFiltersUI.selectModeJSEnum.val==BrowserSelectMode.multi)
                {
                    var guideTextJSS = new JSONStorableString("guideText", "<i>Hold <Ctrl> key to add or remove a clicked resource to these selections. Use left <Shift> to select multiple contiguous resources prior to the clicked resource. Use right <Shift> to select multiple contiguous resources after the clicked resource. Use <Alt> key to change the last selected resource.</i>");
                    var textBox = UI.CreateBasicTextBox(currentRowGO, guideTextJSS, resourceSelectionBoxWidth- currentRowResourcesWidth, 40, 20, currentRowResourcesWidth, 0f, AnchorPresets.topLeft,  backgoundColor: new Color(0f,0f,0f,0f));

                }
            }
            if (browserUI.resourceActionsUI != null)
            {
                browserUI.resourceActionsUI.primaryActionUITab.deleteConfirmationActive = false;
            }
        }

        private GameObject CreateSelectResourceRow()
        {
            GameObject currentRowGO = UI.AddChildGOImage(resourceSelectionScrollableViewportContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, resourceSelectionBoxWidth, 20f, Vector2.zero);
            resourceSelectionRows.Add(currentRowGO);

            return currentRowGO;
        }

        private float GetRVGEButtonWidth(ResourceVersionGroupEntry rvge)
        {
            if (browserFiltersUI.selectModeJSEnum.val!=BrowserSelectMode.multi) return Mathf.Clamp(rvge.resourceDisplayName.Length * 7, 80f, 400f);
            return Mathf.Clamp(rvge.resourceDisplayName.Length * 6,40f,200f);

        }

        public int SelectionCount(int resourceType)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return 0;
            return resourceTypeSelection[resourceType].selectedRVGEs.Count;
        }
        public bool IsSelected(ResourceVersionGroupEntry rvge, int selectionMode, UIDynamicButton updatedButton)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI));
            return resourceTypeSelection[rvge.resourceType].IsSelected(rvge, selectionMode, updatedButton);
        }

        public void SingleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI));
            resourceTypeSelection[rvge.resourceType].SingleSelectRVGE(rvge, button);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUI();
        }

        public void LeftShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI));
            resourceTypeSelection[rvge.resourceType].LeftShiftSelectRVGE(rvge, button, currentPageRVGEs, currentPageButtons);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUI();
        }
        public void RightShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI));
            resourceTypeSelection[rvge.resourceType].RightShiftSelectRVGE(rvge, button, currentPageRVGEs, currentPageButtons);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUI();
        }

        public void ToggleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI));
            resourceTypeSelection[rvge.resourceType].ToggleSelectRVGE(rvge, button);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUI();
        }

        public void UpdateDisplayName(ResourceVersionGroupEntry rvge)
        {
            if (selectedRVGEButtons.ContainsKey(rvge))
            {
                selectedRVGEButtons[rvge].label = rvge.resourceDisplayName;
            }

        }
        public void ClearSelectedRVGEButtons(int resourceType)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection[resourceType].ClearSelectedRVGEButtons();
        }

        public void DeSelectRVGE(ResourceVersionGroupEntry rvge)
        {
            resourceTypeSelection[rvge.resourceType].DeSelectRVGE(rvge);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUI();
        }

        public void DeSelectAll(int resourceType, bool keepFirst = true)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection[resourceType].DeSelectAll(keepFirst);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUI();
        }

        public void SelectAll(int resourceType, List<ResourceVersionGroupEntry> currentFilterResults)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection[resourceType].SelectAll(currentFilterResults);

            RePopulateSelectionsBox();
            if (browserUI.resourceActionsUI != null) browserUI.resourceActionsUI.RefreshUI();
        }

        public void SwitchSelectMode(int resourceType, int selectionMode)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection.Add(resourceType, new ResourceTypeSelections(browserUI));
            resourceTypeSelection[resourceType].SwitchSelectMode(selectionMode);

            RePopulateSelectionsBox();
        }

        public ResourceVersionGroupEntry GetFirstSelectedRVGE(int resourceType)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) return resourceTypeSelection[resourceType].GetFirstSelectedRVGE();
            return null;
        }

        public int IsSelectionTagged(int resourceType, int selectMode, Tag tag)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return TriState.inactive;
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection[resourceType].IsFirstTagged(tag);
            return resourceTypeSelection[resourceType].IsTagged(tag);
        }

        public int IsSelectionFavourite(int resourceType, int selectMode)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return TriState.inactive;
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection[resourceType].isFirstFavourite;
            return resourceTypeSelection[resourceType].isFavourite;
        }
        public int IsSelectionNew(int resourceType, int selectMode)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return TriState.inactive;
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection[resourceType].isFirstNew;
            return resourceTypeSelection[resourceType].isNew;
        }
        public int IsSelectionHidden(int resourceType, int selectMode)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return TriState.inactive;
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection[resourceType].isFirstHidden;
            return resourceTypeSelection[resourceType].isHidden;
        }
        public void SetSelectionTag(int resourceType, int selectMode, Tag tag, int tagTriState)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return;
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection[resourceType].SetFirstTagged(tag, tagTriState);
            else resourceTypeSelection[resourceType].SetTagged(tag, tagTriState);
        }
        public void SetSelectionFavourite (int resourceType, int selectMode, int favTriState)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return ;
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection[resourceType].isFirstFavourite = favTriState;
            else resourceTypeSelection[resourceType].isFavourite = favTriState;
        }

        public void SetSelectionHidden(int resourceType, int selectMode, int hiddenTriState)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return;
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection[resourceType].isFirstHidden = hiddenTriState;
            else resourceTypeSelection[resourceType].isHidden = hiddenTriState;
        }
        public void SetSelectionNew(int resourceType, int selectMode, int hiddenTriState)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return;
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection[resourceType].isFirstNew = hiddenTriState;
            else resourceTypeSelection[resourceType].isNew = hiddenTriState;
        }
    }

    public class ResourceTypeSelections
    {
        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> selectedRVGEButtons = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        public List<ResourceVersionGroupEntry> selectedRVGEs = new List<ResourceVersionGroupEntry>();

        private List<ResourceVersionGroupEntry> randomSelectedRVGEs = new List<ResourceVersionGroupEntry>();
        private ResourceVersionGroupEntry lastRandomRVGE = null;

        private ResourceBrowserUI resourceBrowserUI;
        private BrowserUI browserUI;

        public int isHidden { get
            {
                bool hiddenFound = false;
                bool notHiddenFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isHidden) hiddenFound = true;
                    else notHiddenFound = true;

                    if (hiddenFound && notHiddenFound) return TriState.indeterminate;
                }

                if (hiddenFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool hide = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isHidden = hide;
                    resourceBrowserUI.SetHideFavColorTags(rvge);
                }
            }
        }

        public int isFirstHidden
        {
            get
            {
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isHidden)
                {
                    return TriState.active;
                }
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool hide = (value == TriState.active);
                var rvge = GetFirstSelectedRVGE();

                rvge.isHidden = hide;
                resourceBrowserUI.SetHideFavColorTags(rvge);
            }
        }

        public int isFavourite
        {
            get
            {
                bool favFound = false;
                bool notFavFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isFavorite) favFound = true;
                    else notFavFound = true;

                    if (favFound && notFavFound) return TriState.indeterminate;
                }

                if (favFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool fav = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isFavorite = fav;
                    resourceBrowserUI.SetHideFavColorTags(rvge);
                }
            }
        }

        public bool containsLocal
        {
            get
            {
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.localFileResource) return true;
                }
                return false;
            }
        }
        public int isNew
        {
            get
            {
                bool newFound = false;
                bool notNewFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isNew) newFound = true;
                    else notNewFound = true;

                    if (newFound && notNewFound) return TriState.indeterminate;
                }

                if (newFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool resNew = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isNew = resNew;
                    resourceBrowserUI.SetNewIcon(rvge);
                }
            }
        }

        public int isFirstFavourite
        {
            get
            {
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isFavorite) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool fav = (value == TriState.active);
                var rvge = GetFirstSelectedRVGE();
                rvge.isFavorite = fav;
                resourceBrowserUI.SetHideFavColorTags(rvge);
            }
        }

        public int isFirstNew
        {
            get
            {
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isNew) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool resNew = (value == TriState.active);
                var rvge = GetFirstSelectedRVGE();
                rvge.isNew = resNew;
                resourceBrowserUI.SetNewIcon(rvge);
            }
        }

        public ResourceTypeSelections(BrowserUI _browserUI)
        {
            resourceBrowserUI = _browserUI.resourceBrowserUI;
            browserUI = _browserUI;
        }

        public void SwitchSelectMode(int selectMode)
        {
            if (selectedRVGEs.Count > 0)
            {
                UIDynamicButton firstButton = null;

                if (selectedRVGEButtons.ContainsKey(selectedRVGEs.First()))
                    firstButton = selectedRVGEButtons[selectedRVGEs.First()];
                foreach (var rvgeButton in selectedRVGEButtons)
                {
                    if ( selectMode == BrowserSelectMode.multi || rvgeButton.Value == firstButton ) ResourceBrowserUI.SelectColor(rvgeButton.Value);
                    else ResourceBrowserUI.DeSelectColor(rvgeButton.Value);
                }

            }

        }

        public void ClearSelectedRVGEButtons()
        {
            selectedRVGEButtons.Clear();
        }

        public void RightShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            int selectionIndex = currentPageButtons.FindIndex(x => x == button);
            int altIndex = currentPageButtons.Count-1;

            bool selectToggle = !selectedRVGEs.Contains(rvge);

            for (int searchIndex = selectionIndex +1; searchIndex < currentPageButtons.Count; searchIndex++)
            {
                if (selectToggle && selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex -1;
                    break;
                }
                if (!selectToggle && !selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex -1;
                    break;
                }
            }

            for (int toggleIndex = altIndex; toggleIndex >= selectionIndex; toggleIndex--)
            {
                ToggleSelectRVGE(currentPageRVGEs[toggleIndex], currentPageButtons[toggleIndex]);
            }

        }

        public void LeftShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            int selectionIndex = currentPageButtons.FindIndex(x => x==button);
            int altIndex = 0;

            bool selectToggle = !selectedRVGEs.Contains(rvge);

            for (int searchIndex = selectionIndex - 1; searchIndex >= 0; searchIndex--)
            {
                if (selectToggle && selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex + 1;
                    break;
                }
                if (!selectToggle && !selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex + 1;
                    break;
                }
            }
    
            for (int toggleIndex = altIndex; toggleIndex <= selectionIndex; toggleIndex++)
            {
                ToggleSelectRVGE(currentPageRVGEs[toggleIndex], currentPageButtons[toggleIndex]);
            }

        }
        public void ToggleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (!selectedRVGEs.Contains(rvge)) SelectRVGE(rvge, button);
            else DeSelectRVGE(rvge);
        }
        private void SelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (selectedRVGEs.Contains(rvge)) selectedRVGEs.Remove(rvge);
            selectedRVGEs.Insert(0,rvge);
            

            if (!selectedRVGEButtons.ContainsKey(rvge)) selectedRVGEButtons.Add(rvge, button);
            else selectedRVGEButtons[rvge] = button;
            browserUI.resourceActionsUI.primaryActionUITab.RefreshTargetAtomValue();

            if (button!=null) ResourceBrowserUI.SelectColor(button);

            randomSelectedRVGEs.Clear();
        }
        public void SingleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (selectedRVGEs.Count > 0)
            {
                ResourceVersionGroupEntry firstRVGE = selectedRVGEs.First();
                if (selectedRVGEButtons.ContainsKey(firstRVGE)) ResourceBrowserUI.DeSelectColor(selectedRVGEButtons[firstRVGE]);

               selectedRVGEs.Remove(firstRVGE);
                selectedRVGEButtons.Remove(firstRVGE);
            }
            SelectRVGE(rvge, button);

        }
        public void DeSelectRVGE(ResourceVersionGroupEntry rvge)
        {
            
            selectedRVGEs.Remove(rvge);
            if (selectedRVGEButtons.ContainsKey(rvge))
            {
                UIDynamicButton button = selectedRVGEButtons[rvge];
                selectedRVGEButtons.Remove(rvge);
                if (button!=null) ResourceBrowserUI.DeSelectColor(button);
            }
            randomSelectedRVGEs.Clear();
        }

        public ResourceVersionGroupEntry GetFirstSelectedRVGE()
        {
            if (selectedRVGEs.Count>0) return selectedRVGEs.First();
            return null;
        }

        public void DeSelectAll(bool keepFirst)
        {
            ResourceVersionGroupEntry firstRVGE = null;

            if (selectedRVGEs.Count>0) firstRVGE = selectedRVGEs.First();

            UIDynamicButton firstRVGEButton =null;
            foreach (var kvp in selectedRVGEButtons)
            {
                if (kvp.Key != firstRVGE)
                {
                   if (kvp.Value != null) ResourceBrowserUI.DeSelectColor(kvp.Value);
                } 
                else firstRVGEButton = kvp.Value;
            }
            selectedRVGEs.Clear();
            selectedRVGEButtons.Clear();
            if (firstRVGE!=null && keepFirst)
            {
                selectedRVGEs.Add(firstRVGE);
                selectedRVGEButtons.Add(firstRVGE, firstRVGEButton);
            }
            
        }

        public void SelectAll(List<ResourceVersionGroupEntry> currentFilterResults)
        {
            foreach (var rvge in currentFilterResults)
            {
                SelectRVGE(rvge, resourceBrowserUI.GetUIDButton(rvge));
            }
        }

        public bool IsSelected(ResourceVersionGroupEntry rvge, int selectionMode, UIDynamicButton updatedButton)
        {
            if (!selectedRVGEs.Contains(rvge)) return false;
            if (!selectedRVGEButtons.ContainsKey(rvge)) selectedRVGEButtons.Add(rvge, updatedButton);
            else selectedRVGEButtons[rvge] = updatedButton;

            ResourceVersionGroupEntry firstRVGE = selectedRVGEs.First();
            if (rvge == firstRVGE || selectionMode == BrowserSelectMode.multi) return true;
            return false;
        }

        public int IsFirstTagged(Tag tag)
        {
            if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().IsTagged(tag)) return TriState.active;
            return TriState.inactive;
        }

        public int IsTagged(Tag tag)
        {
            bool tagFound = false;
            bool notTagFound = false;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.IsTagged(tag)) tagFound = true;
                else notTagFound = true;

                if (tagFound && notTagFound) return TriState.indeterminate;
            }

            if (tagFound) return TriState.active;
            return TriState.inactive;
        }
        public void SetTagged(Tag tag, int value)
        {
            if (value == TriState.indeterminate) return;
            bool tagValue = (value == TriState.active);
            foreach (var rvge in selectedRVGEs)
            {
                rvge.SetTag(tag, tagValue);
            }
            
        }
        public int GetFavouriteCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isFavorite) count++;
            }

            return count;
        }
        public int GetHiddenCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isHidden) count++;
            }

            return count;
        }
        public int GetNewCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isNew) count++;
            }

            return count;
        }
        public int GetTaggedCount(Tag tag)
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.IsTagged(tag)) count++;
            }

            return count;
        }

        public int GetSelectedCount()
        {
            return selectedRVGEs.Count();
        }

        public void SetFirstTagged(Tag tag, int value)
        {
            if (value == TriState.indeterminate) return;
            bool tagValue = (value == TriState.active);
            GetFirstSelectedRVGE().SetTag(tag, tagValue);
        }


        public ResourceVersionGroupEntry GetNextRandomRVGE()
        {
            if (randomSelectedRVGEs.Count == 0)
            {
                randomSelectedRVGEs = selectedRVGEs.OrderBy(item => UnityEngine.Random.Range(0, 1f)).ToList();
            }
            if (randomSelectedRVGEs.Count > 0)
            {
                ResourceVersionGroupEntry nextRVGE = null;
                if (randomSelectedRVGEs.Count == 1) nextRVGE = randomSelectedRVGEs.First();
                else
                {
                    while (nextRVGE == null || nextRVGE == lastRandomRVGE)
                    {
                        nextRVGE = randomSelectedRVGEs.First();
                    }
                }
                lastRandomRVGE = nextRVGE;
                randomSelectedRVGEs.Remove(nextRVGE);
                return nextRVGE;
            }
            return null;
        }
    }
    public class ResourceBrowserUI
    {
        private GameObject parentBoxGO;
        private BrowserUI parentBrowser;
        public BrowserFiltersUI browserFiltersUI;

        public JSONStorableString resourceCountTextJSS;

        private int resourceButtonSize
        {
            get
            {
                if (Globals.isVREnabled) return Settings.resourceButtonSizeVR;
                else return Settings.resourceButtonSizeDesktop;
            }
            set
            {
                if (Globals.isVREnabled) Settings.resourceButtonSizeVR = value;
                else Settings.resourceButtonSizeDesktop = value;
            }
        }
        const int resourceButtonlabelHeight = 16;

        private UIDynamicButton firstPageUIDButton;
        private UIDynamicButton prevPageUIDButton;
        private UIDynamicButton nextPageUIDButton;
        private UIDynamicButton lastPageUIDButton;
        private UIDynamicButton incResourceButtonSizesUIDButton;
        private UIDynamicButton decResourceButtonSizesUIDButton;
        private UIDynamicButton clearMultiSelectResourcesUIDButton;
        private UIDynamicButton selectAllResourcesUIDButton;
        private GameObject resourceCountTextBox;

        private Dictionary<UIDynamicButton, ResourceVersionGroupEntry> UIDButtonToRVGEDict = new Dictionary<UIDynamicButton, ResourceVersionGroupEntry>();
        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> RVGEToUIDButtonDict = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        private List<GameObject> activeResourceRowGOs = new List<GameObject>();
        private List<GameObject> inactiveResourceRowGOs = new List<GameObject>();

        private ScrollRect scrollRect;

        private Dictionary<int, float> scrollBarPositionByResourceType = new Dictionary<int, float>();

        private float scrollBarPosition
        {
            get
            {
                
                int currentResourceType = browserFiltersUI.resourceType;
                if (!scrollBarPositionByResourceType.ContainsKey(currentResourceType)) scrollBarPositionByResourceType.Add(currentResourceType, 1);
                return scrollBarPositionByResourceType[currentResourceType];
            }
            set
            {
                int currentResourceType = browserFiltersUI.resourceType;
                if (!scrollBarPositionByResourceType.ContainsKey(currentResourceType)) scrollBarPositionByResourceType.Add(currentResourceType, 1);
                scrollBarPositionByResourceType[currentResourceType] = value;
            }
        }

        private List<UIDynamicButton> activeButtonUIDs = new List<UIDynamicButton>();
        private HashSet<UIDynamicButton> unparentedButtonUIDs = new HashSet<UIDynamicButton>();

        private Dictionary<GameObject, List<UIDynamicButton>> buttonUIDsByRow = new Dictionary<GameObject, List<UIDynamicButton>>();
        private List<ResourceVersionGroupEntry> currentFilterResults { get { return parentBrowser.currentFilterResults; } }

        public void OnDestory()
        {
            foreach (UIDynamicButton buttonUID in activeButtonUIDs)
            {
                GameObject.Destroy(buttonUID);

            }
            foreach (UIDynamicButton buttonUID in unparentedButtonUIDs)
            {
                GameObject.Destroy(buttonUID);

            }
        }

        private int currentFilterResultsPage { get { return parentBrowser.currentFilterResultsPage; } set { parentBrowser.currentFilterResultsPage = value; } }

        private List<ResourceVersionGroupEntry> currentPageRVGEs = new List<ResourceVersionGroupEntry>();
        private List<UIDynamicButton> currentPageButtons = new List<UIDynamicButton>();
        private float resouceButtonWidth
        {
            get
            {
                if (resourceButtonSize == 0) return 60;
                if (resourceButtonSize == 1) return 67;
                if (resourceButtonSize == 2) return 75.5f;
                if (resourceButtonSize == 3) return 87;
                if (resourceButtonSize == 4) return 102;
                if (resourceButtonSize == 5) return 123;
                if (resourceButtonSize == 6) return 154;
                return 123;
            }
        }
        private int buttonsPerRow
        {
            get
            {
                return (int)Math.Floor((parentBrowser.browserScreenWidth-20f) / (float)(resouceButtonWidth + 3));
            }
        }
        private int selectedResourcesVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return -23;
                else return -45;
            }
        }
        
        private float navButtonsHOffset
        {
            get
            {
                return parentBrowser.resourceSelectionsUI.resourceSelectionBoxWidth - (640/2) + BrowserUI.edgeSpacerSize +7.5f ;
            }
        }
        private float navButtonsVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return selectedResourcesVOffset;
                else return selectedResourcesVOffset + 5;
            }
        }
        private float resourceCountHOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return navButtonsHOffset + 47f+(76f/2) +15f;
                else return navButtonsHOffset + (76f / 2);
            }
        }
        private int resourceCountVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return selectedResourcesVOffset;
                else return selectedResourcesVOffset + 25;
            }
        }
        private int resoureButtonSizeHOffset
        {
            get
            {
                return (640/2)-30;
            }
        }

        public void NudgeScrollBar()
        {
            if (scrollRect != null)
            {
                if (scrollBarPosition < 0.5f)
                {
                    scrollBarPosition += 0.1f;
                    scrollBarPosition -= 0.1f;
                }
                else
                {
                    scrollBarPosition -= 0.1f;
                    scrollBarPosition += 0.1f;
                }              
            }
            
        }
        public ResourceBrowserUI(BrowserUI _parentBrowser, GameObject _parentBoxGO)
        {
            parentBrowser = _parentBrowser;
            parentBoxGO = _parentBoxGO;
            browserFiltersUI = parentBrowser.browserFiltersUI;
            scrollRect = parentBoxGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbar.onValueChanged.AddListener(delegate
            {
                if (scrollRect.verticalScrollbar.size > 0.999f) scrollBarPosition = 1f;
                else scrollBarPosition = scrollRect.verticalNormalizedPosition;
            });
            scrollRect.verticalNormalizedPosition = scrollBarPosition;

            firstPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>|<</b>", 26, navButtonsHOffset, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            firstPageUIDButton.buttonImage.sprite = null;
            firstPageUIDButton.button.onClick.AddListener(delegate {
                currentFilterResultsPage = 1;                
                scrollRect.verticalNormalizedPosition = 1f;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(firstPageUIDButton, "Navigate to the first page of browser results");

            prevPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b><</b>", 26, navButtonsHOffset + 15, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            prevPageUIDButton.buttonImage.sprite = null;
            prevPageUIDButton.button.onClick.AddListener(delegate {
                currentFilterResultsPage--;
                scrollRect.verticalNormalizedPosition = 1f;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(prevPageUIDButton, "Navigate to the previous page of browser results");

            nextPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>></b>", 26, navButtonsHOffset + 30, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            nextPageUIDButton.buttonImage.sprite = null;
            nextPageUIDButton.button.onClick.AddListener(delegate {
                if ((currentFilterResultsPage) * Settings.maxButtonsPerPage < currentFilterResults.Count)
                {
                    scrollRect.verticalNormalizedPosition = 1f;
                    currentFilterResultsPage++;
                    RepopulateButtons();
                }

            });
            parentBrowser.ApplyHelp(nextPageUIDButton, "Navigate to the next page of browser results");

            lastPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>>|</b>", 26, navButtonsHOffset +45, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            lastPageUIDButton.buttonImage.sprite = null;
            lastPageUIDButton.button.onClick.AddListener(delegate {
                currentFilterResultsPage = (int) Math.Ceiling(currentFilterResults.Count / Settings.maxButtonsPerPage);
                scrollRect.verticalNormalizedPosition = 1f;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(lastPageUIDButton, "Navigate to the last page of browser results");


            incResourceButtonSizesUIDButton = UI.CreateUIButton(parentBoxGO, 20f, 20f, "<b>+</b>", 26, resoureButtonSizeHOffset, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            incResourceButtonSizesUIDButton.buttonImage.sprite = null;
            incResourceButtonSizesUIDButton.button.onClick.AddListener(delegate {
                if (resourceButtonSize < 6) resourceButtonSize++;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(incResourceButtonSizesUIDButton, "Increase the size of resource icons in the browser results window.");

            decResourceButtonSizesUIDButton = UI.CreateUIButton(parentBoxGO, 20f, 20f, "<b>-</b>", 26, resoureButtonSizeHOffset+20, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            decResourceButtonSizesUIDButton.buttonImage.sprite = null;
            decResourceButtonSizesUIDButton.button.onClick.AddListener(delegate {
                if (resourceButtonSize > 0) resourceButtonSize--;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(decResourceButtonSizesUIDButton, "Decrease the size of resource icons in the browser results window.");

            clearMultiSelectResourcesUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 15f, "X", 22, navButtonsHOffset, selectedResourcesVOffset-31f, anchorPreset: AnchorPresets.bottomMiddle);
            clearMultiSelectResourcesUIDButton.buttonImage.sprite = null;
            clearMultiSelectResourcesUIDButton.buttonColor = Color.cyan;
            clearMultiSelectResourcesUIDButton.button.onClick.AddListener(delegate {
                parentBrowser.resourceSelectionsUI.DeSelectAll(browserFiltersUI.resourceType);
            });
            clearMultiSelectResourcesUIDButton.gameObject.SetActive(false);
            parentBrowser.ApplyHelp(clearMultiSelectResourcesUIDButton, "Clear all resources (other than the very latest selected resource) from the selection window.");

            selectAllResourcesUIDButton = UI.CreateUIButton(parentBoxGO,65f, 15f, "Select All", 22, navButtonsHOffset, navButtonsVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            selectAllResourcesUIDButton.buttonImage.sprite = null;
            selectAllResourcesUIDButton.buttonColor = Color.cyan;
            selectAllResourcesUIDButton.button.onClick.AddListener(delegate {

                if (parentBrowser.isVREnabled || (!Input.GetKey(KeyCode.LeftControl)) && !Input.GetKey(KeyCode.RightControl)) parentBrowser.resourceSelectionsUI.DeSelectAll(browserFiltersUI.resourceType,false);
                parentBrowser.resourceSelectionsUI.SelectAll(browserFiltersUI.resourceType, currentFilterResults);
            });
            selectAllResourcesUIDButton.gameObject.SetActive(false);
            if (parentBrowser.isVREnabled) parentBrowser.ApplyHelp(selectAllResourcesUIDButton, "Select all resources from the the browser results (replacing existing selections).");
            else parentBrowser.ApplyHelp(selectAllResourcesUIDButton, "Select all resources from the the browser results (replacing existing selections). Hold the Ctrl Key whilst pressing 'Select All' to add all resources from the the browser results to the existing selections.");

            resourceCountTextJSS = new JSONStorableString("resourceCountText", "");
            resourceCountTextBox = UI.CreateBasicTextBox(parentBoxGO, resourceCountTextJSS, 90f, 20f, 22, resourceCountHOffset, resourceCountVOffset, AnchorPresets.bottomMiddle, true, 18);
            parentBrowser.ApplyHelp(resourceCountTextBox, "Total count of resources from the broswer filter results and the range displayed on the current page of results.");
        }

        public void UpdateDisplayName(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                RVGEToUIDButtonDict[rvge].label = rvge.resourceDisplayName;
            }
        }

        public void ReloadSprite(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                var button = RVGEToUIDButtonDict[rvge];
                Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
                UnityEngine.UI.Image image = newImageTrans.GetComponent<UnityEngine.UI.Image>();
                image.sprite = rvge.resourceSprite;
            }
        }
        public static void DeSelectColor(UIDynamicButton buttonUID)
        {
            if (buttonUID!=null) buttonUID.buttonColor = BrowserUI.lightGrey;
        }
        public static void SelectColor(UIDynamicButton buttonUID)
        {
            if (buttonUID!=null) buttonUID.buttonColor = Color.cyan;
        }
        public void RepopulateButtons(bool forceSelectedResourceRefresh=false)
        {
            float newScrollBarPosition = scrollBarPosition;

            int buttonsPerPage = (int)Settings.maxButtonsPerPage;
            if ((currentFilterResultsPage - 1) * buttonsPerPage > currentFilterResults.Count) currentFilterResultsPage = 1;
            if (currentFilterResultsPage < 1) currentFilterResultsPage = 1;
            int buttonCount = Math.Min(currentFilterResults.Count, Math.Min(buttonsPerPage, currentFilterResults.Count - ((currentFilterResultsPage - 1) * buttonsPerPage)));

            RefactorButtonRows(buttonCount, buttonsPerRow);

            foreach (var kvp in UIDButtonToRVGEDict) kvp.Value.CancelButtonImageUpdate();

            UIDButtonToRVGEDict.Clear();
            RVGEToUIDButtonDict.Clear();

            int resourceIndexOffset = (currentFilterResultsPage - 1) * buttonsPerPage;

            ResourceVersionGroupEntry firstSelectedRVGE = parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(browserFiltersUI.resourceType);
            if (firstSelectedRVGE == null || (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi && forceSelectedResourceRefresh && !currentFilterResults.Contains(firstSelectedRVGE)))
            {
                if (currentFilterResults.Count!=0) SelectRVGE(currentFilterResults[resourceIndexOffset], activeButtonUIDs[0], true);
            }


            if (currentFilterResultsPage == 1)
            {
                firstPageUIDButton.button.enabled = false;
                firstPageUIDButton.textColor = Color.white;
                prevPageUIDButton.button.enabled = false;
                prevPageUIDButton.textColor = Color.white;
            }
            else
            {
                firstPageUIDButton.button.enabled = true;
                firstPageUIDButton.textColor = Color.black;
                prevPageUIDButton.button.enabled = true;
                prevPageUIDButton.textColor = Color.black;
            }
            if (currentFilterResultsPage >= Math.Ceiling(currentFilterResults.Count / (float)buttonsPerPage))
            {
                nextPageUIDButton.button.enabled = false;
                nextPageUIDButton.textColor = Color.white;
                lastPageUIDButton.button.enabled = false;
                lastPageUIDButton.textColor = Color.white;
            }
            else
            {
                nextPageUIDButton.button.enabled = true;
                nextPageUIDButton.textColor = Color.black;
                lastPageUIDButton.button.enabled = true;
                lastPageUIDButton.textColor = Color.black;
            }

            if (resourceButtonSize == 0)
            {
                decResourceButtonSizesUIDButton.button.enabled = false;
                decResourceButtonSizesUIDButton.textColor = Color.white;
            }
            else {
                decResourceButtonSizesUIDButton.button.enabled = true;
                decResourceButtonSizesUIDButton.textColor = Color.black;
            }
            if (resourceButtonSize == 6)
            {
                incResourceButtonSizesUIDButton.button.enabled = false;
                incResourceButtonSizesUIDButton.textColor = Color.white;
            }
            else
            {
                incResourceButtonSizesUIDButton.button.enabled = true;
                incResourceButtonSizesUIDButton.textColor = Color.black;
            }

            resourceCountTextJSS.val = (resourceIndexOffset + 1).ToString() + " - " + (resourceIndexOffset + buttonCount).ToString() + " of " + currentFilterResults.Count.ToString();

            currentPageRVGEs.Clear();
            currentPageButtons.Clear();

            parentBrowser.resourceSelectionsUI.ClearSelectedRVGEButtons(browserFiltersUI.resourceType);

            for (int buttonId = buttonCount - 1; buttonId >= 0; buttonId--)
            {
                ResourceVersionGroupEntry rvge = currentFilterResults[buttonId + resourceIndexOffset];
                currentPageRVGEs.Insert(0,rvge);

                UIDynamicButton buttonUID = activeButtonUIDs[buttonId];
                currentPageButtons.Insert(0,buttonUID);

                UIDButtonToRVGEDict.Add(buttonUID, rvge);
                RVGEToUIDButtonDict.Add(rvge, buttonUID);
                if (parentBrowser.resourceSelectionsUI.IsSelected(rvge, browserFiltersUI.selectModeJSEnum.val, buttonUID)) buttonUID.buttonColor = Color.cyan;
                else DeSelectColor(buttonUID);

                SetHideFavColorTags(buttonUID.button, rvge);
                SetNewIcon(buttonUID.button, rvge);

                buttonUID.label = rvge.resourceDisplayName;

                buttonUID.button.gameObject.SetActive(true);

                buttonUID.button.onClick.RemoveAllListeners();
                buttonUID.button.onClick.AddListener(delegate {
                    SelectRVGE(rvge, buttonUID);
                });

                currentFilterResults[buttonId + resourceIndexOffset].LoadSprite(activeButtonUIDs[buttonId]);
            }

            for (int i = buttonCount; i < (activeResourceRowGOs.Count) * buttonsPerRow; i++)
            {
                activeButtonUIDs[i].label = "";
                activeButtonUIDs[i].button.gameObject.SetActive(false);
            }

            if (forceSelectedResourceRefresh)
            {
                Globals.mvrScript.StartCoroutine(SetScollBarPosition(newScrollBarPosition));
            }

            
        }

        private IEnumerator SetScollBarPosition(float newScrollBarPosition)
        {
            yield return new WaitForEndOfFrame();
            scrollRect.verticalNormalizedPosition = newScrollBarPosition;
        }

        public void SetButtonLoading(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                Button button = RVGEToUIDButtonDict[rvge].button;
                Transform newIconTransform = button.transform.Find("LoadingIcon");
                newIconTransform.gameObject.SetActive(true);
            }
                
        }

        public void UnsetButtonLoading(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                Button button = RVGEToUIDButtonDict[rvge].button;
                Transform newIconTransform = button.transform.Find("LoadingIcon");
                newIconTransform.gameObject.SetActive(false);
            }
        }

        private void SetNewIcon(Button button, ResourceVersionGroupEntry rvge)
        {
            Transform newIconTransform = button.transform.Find("NewIcon");
            Transform updatedIconTransform = button.transform.Find("UpdatedIcon");

            if (rvge.isMultiVARVersion)
            {
                newIconTransform.gameObject.SetActive(false);
                updatedIconTransform.gameObject.SetActive(rvge.isNew);
            }
            else
            {
                updatedIconTransform.gameObject.SetActive(false);
                newIconTransform.gameObject.SetActive(rvge.isNew);
            }
        }

        public void SetNewIcon(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                SetNewIcon(RVGEToUIDButtonDict[rvge].button, rvge);
            }
        }

        private void SetHideFavColorTags(Button button, ResourceVersionGroupEntry rvge)
        {
            Transform hiddenImageTransform = button.transform.Find("HiddenIconImage");
            Transform favImageTransform = button.transform.Find("FavIconImage");

            favImageTransform.gameObject.SetActive(rvge.isFavorite);
            hiddenImageTransform.gameObject.SetActive(rvge.isHidden || rvge.isTagHidden);
        }

        public void SetHideFavColorTags(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge))
            {
                SetHideFavColorTags(RVGEToUIDButtonDict[rvge].button, rvge);
            }
        }

        public UIDynamicButton GetUIDButton(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge)) return RVGEToUIDButtonDict[rvge];
            return null;
        }

        public void RefactorButtonRows(int requiredButtonCount, int requiredButtonsPerRow)
        {
            int rowsRequired = (int)Math.Ceiling((double)requiredButtonCount / requiredButtonsPerRow);
            GameObject buttonListContentGO = parentBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            int newRowsRequired = rowsRequired - activeResourceRowGOs.Count;

            if (buttonUIDsByRow.Count > 0)
            {
                int startingButtonsPerRow = buttonUIDsByRow.First().Value.Count();

                // Change in number of buttons per row
                if (startingButtonsPerRow != requiredButtonsPerRow)
                {
                    // Decrease in buttons perRow
                    if (startingButtonsPerRow > requiredButtonsPerRow)
                    {
                        // Remove spare buttons from every row (active or not)
                        foreach (var kvp in buttonUIDsByRow)
                        {
                            // Remove the excess buttons and add them to the unparented buttons
                            for (int i = requiredButtonsPerRow; i < startingButtonsPerRow; i++)
                            {
                                UIDynamicButton button = kvp.Value.Last();
                                kvp.Value.Remove(button);
                                unparentedButtonUIDs.Add(button);
                                button.gameObject.SetActive(false);
                                UIDButtonToRVGEDict.Remove(button);
                                activeButtonUIDs.Remove(button);
                            }
                        }
                    }
                    // Increase in buttons per Row
                    else if (startingButtonsPerRow < requiredButtonsPerRow)
                    {
                        HashSet<GameObject> recycledRows = new HashSet<GameObject>();
                        // Cycle through every row
                        foreach (var kvp in buttonUIDsByRow)
                        {
                            // Check row is not tagged for recycling
                            if (!recycledRows.Contains(kvp.Key))
                            {
                                // Cycle through missing button
                                for (int i = startingButtonsPerRow; i < requiredButtonsPerRow; i++)
                                {
                                    // Check if there are no spare unparented buttons available
                                    if (unparentedButtonUIDs.Count == 0)
                                    {
                                        // Attempt to recycle an inactiveRow if it exists
                                        if (inactiveResourceRowGOs.Count > 0)
                                        {
                                            var rowGO = inactiveResourceRowGOs.First();

                                            // Remove all buttons from the inactive row
                                            // and add them to the unparented buttons
                                            var buttonsList = buttonUIDsByRow[rowGO];
                                            foreach (var button in buttonsList)
                                            {
                                                unparentedButtonUIDs.Add(button);
                                                UIDButtonToRVGEDict.Remove(button);
                                            }
                                            buttonsList.Clear();
                                            recycledRows.Add(rowGO);
                                            inactiveResourceRowGOs.Remove(rowGO);

                                        }
                                        // If rows are due to be removed, then recycle a row
                                        // and add the buttons to the unparented buttons
                                        else if (newRowsRequired < 0)
                                        {
                                            GameObject rowGO = activeResourceRowGOs.Last();

                                            activeResourceRowGOs.Remove(rowGO);
                                            var buttonsList = buttonUIDsByRow[rowGO];
                                            foreach (UIDynamicButton button in buttonsList)
                                            {
                                                activeButtonUIDs.Remove(button);
                                                unparentedButtonUIDs.Add(button);
                                                UIDButtonToRVGEDict.Remove(button);
                                                button.gameObject.SetActive(false);
                                            }
                                            buttonsList.Clear();
                                            recycledRows.Add(rowGO);
                                            rowGO.SetActive(false);
                                            newRowsRequired++;
                                        }
                                    }
                                    // If unparentedButtons exist then take one and add it to the row
                                    if (unparentedButtonUIDs.Count > 0)
                                    {
                                        UIDynamicButton button = unparentedButtonUIDs.Last();
                                        unparentedButtonUIDs.Remove(button);
                                        kvp.Value.Add(button);
                                        button.gameObject.SetActive(true);
                                    }
                                    else
                                    {
                                        kvp.Value.Add(CreateResourceButtonInRow(kvp.Key, i));
                                    }
                                }
                            }
                        }
                        foreach (var rowGO in recycledRows) buttonUIDsByRow.Remove(rowGO);
                        recycledRows.Clear();
                    }

                    activeButtonUIDs.Clear();
                    foreach (var rowGO in activeResourceRowGOs) ResizeResourceButtonRow(rowGO, true);
                    foreach (var rowGO in inactiveResourceRowGOs) ResizeResourceButtonRow(rowGO);
                }
            }

            if (newRowsRequired > 0)
            {
                for (int i = 0; i < newRowsRequired; i++)
                {
                    if (inactiveResourceRowGOs.Count == 0)
                    {
                        GameObject rowGO = UI.AddChildGOImage(buttonListContentGO, new Color(0f, 0.8f, 0f,0f), AnchorPresets.hStretchTop, parentBrowser.browserScreenWidth-20f, resouceButtonWidth + resourceButtonlabelHeight + 2, Vector2.zero);
                        activeResourceRowGOs.Add(rowGO);
                        List<UIDynamicButton> buttonsInRow = new List<UIDynamicButton>();
                        buttonUIDsByRow.Add(rowGO, buttonsInRow);
                        for (int column = 0; column < requiredButtonsPerRow; column++)
                        {
                            UIDynamicButton resourceButton = CreateResourceButtonInRow(rowGO, column);
                            buttonsInRow.Add(resourceButton);
                            activeButtonUIDs.Add(resourceButton);
                        }
                    }
                    else
                    {
                        GameObject rowGO = inactiveResourceRowGOs.Last();
                        inactiveResourceRowGOs.Remove(rowGO);
                        rowGO.SetActive(true);
                        activeResourceRowGOs.Add(rowGO);
                        foreach (UIDynamicButton button in buttonUIDsByRow[rowGO])
                        {
                            activeButtonUIDs.Add(button);
                        }
                        
                    }

                }
            }
            else if (newRowsRequired < 0)
            {
                for (int i = 0; i < Math.Abs(newRowsRequired); i++)
                {
                    GameObject rowGO = activeResourceRowGOs.Last();

                    activeResourceRowGOs.Remove(rowGO);
                    foreach (UIDynamicButton button in buttonUIDsByRow[rowGO]) activeButtonUIDs.Remove(button);
                    inactiveResourceRowGOs.Add(rowGO);
                    rowGO.SetActive(false);
                }
            }
        }
        private void ResizeResourceButtonRow(GameObject rowGO, bool addtoActiveButtons = false)
        {
            UI.ResizeHeight(rowGO, resouceButtonWidth + resourceButtonlabelHeight + 2);
            int column = 0;
            foreach (var button in buttonUIDsByRow[rowGO])
            {
                button.transform.SetParent(rowGO.transform);
                button.buttonImage.enabled = true;
                button.button.enabled = true;

                UI.ResizeRecTransform(button.gameObject, resouceButtonWidth, resouceButtonWidth + resourceButtonlabelHeight, (column * resouceButtonWidth) + 2 + (column * 3), 0, AnchorPresets.middleLeft);

                Transform textTransform = button.button.transform.Find("Text");
                UI.ResizeRecTransform(textTransform.gameObject, resouceButtonWidth * 2, (resourceButtonlabelHeight - 1) * 2, 0, 0, AnchorPresets.bottomLeft);

                Transform imageTransform = button.button.transform.Find("ButtonSubImage");
                UI.ResizeRecTransform(imageTransform.gameObject, resouceButtonWidth, resouceButtonWidth, 0, -2f, AnchorPresets.topMiddle);

                if (addtoActiveButtons) activeButtonUIDs.Add(button);

                column++;
            }
        }
        private UIDynamicButton CreateResourceButtonInRow(GameObject rowGO, int column)
        {
            UIDynamicButton button = UI.CreateUIButton(rowGO, resouceButtonWidth, resouceButtonWidth + resourceButtonlabelHeight, "", 20, (column * resouceButtonWidth) + 2 + (column * 3), 0, AnchorPresets.middleLeft);

            GameObject newImageGO = UI.AddChildGOImage(button.button.gameObject, Color.white, AnchorPresets.topMiddle, button.height - 18f, button.height - 18f, new Vector2(0f, -2f));
            newImageGO.name = "ButtonSubImage";

            GameObject hiddenIconImageGO = UI.AddChildGOImage(button.button.gameObject, Color.red, AnchorPresets.bottomRight, 7f, 7f, new Vector2(0f, 0f));
            hiddenIconImageGO.name = "HiddenIconImage";
            hiddenIconImageGO.SetActive(false);
            GameObject favIconImageGO = UI.AddChildGOImage(button.button.gameObject, new Color(0f,0.7f,0f,1f), AnchorPresets.bottomRight, 7f, 7f, new Vector2(0f, 7.5f));
            favIconImageGO.name = "FavIconImage";
            favIconImageGO.SetActive(false);

            var newIconGO = UI.CreateBasicTextBox(button.button.gameObject, new JSONStorableString("newIcon", "NEW"), 30, 15, 20,-1,-2, anchorPreset: AnchorPresets.topRight, backgoundColor: Color.red);
            newIconGO.name = "NewIcon";
            newIconGO.SetActive(false);

            var updatedIconGO = UI.CreateBasicTextBox(button.button.gameObject, new JSONStorableString("updatedIcon", "UPDATED"), 56, 15, 20, -1, -2, anchorPreset: AnchorPresets.topRight, backgoundColor: Color.red);
            updatedIconGO.name = "UpdatedIcon";
            updatedIconGO.SetActive(false);

            var loadingIconGO = UI.CreateBasicTextBox(button.button.gameObject, new JSONStorableString("loadingIcon", "LOADING..."), 60, 20, 21, 0, 0, anchorPreset: AnchorPresets.centre, backgoundColor: Color.blue);
            loadingIconGO.name = "LoadingIcon";
            loadingIconGO.SetActive(false);

            Transform textTransform = button.button.transform.Find("Text");
            UI.ResizeRecTransform(textTransform.GetComponent<RectTransform>(), resouceButtonWidth * 2, (resourceButtonlabelHeight - 1) * 2, 0, 0, AnchorPresets.bottomLeft);
            button.buttonText.color = Color.black;
            button.buttonImage.sprite = null;
            button.buttonText.resizeTextForBestFit = true;
            button.buttonText.resizeTextMinSize = 15;
            button.buttonText.resizeTextMaxSize = 21;

            //            activeButtonUIDs.Add(button);

            ResourceButtonPointerEnterExitClick context = button.button.gameObject.AddComponent<ResourceButtonPointerEnterExitClick>();
            context.Init(button, UIDButtonToRVGEDict, parentBrowser.helpTextJSS,parentBrowser);


            return button;
        }

        private bool tempMultiMode =false;
        private int previousSelecteMode;

        private bool multiSelectKeyPressed { get
            {
                return Input.GetKey(KeyCode.LeftShift) || Input.GetKey(KeyCode.RightShift) || Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt) || Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl);
            } }
        public void Update()
        {
            if (!parentBrowser.isVREnabled)
            {                
                if (multiSelectKeyPressed)
                {
                    if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi)
                    {
                        previousSelecteMode = browserFiltersUI.selectModeJSEnum.val;
                        browserFiltersUI.selectModeJSEnum.val = BrowserSelectMode.multi;
                        tempMultiMode = true;
                    }
                    
                }
                else if (tempMultiMode)
                {
                    browserFiltersUI.selectModeJSEnum.val = previousSelecteMode;
                    tempMultiMode = false;
                }
            }

        }

        private void SelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton buttonUID, bool suppressQuickActivate=false)
        {
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi)
            {
                parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.instant && !suppressQuickActivate) Globals.mvrScript.StartCoroutine(rvge.Activate(parentBrowser.resourceActionsUI.primaryActionUITab));
            }
            else if (!parentBrowser.isVREnabled)
            {
                if (multiSelectKeyPressed) tempMultiMode = false;
                if (Input.GetKey(KeyCode.LeftShift)) parentBrowser.resourceSelectionsUI.LeftShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.RightShift)) parentBrowser.resourceSelectionsUI.RightShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl)) parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
                else if (Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt)) parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                else
                {
                    parentBrowser.resourceSelectionsUI.DeSelectAll(rvge.resourceType);
                    parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                }
            }
            else parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
        }

        public void RepositionNavButtons()
        {
            UI.ResetAnchorPosition(firstPageUIDButton.gameObject, navButtonsHOffset, navButtonsVOffset);
            UI.ResetAnchorPosition(prevPageUIDButton.gameObject, navButtonsHOffset + 15, navButtonsVOffset);
            UI.ResetAnchorPosition(nextPageUIDButton.gameObject, navButtonsHOffset + 30, navButtonsVOffset);
            UI.ResetAnchorPosition(lastPageUIDButton.gameObject, navButtonsHOffset + 45, navButtonsVOffset);

            UI.ResetAnchorPosition(incResourceButtonSizesUIDButton.gameObject, resoureButtonSizeHOffset , navButtonsVOffset);
            UI.ResetAnchorPosition(decResourceButtonSizesUIDButton.gameObject, resoureButtonSizeHOffset+20, navButtonsVOffset);

            UI.ResetAnchorPosition(resourceCountTextBox.gameObject, resourceCountHOffset, resourceCountVOffset);

            UI.ResetAnchorPosition(clearMultiSelectResourcesUIDButton.gameObject, navButtonsHOffset, navButtonsVOffset-17f);
            UI.ResetAnchorPosition(selectAllResourcesUIDButton.gameObject, navButtonsHOffset+53f, navButtonsVOffset - 17f);
            SetMultiSelectClear();
        }

        public void SetMultiSelectClear()
        {
          if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi && parentBrowser.resourceSelectionsUI.SelectionCount(browserFiltersUI.resourceType) > 1) clearMultiSelectResourcesUIDButton.gameObject.SetActive(true);
           else clearMultiSelectResourcesUIDButton.gameObject.SetActive(false);

            if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) selectAllResourcesUIDButton.gameObject.SetActive(true);
            else selectAllResourcesUIDButton.gameObject.SetActive(false);
        }
    }



    public class BrowserFiltersUI
    {
        private JSONStorableEnumStringChooser resourceTypeJSEnum ;
        public int resourceType { get
            {
                return resourceTypeJSEnum.val;
            }
            set {
                resourceTypeJSEnum.val = value;
            }
        }
        public string resourceTypeName
        {
            get
            {
                return resourceTypeJSEnum.displayVal;
            }
        }


        public ResourceBrowserFilterSettings currentFilterSettings { get { return filterSettingsPerResourceType[resourceType]; } }
        public JSONStorableMultiEnumStringChooser creatorJSMultiEnum { get { return filterSettingsPerResourceType[resourceType].creatorJSMultiEnum; } }

        public JSONStorableString folderFilterJSS { get { return filterSettingsPerResourceType[resourceType].folderFilterJSS; } }

        public Folder selectedFolderFilter {
            get { return filterSettingsPerResourceType[resourceType].selectedFolderFilter; }
            set { filterSettingsPerResourceType[resourceType].selectedFolderFilter = value; } }

        public JSONStorableBool inludeSubFoldersJSB { get { return filterSettingsPerResourceType[resourceType].inludeSubFoldersJSB; } }
        public JSONStorableString resourceNameFilterJSS { get { return filterSettingsPerResourceType[resourceType].resourceNameFilterJSS; } }

        public JSONStorableBool showHiddenJSB { get { return filterSettingsPerResourceType[resourceType].showHiddenJSB; } }
        public JSONStorableBool onlyFavouritesJSB { get { return filterSettingsPerResourceType[resourceType].onlyFavouritesJSB; } }
        public JSONStorableBool onlyNewJSB { get { return filterSettingsPerResourceType[resourceType].onlyNewJSB; } }

        public JSONStorableMultiEnumStringChooser tagsFilterJSMultiEnum { get { return filterSettingsPerResourceType[resourceType].tagsFilterJSMultiEnum; } }
        public JSONStorableEnumStringChooser licenseFilterJSEnum { get { return filterSettingsPerResourceType[resourceType].licenseFilterJSEnum; } }

        public JSONStorableEnumStringChooser selectModeJSEnum { get { return filterSettingsPerResourceType[resourceType].selectModeJSEnum; } }
        public JSONStorableEnumStringChooser sortModeJSEnum { get { return filterSettingsPerResourceType[resourceType].sortModeJSEnum; } }

        public string resourceNameFilterQuick { get
            { return filterSettingsPerResourceType[resourceType].resourceNameFilterQuick; }
            set { filterSettingsPerResourceType[resourceType].resourceNameFilterQuick = value; }
        }

        public bool isCombiTagFilterActive { get
            {
                return filterSettingsPerResourceType[resourceType].combiTagFilterEnabled;
            }
            set
            {
                filterSettingsPerResourceType[resourceType].combiTagFilterEnabled = value;
            }
        }

        public CombiTag combiTagFilter { get
            {
                return filterSettingsPerResourceType[resourceType].combiTag;
            }
        }

        private UIDynamicPopup resourceTypePopupUID;
        private UIDynamicButton refreshButtonUID;
        private UIDynamicPopup tagPopupUID;
        private UIDynamicButton combiTagEnableButtonUID;
        private UIDynamicButton combiTagSummaryButtonUID;
        private UIDynamicPopup creatorPopupUID;

        private UIDynamicPopup licensePopupUID;

        private UIDynamicToggle hiddenUID;
        private UIDynamicToggle onlyFavUID;
        private UIDynamicToggle onlyNewUID;
        private UIDynamicPopup sortModePopup;
        private UIDynamicPopup selectModePopup;

        private UIDynamicButton clearResourceNameFilterButton;
        private UIDynamicButton resetCreatorFilterButton;
        private UIDynamicButton openFolderBrowserButtonUID;

        private UIDynamicTextField resourceNameTextUID;
        private UIDynamicTextField folderTextUID;
        private UIDynamicToggle incSubFoldersToggleUID;

        public Dictionary<int, ResourceBrowserFilterSettings> filterSettingsPerResourceType = new Dictionary<int, ResourceBrowserFilterSettings>();

        private GameObject filtersBoxGO;
        private GameObject folderBrowserBoxGO;
        private GameObject combiTagFilterBoxGO;

        private FolderBrowserUI folderBrowserUI;
        private CombiTagFilterUI combiTagFilterUI;

        public BrowserUI parentBrowser;
        public BrowserFiltersUI(BrowserUI _parentBrowser)
        {
            parentBrowser = _parentBrowser;
            resourceTypeJSEnum = new JSONStorableEnumStringChooser("resourceType", ResourceType.enumManifestName, Settings.defaultResourceTypeJESC.val, "Resource Type", ResourceTypeFilterCallback);

            List<int> resourceTypesToExclude = new List<int>();
            if (!PC.patreonContentEnabled)
            {
                resourceTypesToExclude = EnumManifest.GetEnumsList(ResourceType.enumManifestName);
                resourceTypesToExclude.Remove(ResourceType.scene);
                resourceTypesToExclude.Remove(ResourceType.presetAppearance);
            }

            resourceTypeJSEnum.SetEnumChoices(ResourceType.enumManifestName, resourceTypesToExclude);

            foreach (int resourceType in EnumManifest.GetEnumsList(ResourceType.enumManifestName))
            {
                filterSettingsPerResourceType.Add(resourceType, new ResourceBrowserFilterSettings(this, resourceType));
            }
        }

        public void InitUI(GameObject _filtersBoxGO)
        {
            filtersBoxGO = _filtersBoxGO;
            if (resourceTypePopupUID != null) GameObject.Destroy(resourceTypePopupUID.gameObject);
            if (refreshButtonUID != null) GameObject.Destroy(refreshButtonUID.gameObject);
            if (openFolderBrowserButtonUID != null) GameObject.Destroy(openFolderBrowserButtonUID.gameObject);
            
            if (creatorPopupUID != null) GameObject.Destroy(creatorPopupUID.gameObject);
            if (folderTextUID != null) GameObject.Destroy(folderTextUID.gameObject);
            if (incSubFoldersToggleUID != null) GameObject.Destroy(incSubFoldersToggleUID.gameObject);
            if (licensePopupUID != null) GameObject.Destroy(licensePopupUID.gameObject);
            if (hiddenUID != null) GameObject.Destroy(hiddenUID.gameObject);
            if (onlyFavUID != null) GameObject.Destroy(onlyFavUID.gameObject);
            if (onlyNewUID != null) GameObject.Destroy(onlyNewUID.gameObject);
            if (sortModePopup != null) GameObject.Destroy(sortModePopup.gameObject);
            if (selectModePopup != null) GameObject.Destroy(selectModePopup.gameObject);
            if (clearResourceNameFilterButton != null) GameObject.Destroy(clearResourceNameFilterButton.gameObject);
            if (resourceNameTextUID != null) GameObject.Destroy(resourceNameTextUID.gameObject);
            if (resetCreatorFilterButton != null) GameObject.Destroy(resetCreatorFilterButton.gameObject);
            if (folderBrowserBoxGO != null) GameObject.Destroy(folderBrowserBoxGO);
            if (combiTagFilterBoxGO != null) GameObject.Destroy(combiTagFilterBoxGO);

            resourceTypePopupUID = UI.CreatePopup(filtersBoxGO, resourceTypeJSEnum, 160f, 30f, null, 20, 0f, 0f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(resourceTypePopupUID, "Select a type of VAM resource to apply as a filter to the Browser results.");

            RefreshCombiTagUI();

            creatorPopupUID = UI.CreateFilterablePopup(filtersBoxGO, creatorJSMultiEnum, 160f, 30f, null, 20, 330f, 0f, 300f, AnchorPresets.topLeft);
            creatorPopupUID.popup.onOpenPopupHandlers += RefreshCurrentCreatorFilters;
            parentBrowser.ApplyHelp(creatorPopupUID, "Select a Creator to filter Browser results by a specific VAR package creator. Non VAR packaged resources can be selected with the 'Local' option.");

            licensePopupUID = UI.CreatePopup(filtersBoxGO, licenseFilterJSEnum, 115f, 30f, null, 20, -35f, 0f, AnchorPresets.topRight);
            licensePopupUID.popup.onOpenPopupHandlers += RefreshCurrentLicenseFilters;
            parentBrowser.ApplyHelp(licensePopupUID, "Select a License type to filter Browser results by the license type applicable to the containing VAR package.");


            refreshButtonUID = UI.CreateUIButton(filtersBoxGO, 33f, 30f, "Rescan for new VARs", 16, 0f, 0f, AnchorPresets.topRight);
            refreshButtonUID.button.image.sprite = null;
            refreshButtonUID.buttonColor = Color.blue;
            refreshButtonUID.textColor = Color.white;

            refreshButtonUID.button.onClick.AddListener(delegate
            {
                SuperController.singleton.StartCoroutine(RefreshContents(refreshButtonUID));
            });
            parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new VARs in the AddonPackages folder. Also updates Hidden & Favourite states of resources changed in VAM UI.");


            selectModePopup = UI.CreatePopup(filtersBoxGO, selectModeJSEnum, 120f, 20f, null, 20, 0f, -30f, AnchorPresets.topLeft, labelProportion: 0.55f);
            parentBrowser.ApplyHelp(selectModePopup, "Instant Mode: the default action (e.g. Load Scene) will be performed instantly when selecting a resource in the browser. Single Mode: only one resource is selectable at a time. Multi Mode: many resources can be selected and modified simultaneously.");

            hiddenUID = UI.CreateToggle(filtersBoxGO, showHiddenJSB, 85f, 20f, "Show Hidden", 20, 120f, -30f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(hiddenUID, "Show resources that are marked as Hidden in the resource browser.");
            onlyFavUID = UI.CreateToggle(filtersBoxGO, onlyFavouritesJSB, 90f, 20f, "Only Favorites", 20, 205f, -30f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(onlyFavUID, "Only show resources that are marked as Favourite in the resource browser.");

            onlyNewUID = UI.CreateToggle(filtersBoxGO, onlyNewJSB, 70f, 20f, "Only New", 20, 295f, -30f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(onlyNewUID, "Only show resources that are marked as New or Updated in the resource browser.");

            resourceNameTextUID = UI.CreateTextBox(filtersBoxGO, resourceNameFilterJSS, 150f, 20f, 20, 365f, -30f, AnchorPresets.topLeft, true, InputField.LineType.MultiLineNewline, "Search Resource...", vScrollBarWidth: 0f, textEntryChangeCallback: ResourceNameFilterCallback);
            resourceNameTextUID.UItext.alignment = TextAnchor.MiddleLeft;
            SetResourceNameFilterHelpText();

            clearResourceNameFilterButton = UI.CreateUIButton(filtersBoxGO, 16f, 16f, "X", 20, 497f, -32f, AnchorPresets.topLeft);
            clearResourceNameFilterButton.button.onClick.AddListener(delegate { resourceNameFilterJSS.val = ""; });
            parentBrowser.ApplyHelp(clearResourceNameFilterButton, "Clear the Resource Name search field of user entered text.");
            if (resourceNameFilterJSS.val == "") clearResourceNameFilterButton.gameObject.SetActive(false);

            resetCreatorFilterButton = UI.CreateUIButton(filtersBoxGO, 16f, 16f, "X", 20, 472.5f, -13.3f, AnchorPresets.topLeft);
            resetCreatorFilterButton.button.onClick.AddListener(delegate { creatorJSMultiEnum.val = Settings.resourceSettings[resourceType].defaultFilterSettings.creatorJSMultiEnum.val; });
            parentBrowser.ApplyHelp(resetCreatorFilterButton, "Reset the Creator/Local filter to it's default value");
            if (creatorJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.mainValFlag || creatorJSMultiEnum.val == Settings.resourceSettings[resourceType].defaultFilterSettings.creatorJSMultiEnum.val) resetCreatorFilterButton.gameObject.SetActive(false);

            sortModePopup = UI.CreatePopup(filtersBoxGO, sortModeJSEnum, 125f, 20f, null, 20, 0f, -30f, AnchorPresets.topRight, labelProportion: 0.35f, popupRowHeight:24f);
            parentBrowser.ApplyHelp(sortModePopup, "Order Browser resource results Alphabetically, by Package and Folder or by Date. New dates are when items are first detected. For packaged items, the Updated date is when the highest VAR version was detected. Updated and New dates are the same for local resources.");

            combiTagFilterBoxGO = UI.AddChildGOImage(filtersBoxGO, new Color(0.2f, 0.2f, 0.2f, 0.95f), AnchorPresets.topRight, 450f, 275f, new Vector2(-160f, -30f));

            combiTagFilterUI = new CombiTagFilterUI(combiTagFilterBoxGO, parentBrowser);
            combiTagFilterBoxGO.SetActive(false);

            folderBrowserBoxGO = UI.CreateVScrollableContent(filtersBoxGO, new Color(0.2f, 0.2f, 0.2f, 0.95f), AnchorPresets.topRight, 350f, 500f, new Vector2(-20f, -30f));
            folderBrowserBoxGO.SetActive(false);
            folderBrowserUI = new FolderBrowserUI(folderBrowserBoxGO, FolderBrowserType.filterSelect, FolderFilterSelected, parentBrowser);

            folderTextUID = UI.CreateTextBox(filtersBoxGO, folderFilterJSS, 100f, 15f, 20, -50f, 0f, AnchorPresets.topRight, false, InputField.LineType.SingleLine, vScrollBarWidth: 0f);
            folderTextUID.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
            folderTextUID.UItext.alignment = TextAnchor.MiddleLeft;
            parentBrowser.ApplyHelp(folderTextUID, "Current folder filter: " + selectedFolderFilter.GetFolderPath());

            incSubFoldersToggleUID = UI.CreateToggle(filtersBoxGO, inludeSubFoldersJSB, 100f, 15f, "Include subfolders", 20, -50f, -15f, AnchorPresets.topRight, 15f);
            parentBrowser.ApplyHelp(incSubFoldersToggleUID, "Include resources from sub-folders of the selected folder.");

            openFolderBrowserButtonUID = UI.CreateUIButton(filtersBoxGO, 15f, 30f, "V", 20, -35f, 0, AnchorPresets.topRight, false);
            parentBrowser.ApplyHelp(openFolderBrowserButtonUID, "Open the Folder Browser to select a folder for filtering resources");
            openFolderBrowserButtonUID.button.onClick.AddListener(delegate
            {
                if (!folderBrowserBoxGO.activeSelf) SetFolderBrowserActive();
                else SetFolderBrowserInActive();
            });
            RefreshLicenseFolderFilterVisibility();
            RefreshRescanButton();
        }

        public void ToggleCombiTagUI()
        {
            if (combiTagFilterBoxGO.activeSelf) combiTagFilterBoxGO.SetActive(false);
            else
            {
                combiTagFilterBoxGO.SetActive(true);
                combiTagFilterUI.RefreshUI(resourceType, combiTagFilter);
            }
        }

        public void RefreshCombiTagUI()
        {
            if (tagPopupUID != null) GameObject.Destroy(tagPopupUID.gameObject);
            if (combiTagEnableButtonUID != null) GameObject.Destroy(combiTagEnableButtonUID.gameObject);
            if (combiTagSummaryButtonUID !=null) GameObject.Destroy(combiTagSummaryButtonUID.gameObject);

            if (isCombiTagFilterActive && PC.patreonContentEnabled)
            {
                combiTagSummaryButtonUID = UI.CreateUIButton(filtersBoxGO, 145f, 27f, combiTagFilter.description, 20,165f,-2f, AnchorPresets.topLeft,true);
                combiTagSummaryButtonUID.button.onClick.AddListener(delegate {
                    ToggleCombiTagUI();

                });
                parentBrowser.ApplyHelp(combiTagSummaryButtonUID, "Toggle disaplying the Tag Combination filter setup window.");


                combiTagEnableButtonUID = UI.CreateUIButton(filtersBoxGO, 15f, 30f, "X", 20, 165f + 145f, 0f, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(combiTagEnableButtonUID, "Cancel tag combination filtering and use single tag filtering");
                combiTagEnableButtonUID.button.onClick.AddListener(delegate
                {
                    isCombiTagFilterActive = false;
                    combiTagFilterBoxGO.SetActive(false);
                    RefreshCombiTagUI();
                    parentBrowser.RefreshFilterResults();
                });
               
            }
            else
            {

                tagPopupUID = UI.CreateFilterablePopup(filtersBoxGO, tagsFilterJSMultiEnum, 145f, 30f, null, 20, 165f, 0f, 150f, AnchorPresets.topLeft);
                tagPopupUID.popup.onOpenPopupHandlers += RefreshCurrentTagFilters;
                parentBrowser.ApplyHelp(tagPopupUID, "Select a Tag to apply as a filter to the Browser results.");

                if (PC.patreonContentEnabled)
                {
                    combiTagEnableButtonUID = UI.CreateUIButton(filtersBoxGO, 15f, 30f, "V", 20, 165f + 145f, 0f, AnchorPresets.topLeft);
                    parentBrowser.ApplyHelp(combiTagEnableButtonUID, "Filter resources using a combination of Tags");
                    combiTagEnableButtonUID.button.onClick.AddListener(delegate
                    {
                        isCombiTagFilterActive = true;
                        if (folderBrowserBoxGO.activeSelf) SetFolderBrowserInActive();

                        combiTagFilterBoxGO.SetActive(true);
                        combiTagFilterUI.RefreshUI(resourceType, combiTagFilter);
                        RefreshCombiTagUI();
                        parentBrowser.RefreshFilterResults();
                    });
                }

            }
        }

        public void SetFolderBrowserActive()
        {
            selectedFolderFilter.ExpandParents();
            folderBrowserUI.SetNormalFolderMode();
            folderBrowserUI.RefreshUI(resourceType);
            folderBrowserBoxGO.SetActive(true) ;
        }
        public void SetFolderBrowserInActive()
        {
            folderBrowserBoxGO.SetActive(false);
        }

        public void FolderFilterSelected(Folder selectedFolder)
        {
            folderFilterJSS.val = selectedFolder.folderName;
            selectedFolderFilter = selectedFolder;
            if (!Settings.keepFolderBrowserOpenJSB.val) folderBrowserBoxGO.SetActive(false);
            if (selectedFolder.isRootFolder && folderFilterJSS.val.Length > 16) folderTextUID.UItext.alignment = TextAnchor.MiddleRight;
            else folderTextUID.UItext.alignment = TextAnchor.MiddleLeft;

            parentBrowser.ApplyHelp(folderTextUID, "Filter to local folder: " + selectedFolderFilter.GetFolderPath());

            currentFilterSettings.RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
        }

        private void RefreshLicenseFolderFilterVisibility()
        {
            licensePopupUID.gameObject.SetActive(false);
            folderTextUID.gameObject.SetActive(false);
            incSubFoldersToggleUID.gameObject.SetActive(false);
            openFolderBrowserButtonUID.gameObject.SetActive(false);
            licensePopupUID.popup.visible = false;


            if (creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag || creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs)
            {
                licensePopupUID.gameObject.SetActive(true);
                folderBrowserBoxGO.SetActive(false);
            }
            else if (creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.localOnly)
            {
                folderTextUID.gameObject.SetActive(true);
                incSubFoldersToggleUID.gameObject.SetActive(true);
                openFolderBrowserButtonUID.gameObject.SetActive(true);
            }
            else folderBrowserBoxGO.SetActive(false);
        }

        public void SetResourceNameFilterHelpText()
        {
            if (Settings.resourceNameSearchModeJSESC.val == ResourceNameSearchMode.StartsWith) parentBrowser.ApplyHelp(resourceNameTextUID, "Filter resources where the resource name starts with the specified search text");
            else if (Settings.resourceNameSearchModeJSESC.val == ResourceNameSearchMode.Contains) parentBrowser.ApplyHelp(resourceNameTextUID, "Filter resources where the resource name contains the specified search text");
            else parentBrowser.ApplyHelp(resourceNameTextUID, "Filter resources where the resource name or its full file path contains the specified search text");
        }

        private IEnumerator RefreshContents(UIDynamicButton refreshButtonUID)
        {
            string tempLabel = refreshButtonUID.label;
            Color buttonCol = refreshButtonUID.buttonColor;
            refreshButtonUID.label = "Scan\n-ing";
            refreshButtonUID.buttonColor = Color.cyan;
            refreshButtonUID.textColor = Color.white;

            yield return new WaitForSeconds(0.01f);

            DateTime scanTime = DateTime.Now;

            if ((creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag && creatorJSMultiEnum.valTopEnum != CreatorFilterChoices.localOnly) || creatorJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                SuperController.singleton.RescanPackages();
                HashSet<string> varFiles = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var");
                VARPackageManifest.RefreshManifestWithNewVARs(varFiles, scanTime, new HashSet<int>());
                VARPackageManifest.RefreshVARUserFilePrefs();
            }

            if (creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag && (creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.localOnly || creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.all)) ResourceManifest.LocalResourcesInit(scanTime, new HashSet<int>());

            parentBrowser.RefreshFilterResults(true);

            refreshButtonUID.buttonText.text = tempLabel;
            refreshButtonUID.buttonColor = Color.blue;
        }

        private void RefreshCurrentTagFilters()
        {
            currentFilterSettings.RefreshTagFilters();
        }
        private void RefreshCurrentCreatorFilters()
        {
            currentFilterSettings.RefreshCreatorFilters();
        }

        private void RefreshCurrentLicenseFilters()
        {
            currentFilterSettings.RefreshLicenseFilters();
        }

        private void RefreshRescanButton()
        {
            if (creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag || creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs)
            {
                refreshButtonUID.label = "Rescan for new VARs";
                parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new VARs in the AddonPackages folder. Also updates Hidden & Favourite states of resources changed in VAM UI.");
            }
            else
            {
                if (creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.all)
                {
                    refreshButtonUID.label = "Rescan Local & VARs";
                    parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new VARs in the AddonPackages folder and new (or removed) files in the local Saves and Custom folders. Also updates Hidden & Favourite states of resources changed in VAM UI.");
                }
                else
                {
                    refreshButtonUID.label = "Rescan Local files";
                    parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new (or removed) files in the local Saves and Custom folders. Also updates Hidden & Favourite states of resources changed in VAM UI.");
                }
            }
        }

        public void CreatorsTopValCallback(int i)
        {
            RefreshLicenseFolderFilterVisibility();
            RefreshRescanButton();


            if (creatorJSMultiEnum.valType!=JSONStorableMultiEnumStringChooser.mainValFlag || creatorJSMultiEnum.val == Settings.resourceSettings[resourceType].defaultFilterSettings.creatorJSMultiEnum.val) resetCreatorFilterButton.gameObject.SetActive(false);
            else resetCreatorFilterButton.gameObject.SetActive(true);
            currentFilterSettings.RefreshLicenseFilters();
            parentBrowser.RefreshFilterResults(true);
        }
        public void CreatorsMainValCallback(string i)
        {
            RefreshRescanButton();
            currentFilterSettings.RefreshLicenseFilters();

            if (creatorJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.mainValFlag || creatorJSMultiEnum.val == Settings.resourceSettings[resourceType].defaultFilterSettings.creatorJSMultiEnum.val) resetCreatorFilterButton.gameObject.SetActive(false);
            else resetCreatorFilterButton.gameObject.SetActive(true);

            parentBrowser.RefreshFilterResults(true);
            RefreshLicenseFolderFilterVisibility();
        }

        public void TagTopValCallback(int i)
        {
            if (i == TagGroups.hiddenOnly)
            {
                showHiddenJSB.valNoCallback = true;
                ShowHiddenFilterCallback(true);
                return;
            }
            currentFilterSettings.RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
           
        }
        public void TagMainValCallback(string i)
        {
            currentFilterSettings.RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
            
        }

        public void ResourceNameFilterCallback(string i)
        {
            resourceNameFilterQuick = i.ToLower();
            if (resourceNameFilterQuick == "") clearResourceNameFilterButton.gameObject.SetActive(false);
            else clearResourceNameFilterButton.gameObject.SetActive(true);
            parentBrowser.RefreshFilterResults(true);
            
        }
        public void ResourceTypeFilterCallback(int value)
        {
            InitUI(filtersBoxGO);
            parentBrowser.RefreshFilterResults(true);
            parentBrowser.RefreshBrowserUILayout();

            parentBrowser.resourceActionsUI.RefreshUI();
            
            parentBrowser.resourceActionsUI.primaryActionUITab.RefreshTargetAtomPopup(true);
            parentBrowser.resourceSelectionsUI.RePopulateSelectionsBox();

        }
        public void ShowHiddenFilterCallback(bool value)
        {
            //            if (!value) VARPackageManifest.RefreshVARUserFilePrefs();
            currentFilterSettings.RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
            parentBrowser.resourceActionsUI.RefreshUI();
        }
        public void OnlyFavFilterCallback(bool value)
        {
            //            if (value) VARPackageManifest.RefreshVARUserFilePrefs();
            currentFilterSettings.RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
            parentBrowser.resourceActionsUI.RefreshUI();
        }

        public void SubFoldersCallback(bool value)
        {
            parentBrowser.RefreshFilterResults(true);
            parentBrowser.resourceActionsUI.RefreshUI();
        }
        public void OnlyNewFilterCallback(bool value)
        {
            currentFilterSettings.RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
            parentBrowser.resourceActionsUI.RefreshUI();
        }

        public void SortByCallback(string i)
        {
            parentBrowser.RefreshFilterResults();
        }
        public void SelectModeCallback(int i)
        {
            parentBrowser.resourceSelectionsUI.SwitchSelectMode(resourceType,i);
            if (parentBrowser.resourceActionsUI.currentActionUITabType != ActionUITabType.infoTab) parentBrowser.resourceActionsUI.RefreshUI();
            parentBrowser.RefreshBrowserUILayout();
        }

        public void LicenseCallback(int i)
        {
            parentBrowser.RefreshFilterResults();
        }
    }
}
